/***************************************************************************
	NIC-independent macros for the GAMMA protocol
	Read COPYRIGHT file in this directory
****************************************************************************/


#define gmin(a,b) ((a)<=(b)?(a):(b))
#define gmax(a,b) ((a)>=(b)?(a):(b))

#ifdef MORE_INSTANCES_ON_SAME_NODE
#define DELAY_FOR_SMP busywait(50)
#else
#define DELAY_FOR_SMP
#endif

#define printg(a...) printk(KERN_CRIT a)

#ifdef DRIVER_DEBUG_EVERYTHING
#define DEBUG_E(a) {a;DELAY_FOR_SMP;}
#define DEBUG(a) {a;DELAY_FOR_SMP;}
#define DEBUG_F(a) {a;DELAY_FOR_SMP;}
#define DEBUG_L(a) {a;DELAY_FOR_SMP;}
#else
#ifdef DRIVER_DEBUG
#define DEBUG_E(a)
#define DEBUG(a) {a;DELAY_FOR_SMP;}
#define DEBUG_F(a) {a;DELAY_FOR_SMP;}
#define DEBUG_L(a) {a;DELAY_FOR_SMP;}
#else
#ifdef DRIVER_DEBUG_FLWCTL_RETX
#define DEBUG_E(a)
#define DEBUG(a)
#define DEBUG_F(a) {a;DELAY_FOR_SMP;}
#define DEBUG_L(a) {a;DELAY_FOR_SMP;}
#else
#ifdef DRIVER_DEBUG_LIGHT
#define DEBUG_E(a)
#define DEBUG(a)
#define DEBUG_F(a)
#define DEBUG_L(a) {a;DELAY_FOR_SMP;}
#else
#define DEBUG_E(a)
#define DEBUG(a)
#define DEBUG_F(a)
#define DEBUG_L(a)
#endif
#endif
#endif
#endif

#define incr_modulo(x,n) if ((++(x)) == (n)) (x) = 0

/* the pte_offset_map() should be followed by a pte_unmap() if HIGHMEM in use */

#define user_space_to_phys(P,addr) \
(unsigned long)virt_to_phys( \
	(void *)( \
		(unsigned long)page_address(pte_page( \
			*pte_offset_map( \
				pmd_offset( \
					pgd_offset( \
						(P)->mm, \
						(unsigned long)(addr) \
					), \
					(unsigned long)(addr) \
				), \
				(unsigned long)(addr) \
			) \
		)) \
		| ((unsigned long)(addr) & 0x00000fff) \
	) \
)

#define recv_payload_1st_chunk(dest,skip,len) \
copy_to_user((void *)(dest),(void *)(((u32)h)+skip),len)

#define recv_payload_2nd_chunk(dest,skip1,skip2,len) \
copy_to_user((void *)(dest),(void *)(((u32)h)+skip1+skip2),len)

/*
 *  NOTE: most of what follows is NOT thread safe on SMP machines.
 *  Caller should raise spinlock.
 */

/*
 *  Shift user-space rec buffer's current ptr forward.
 *  Check for room availability.
 *  If not enough room in rec buffer, truncate the msg chunk
 *  by setting cnk_len equal to the available room.
 *  Decrement the available room by cnk_len.
 *  This macro is to prepare the download of a fresh packet into final dest.
 *  The fresh packet will be stored starting from address "rx_buffer".
 *  Check whether user-space or kernel-space buffer.
 */
#define update_ptr_to_user_recv_buffer(par_pid,port,procnum,len,rx_buffer) \
{ \
rx_buffer = (unsigned long) \
par_proc_table[par_pid]->inst_table[procnum].port_table[port].in_progress; \
if \
(par_proc_table[par_pid]->inst_table[procnum].port_table[port].in_room < len) \
len = par_proc_table[par_pid]->inst_table[procnum].port_table[port].in_room; \
par_proc_table[par_pid]->inst_table[procnum].port_table[port].in_room \
-= len; \
par_proc_table[par_pid]->inst_table[procnum].port_table[port].in_progress \
+= len; \
}

/*
 *  Get the pointer to current user-level recv buffer.
 */
#define get_ptr_to_user_recv_buffer(par_pid,port,procnum,rx_buffer) \
rx_buffer = (unsigned long) \
par_proc_table[par_pid]->inst_table[procnum].port_table[port].in_progress

/*  Set len of a just arrived msg.  */
#define set_msglen(par_pid,procnum,len) \
copy_to_user( \
	(u32 *)(par_proc_table[par_pid]->inst_table[procnum].msglen_ptr), \
	&(len), \
	sizeof(u32) \
)

#ifdef USE_DAMNICK
/*  Pre-notify a message arrival to the port semaphore.  */
#define pre_notify_arrival(par_pid,port,procnum,pre_notif_val) \
{ \
copy_from_user( \
	&pre_notif_val, \
	((unsigned long *) \
(par_proc_table[par_pid]->inst_table[procnum].pre_sem_table)+ \
par_proc_table[par_pid]->inst_table[procnum].port_table[port].semaphore), \
	sizeof(pre_notif_val) \
); \
pre_notif_val++; \
copy_to_user( \
	((unsigned long *) \
(par_proc_table[par_pid]->inst_table[procnum].pre_sem_table)+ \
par_proc_table[par_pid]->inst_table[procnum].port_table[port].semaphore), \
	&pre_notif_val, \
	sizeof(pre_notif_val) \
); \
}
#endif

#define reset_current_ptr(par_pid,port,procnum) \
{ \
par_proc_table[par_pid]->inst_table[procnum].port_table[port].in_progress \
= par_proc_table[par_pid]->inst_table[procnum].port_table[port].in_buffer; \
par_proc_table[par_pid]->inst_table[procnum].port_table[port].in_room \
= par_proc_table[par_pid]->inst_table[procnum].port_table[port].in_size; \
}

#define \
check_available_room_and_possibly_reset_current_ptr(par_pid,port,procnum) \
{ \
if (!(par_proc_table[par_pid]->inst_table[procnum].port_table[port].in_room)) \
{ \
par_proc_table[par_pid]->inst_table[procnum].port_table[port].in_progress \
= par_proc_table[par_pid]->inst_table[procnum].port_table[port].in_buffer; \
par_proc_table[par_pid]->inst_table[procnum].port_table[port].in_room \
= par_proc_table[par_pid]->inst_table[procnum].port_table[port].in_size; \
} \
}

#define read_credit(node,credit) \
for (;;) { \
	unsigned long iii; \
	if (node != BROADCAST) { \
		credit = local_credit[node]; \
		break; \
	} \
	credit = 999999; \
	for (iii=0; iii<total_num_nodes; iii++) \
		if (iii != phys_whoami) \
			if (local_credit[iii] < credit) \
				credit = local_credit[iii]; \
	break; \
}

#define dec_credit(node,amount) \
for (;;) { \
	unsigned long iii; \
	if (node != BROADCAST) { \
		local_credit[node] -= amount; \
		break; \
	} \
	for (iii=0; iii<total_num_nodes; iii++) \
		if (iii != phys_whoami) \
			local_credit[iii] -= amount; \
	break; \
}


#if 0

#define FROM_SEND_ROUTINE 2

#ifdef MORE_INSTANCES_ON_SAME_NODE
#define TX_SPIN_LOCK spin_lock(&gamma_tx_lock)
#define TX_SPIN_UNLOCK spin_unlock(&gamma_tx_lock)
#define CREDIT_SPIN_TRYLOCK spin_trylock(&gamma_cr_lock)
#define CREDIT_SPIN_UNLOCK spin_unlock(&gamma_cr_lock)
#else
#define TX_SPIN_LOCK
#define TX_SPIN_UNLOCK
#define CREDIT_SPIN_TRYLOCK 1
#define CREDIT_SPIN_UNLOCK
#endif

#define wait_credit_refill(i,dest,par_pid,procnum) \
{ \
int res; \
time_586 a,b; \
i = 1; \
gamma_time(a); \
if (CREDIT_SPIN_TRYLOCK) \
for (;;) { \
	if (dest != BROADCAST) { \
		if (local_credit1[dest]) { \
			local_credit[dest] = local_credit1[dest]; \
			local_credit1[dest] = 0; \
			credit = local_credit[dest]; \
			curr_credit = credit; \
DEBUG_F( \
if (credit) { \
unsigned long i; \
printg("wait_credit_refill():  credit refilled: "); \
for (i=0;i<total_num_nodes;i++) \
	printg("%ld ", local_credit[i]); \
printg("\n"); \
}) \
			CREDIT_SPIN_UNLOCK; \
			break; \
		} \
	} \
	else { \
		unsigned long j; \
		credit = 999999; \
		for (j=0;j<total_num_nodes;j++) { \
			if (j != phys_whoami) { \
				if (local_credit1[j]) { \
					local_credit[j] = local_credit1[j]; \
					local_credit1[j] = 0; \
				} \
				if (local_credit[j] < credit) \
					credit = local_credit[j]; \
			} \
		} \
		if (credit) { \
			curr_credit = credit; \
DEBUG_F({ \
unsigned long i; \
printg("wait_credit_refill():  credit refilled: "); \
for (i=0;i<total_num_nodes;i++) \
	printg("%ld ", local_credit[i]); \
printg("\n"); \
}) \
			CREDIT_SPIN_UNLOCK; \
			break; \
		} \
	} \
	/* avoid deadlock: poll() might invoke send() in turn,  */ \
	/* so we must deassert "gamma_tx_lock" (SMP only).      */ \
	/* this however opens a window to reentrance in send(). */ \
	/* never mind, provided this macro does not reenter.    */ \
	/* that's the reason for the "gamma_cr_lock" (SMP only) */ \
	TX_SPIN_UNLOCK; \
	res = sys_gamma_poll(FROM_SEND_ROUTINE); \
	TX_SPIN_LOCK; \
	if (unlikely((res == -GAMMA_RX_SUSP) || (res == -GAMMA_ERR_MUSTYIELD))) { \
		i = 0; \
DEBUG_F(printg("wait_credit_refill():  leave wait_credit_refill because end-of-msg catched or msg for another process catched\n")) \
		CREDIT_SPIN_UNLOCK; \
		break; \
	} \
	gamma_time(b); \
	if ( \
		gamma_time_diff_ticks(b,a) \
		 > (unsigned long long)MAX_CREDIT_WAIT*CLOCK \
	) { \
		i = 0; \
DEBUG_F({ \
unsigned long i; \
printg("wait_credit_refill():  credit NOT refilled yet: "); \
for (i=0;i<total_num_nodes;i++) \
	printg("%ld ", local_credit[i]); \
printg("\n"); \
}) \
		CREDIT_SPIN_UNLOCK; \
		break; \
	} \
} \
else i = 0; \
}

#else

#define wait_credit_refill(i,dest,par_pid,procnum) \
{ \
i = 1; \
for (;;) { \
	if (dest != BROADCAST) { \
		if (local_credit1[dest]) { \
			local_credit[dest] = local_credit1[dest]; \
			local_credit1[dest] = 0; \
			credit = local_credit[dest]; \
			curr_credit = credit; \
DEBUG_F( \
if (credit) { \
unsigned long i; \
printg("wait_credit_refill():  credit refilled: "); \
for (i=0;i<total_num_nodes;i++) \
	printg("%ld ", local_credit[i]); \
printg("\n"); \
}) \
			break; \
		} \
	} \
	else { \
		unsigned long j; \
		credit = 999999; \
		for (j=0;j<total_num_nodes;j++) { \
			if (j != phys_whoami) { \
				if (local_credit1[j]) { \
					local_credit[j] = local_credit1[j]; \
					local_credit1[j] = 0; \
				} \
				if (local_credit[j] < credit) \
					credit = local_credit[j]; \
			} \
		} \
		if (credit) { \
			curr_credit = credit; \
DEBUG_F({ \
unsigned long i; \
printg("wait_credit_refill():  credit refilled: "); \
for (i=0;i<total_num_nodes;i++) \
	printg("%ld ", local_credit[i]); \
printg("\n"); \
}) \
			break; \
		} \
	} \
	i = 0; \
DEBUG_E({ \
unsigned long i; \
printg("wait_credit_refill():  credit NOT refilled yet: "); \
for (i=0;i<total_num_nodes;i++) \
	printg("%ld ", local_credit[i]); \
printg("\n"); \
}) \
	break; \
} \
}

#endif


#if 0

#define wait_room_in_shared_ring(i,par_pid,procnum) \
{ \
int res; \
time_586 a,b; \
i = 1; \
gamma_time(a); \
if (CREDIT_SPIN_TRYLOCK) \
for (;;) { \
	ring_room = SHARED_RING_SIZE-(shared_ring.tail-shared_ring.head); \
	if (ring_room > 4) { \
		CREDIT_SPIN_UNLOCK; \
		break; \
	} \
	/* avoid deadlock: poll() might invoke send() in turn,  */ \
	/* so we must deassert "gamma_tx_lock" (SMP only).	*/ \
	/* this however opens a window to reentrance in send(). */ \
	/* never mind, provided this macro does not reenter.    */ \
	/* that's the reason for the "gamma_cr_lock" (SMP only) */ \
	TX_SPIN_UNLOCK; \
	res = sys_gamma_poll(FROM_SEND_ROUTINE); \
	TX_SPIN_LOCK; \
	if (unlikely((res == -GAMMA_RX_SUSP) || (res == -GAMMA_ERR_MUSTYIELD))) { \
		i = 0; \
DEBUG_F(printg("wait_room_in_shared_ring():  leave wait_credit_refill because end-of-msg catched or msg for another process catched\n")) \
		CREDIT_SPIN_UNLOCK; \
		break; \
	} \
	gamma_time(b); \
	if ( \
		gamma_time_diff_ticks(b,a) \
		> (unsigned long long)MAX_CREDIT_WAIT*CLOCK \
	) { \
		i = 0; \
		CREDIT_SPIN_UNLOCK; \
		break; \
	} \
} \
else i = 0; \
}

#else

#define wait_room_in_shared_ring(i,par_pid,procnum) \
{ \
time_586 a,b; \
i = 1; \
gamma_time(a); \
for (;;) { \
	ring_room = SHARED_RING_SIZE-(shared_ring.tail-shared_ring.head); \
	if (ring_room > 4) { \
		break; \
	} \
	gamma_time(b); \
	if ( \
		gamma_time_diff_ticks(b,a) \
		> (unsigned long long)MAX_CREDIT_WAIT*CLOCK \
	) { \
		i = 0; \
		break; \
	} \
} \
}

#endif

#define busywait(n) \
{ \
	time_586 a,b; \
	gamma_time(a); \
	for (;;) { \
		gamma_time(b); \
		if ( \
gamma_time_diff_ticks(b,a) > (unsigned long long)(n)*CLOCK \
		) \
			break; \
	} \
}

#define tx_ring_is_full() (tx_ring_occupancy >= GAMMA_TX_RING_SIZE-5)

#define tx_ring_has_room() (tx_ring_occupancy < GAMMA_TX_RING_SIZE-20)

#define IRQ_enable_for_5usec(t_elaps,t_start) \
if ( \
gamma_time_diff_ticks((t_elaps),(t_start)) > (unsigned long long)9000*CLOCK \
) { \
/* \
 *  Open a 5 usec window for IRQs every 9 msec, necessary for \
 *  the correct management of the OS clock (gettimeoday()). \
 */ \
 local_save_flags(flags); \
 local_irq_enable(); \
 busywait(5); \
 local_irq_restore(flags); \
 gamma_time(t_start); \
}
