#define GAMMA_USERLEVEL
#ifdef GAMMA_INSTALLING
#include "gamma_userlev.h"
#else
#include <linux/gamma/gamma_userlev.h>
#endif
#undef GAMMA_USERLEVEL


#define MAX_NAME_LEN	200 /*  Max absolute name length of a GAMMA app. */


/*  GAMMA ports reserved for collective routines.  */

#define SYNC_PORT	MAX_PORT_NUMBER	/*  Port & sem. for barrier.             */
#define BCAST_PORT	(SYNC_PORT-1)	/*  Port & sem. for MPI_Bcast-like bcast */


/*  GAMMA system calls handles.  */
/*  Must be consistent with ../gamma_entry.S  */

#define GAMMA_INIT_PART1	0
#define GAMMA_CONNECTPORT	1
#define GAMMA_SEND		2
#define GAMMA_EXIT		3
#define GAMMA_VM_CREATE		4
#define GAMMA_VM_DESTROY	5
#define GAMMA_REGISTER		6
#define GAMMA_GETCONFIG		7
#define GAMMA_POLL		8
#define GAMMA_TIMERS		9
#define GAMMA_MLOCK		10
#define GAMMA_MUNLOCK		11
#define GAMMA_MUNLOCKALL	12
#define GAMMA_INIT_PART2	13
#define GAMMA_SEND_FLOWCTL	14
#define GAMMA_SEND_2P_FLOWCTL	15
#define GAMMA_SEND_2P		16
#define GAMMA_PRINTK		17
#define GAMMA_VM_ABORT		18
#define GAMMA_INIT_PART3	19
#define GAMMA_SUSPEND_IRQ	20
#define GAMMA_RESUME_IRQ	21
#define GAMMA_ISEND		22
#define GAMMA_ISEND_FLOWCTL	23
#define GAMMA_ISEND_2P		24
#define GAMMA_ISEND_2P_FLOWCTL	25
#define GAMMA_WSEND		26
#define GAMMA_TSEND		27
#define GAMMA_VM_CREATE_PID	28
#define GAMMA_VM_WISH		29
#define GAMMA_VM_GETINFO	30
#define GAMMA_REGISTER_NOSPAWN	31
#define GAMMA_SETCREDIT		32
#define GAMMA_RESETCREDIT	33
#define GAMMA_GETCREDIT		34
#define GAMMA_READ_RXTMPBUFF	35
#define GAMMA_SETMAXPOLLS	36
#define GAMMA_RESETMAXPOLLS	37
#define GAMMA_GETMAXPOLLS	38
#define GAMMA_POLL_NOWAIT	39
#define GAMMA_POST_RECV		40
#define GAMMA_SWITCH_TO_BUFFER	41
#define GAMMA_RECVPTR		42


/*
 *  During execution of a receive/error handler,
 *  this is the port that launched the handler itself.
 */
extern int gamma_active_port;

/*  The length of the last arrived message.  */
extern int gamma_msglen;


/****************************************************************/
/*	C prototypes						*/
/****************************************************************/

int gamma_init(unsigned char, int, char **);

int gamma_init_argv(int *, char ***);

int gamma_init_nospawn(unsigned char, unsigned long);

int gamma_post_recv(unsigned short, void *, unsigned long);

int gamma_switch_to_buffer(unsigned short, void *, unsigned long);

int gamma_recvptr(unsigned short);

int gamma_set_passive_port (
			unsigned short port,
			unsigned short partner_node,
			unsigned char partner_par_pid,
			unsigned short partner_port,
			unsigned short semaphore,
			unsigned char queue_kind,
			void *buffer_in,
			unsigned long buffer_in_len);

int gamma_set_active_port (
			unsigned short port,
			unsigned short partner_node,
			unsigned char partner_par_pid,
			unsigned short partner_port,
			void (*handler_rec)(),
			unsigned short semaphore,
			unsigned char queue_kind,
			void *buffer_in,
			unsigned long buffer_in_len);

int gamma_set_active_port_2p (
			unsigned short port,
			unsigned short partner_node,
			unsigned char partner_par_pid,
			unsigned short partner_port,
			void (*handler_rec)(),
			unsigned short semaphore,
			unsigned char queue_kind,
			void *buffer_in,
			unsigned long buffer_in_len,
			unsigned char boundary_1p);

int gamma_send(unsigned short, void *, unsigned long);

int gamma_send_flowctl(unsigned short, void *, unsigned long);

int gamma_send_2p(
		unsigned short,
		void *,
		unsigned long,
		void *,
		unsigned long);

int gamma_send_2p_flowctl(
		unsigned short,
		void *,
		unsigned long,
		void *,
		unsigned long);

int gamma_isend(unsigned short, void *, unsigned long);

int gamma_isend_flowctl(unsigned short, void *, unsigned long);

int gamma_isend_2p(
		unsigned short,
		void *,
		unsigned long,
		void *,
		unsigned long);

int gamma_isend_2p_flowctl(
		unsigned short,
		void *,
		unsigned long,
		void *,
		unsigned long);

int gamma_wsend(int);

int gamma_tsend(int);

int gamma_atomic(void (*)());

int gamma_exit(void);

int gamma_exit_nospawn(void);

int gamma_abort(void);

int gamma_exit_nospawn(void);

int gamma_vm_create (unsigned char, void *, void *);

int gamma_vm_destroy (unsigned char);

int gamma_getconfig(unsigned long);

int gamma_mlock(void *, unsigned long);
int gamma_mlock_funct(void *, unsigned long);
int gamma_munlock(void *, unsigned long);
int gamma_munlockall(void);

int gamma_how_many_nodes(void);

int gamma_my_par_pid(void);

int gamma_my_node(void);

int gamma_signal(unsigned short);

int gamma_sigerr(unsigned short);

int gamma_wait(unsigned short, unsigned long);

int gamma_test(unsigned short);

int gamma_suspend_irq(void);

int gamma_resume_irq(void);

int gamma_timers(time_586, time_586);

int gamma_sync(void);

int gamma_bcast(void *, unsigned long, unsigned long);

void gamma_busywait_usec(unsigned long);

int gamma_sched_yield(void);

int gamma_printk(int);

int gamma_name2i(char *);

int gamma_setcredit(unsigned long);

int gamma_resetcredit(void);

int gamma_getcredit(void);

int gamma_setmaxpolls(unsigned long);

int gamma_resetmaxpolls(void);

int gamma_getmaxpolls(void);
