/***************************************************************************
	Read COPYRIGHT file in this directory
****************************************************************************/


#include "gamma.h"
#include "gamma_leading.h"
static char *gamma_authors = "{ciaccio,chiola}@disi.unige.it\n";


/************************************************************************
 *  GAMMA Data structures						*
 ************************************************************************/

task_t *gamma_run_handler;

task_t *gamma_run_handler_samenode;

task_t *gamma_must_yield;

task_t *gamma_must_yield_samenode;

unsigned long gamma_irq_disabled;

unsigned long total_num_nodes = 0;

unsigned long phys_whoami;

unsigned long local_credit[MAX_NUM_NODES];
unsigned long local_credit1[MAX_NUM_NODES];

atomic_t gamma_susp_irq_count;
atomic_t gamma_IP_tx_block;

struct net_device *gamma_device;

/*
 *  Table of (ptrs to) active parallel applications (one entry for each).
 *  Initialized by gamma_start(); entries allocated by sys_gamma_vm_create()
 *  and gamma_kernel_rec().
 */
par_entry *par_proc_table[MAX_PAR_PID+1];

/*
 *  Current initial sizes of flow ctl credits.
 *  Initialized by sys_gamma_getconfig() and gamma_start().
 */
unsigned long initial_credit[MAX_NUM_NODES];
unsigned long credit_ceil[MAX_NUM_NODES];
unsigned long credit_growth_cnt[MAX_NUM_NODES];
unsigned long credit_size;

/*
 *  Per-destination rings of (ptrs to) per-packet tx temp buffers.
 *  Bcast pkts use entry #phys_whoami ("self", which would otherwise be
 *  unused).
 *  Alloc and init by gamma_start(), init by gamma_getconfig().
 */
typedef struct {
	unsigned char *b;	/*  Ptr to temp buffer  */
	int l;			/*
				 *  pkt from blocking send routines:
				 *    pkt size (>=0);
				 *  pkt from non-blocking send routines:
				 *    1 plus the NIC Tx ring tail at the time
				 *    when the pkt was enqueued,
				 *    then change sign (<0 therefore).
				 */
	unsigned long hlen;	/*
				 *  non-blocking send only:
				 *  size of header (+ possible 1st chunk,
				 *  in case 2-way gather send was used).
				 */
	unsigned long plen;	/*
				 *  non-blocking send only:
				 *  size of of payload (2nd chunk only,
				 *  in case 2-way gather send was used).
				 */
} tx_tmpbuf_type;
typedef struct {
	tx_tmpbuf_type buf[GAMMA_TX_WIN_SIZE];
	volatile unsigned short tail;
#ifdef SEQCHECK
/*
 *  Pkt numbering.  Each pkt is stamped by a sequence number.
 *  The current pkt number for destination node #N is stored here;
 *  the expected pkt number from source node #N is stored at rx_seq[N].
 */
	volatile unsigned char tx_seq;
/*  Pad to doubleword.  */
	unsigned char dummy;
#endif
} tx_tmp_ring_type;
tx_tmp_ring_type tx_tmp_ring[MAX_NUM_NODES];
tx_tmp_ring_type kernel_tx_tmp_ring;

/*
 *  Ring of ptrs to packets incoming to non-running processes.
 *  Init and alloc by gamma_start().
 */
struct {
	void *buf[RX_TMP_RING_SIZE];
	unsigned long head;
	unsigned long tail;
} rx_tmp_ring;

#ifdef MORE_INSTANCES_ON_SAME_NODE
/*
 *  Ring of (ptrs to) temp buffers for intra-node message exchange.
 *  Init and alloc by gamma_start().
 */
#define IS_1ST_CNK 0x1
#define IS_LAST_CNK 0x2
typedef struct {
	unsigned long par_pid;
	unsigned long procnum;
	unsigned long port;
	unsigned long cnk_len;
	unsigned long bitmap;
	unsigned long num_locdest;
	time_586 watch_time;
	task_t *sender;
	void *buf;
} shared_ring_entry;
struct {
	volatile unsigned long head;
	volatile unsigned long dirty;
	volatile unsigned long tail;
	shared_ring_entry desc[SHARED_RING_SIZE];
} shared_ring;
#endif

#ifdef SEQCHECK
/*
 *  Table of pkt numbering.  Each pkt is stamped by a sequence number.
 *  The current pkt number for destination node #N is stored at tx_seq[N];
 *  the expected pkt number from source node #N is stored at rx_seq[N].
 *  Initialized by gamma_getconfig().
 *  Bcast pkts have distinct numbering.
 */
unsigned char rx_seq[MAX_NUM_NODES];
unsigned char rx_seq_bcast[MAX_NUM_NODES];

/*  Per-node table of flags: 1 iff pkts from given node must be rejected.  */
unsigned char rx_discard[MAX_NUM_NODES];

/*
 *  Per-destination flags:
 *  1 iff given node has experienced pkt losses (congestion?).
 *  Initialized by sys_gamma_getconfig()
 */
unsigned char congested[MAX_NUM_NODES];
#endif

/*
 *  Counter of invocations to sys_gamma_poll(), plus a threshold.
 *  The counter is incremented each time a polling phase ends without getting
 *  any new pkt.
 *  When its value get higher than the threshold, the waiting node
 *  may suppose there is a dealock due to one or more missing pkts;
 *  it then asks everybody for their current frame number (by sending TEST
 *  pkt).
 *  This helps detecting missing pkts and thus keeps the GAMMA protocol alive.
 */
#define REPEAT_POLL 200                  /*  200 usec poll phase duration */
#define MAX_POLLS (1000000/REPEAT_POLL)  /*  1 sec timeout                */
#define CEILING_MAX_POLLS 1*MAX_POLLS    /*  Can be increased             */
int how_many_polls = 0;
int max_polls = MAX_POLLS;
int base_max_polls = MAX_POLLS;

/*  spin/lock semaphores for mutual excl on send/recv/scheduler/wait_credit  */
spinlock_t gamma_rx_lock = SPIN_LOCK_UNLOCKED;
spinlock_t gamma_tx_lock = SPIN_LOCK_UNLOCKED;
#if 0
spinlock_t gamma_cr_lock = SPIN_LOCK_UNLOCKED;
#endif
spinlock_t gamma_rxsamenode_lock = SPIN_LOCK_UNLOCKED;

/*
 *  Order of magnitude of sizeof(par_entry), needed for get_free_pages().
 *  Initialized by gamma_start();
 */
unsigned long par_entry_magnitude;

eth_addr eth_table[MAX_NUM_NODES];

#ifdef MEASURE_OUT_TH
unsigned long out_counter = 0;
unsigned long out_total_amount = 0;
#endif

#ifdef MEASURE_OUT_TH_MEM
unsigned long out_counter_mem = 0;
unsigned long out_total_amount_mem_delay = 0;
unsigned long out_total_amount_mem_bytes = 0;
#endif

#ifdef MEASURE_IN_TH
unsigned long in_counter = 0;
unsigned long in_total_amount = 0;
#endif

#ifdef MEASURE_IN_TH_MEM
unsigned long in_counter_mem = 0;
unsigned long in_total_amount_mem_delay = 0;
unsigned long in_total_amount_mem_bytes = 0;
#endif

static char *gamma_version =
"\nGAMMA  Genoa Active Message MAchine (Linux 2.6)  16 Sep 2004\n";

/************************************************************************
 *	Some function protos						*
 ************************************************************************/


int sys_gamma_poll(unsigned long);
static inline int is_low_watermark(unsigned long,unsigned long,unsigned long);
static inline int min_low_watermark(unsigned long);
static inline int max_low_watermark(unsigned long);


/************************************************************************
 *	Init routine							*
 ************************************************************************/


void gamma_start
(struct net_device *dev)
{

	int i,j;


	rx_tmp_ring.head = 0;
	rx_tmp_ring.tail = 0;

/*
 *  Allocate per-packet tx tmp bufs (one for each Tx ring entry).
 *  We don't yet know how many nodes are actually present in the cluster;
 *  in doubt, we must allocate as many Tx rings as MAX_NUM_NODES.
 *  Later, we de-allocate those one unneeded (gamma_getconfig()).
 */
	for (i=0;i<MAX_NUM_NODES;i++) {
		for (j=0;j<GAMMA_TX_WIN_SIZE;j++) {
			tx_tmp_ring[i].buf[j].b = (unsigned char *)
				kmalloc(FRAME_SIZ, GFP_KERNEL);
			if (tx_tmp_ring[i].buf[j].b == NULL) {
				printg("gamma_start():  kmalloc() failed\n");
				return;
			}
			tx_tmp_ring[i].buf[j].l = 0;
		}
		tx_tmp_ring[i].tail = 0;
#ifdef SEQCHECK
		tx_tmp_ring[i].tx_seq = 0;
#endif
	}
	for (j=0;j<GAMMA_TX_WIN_SIZE;j++) {
		kernel_tx_tmp_ring.buf[j].b = (unsigned char *)
			kmalloc(FRAME_SIZ, GFP_KERNEL);
		if (kernel_tx_tmp_ring.buf[j].b == NULL) {
			printg("gamma_start():  kmalloc() failed\n");
			return;
		}
		kernel_tx_tmp_ring.buf[j].l = 0;
	}
	kernel_tx_tmp_ring.tail = 0;
#ifdef SEQCHECK
	kernel_tx_tmp_ring.tx_seq = 0;
#endif

	gamma_run_handler = NULL;

	gamma_run_handler_samenode = NULL;

	gamma_must_yield = NULL;

	gamma_must_yield_samenode = NULL;

	gamma_irq_disabled = 0;

	atomic_set(&gamma_susp_irq_count,0);
	atomic_set(&gamma_IP_tx_block,0);

	gamma_device = dev;

/*  The table of parallel processes is empty initially  */
	for (i=1;i<=MAX_PAR_PID;i++)
		par_proc_table[i] = (par_entry *)NULL;

#ifdef MORE_INSTANCES_ON_SAME_NODE
/*
 *  Init and allocate the ring of temp buffers
 *  for intra-node communication.
 */
	shared_ring.head = shared_ring.dirty = shared_ring.tail = 0;
	for (i=0;i<SHARED_RING_SIZE;i++) {
		shared_ring.desc[i].par_pid = 0;
		shared_ring.desc[i].procnum = 0;
		shared_ring.desc[i].port = 0;
		shared_ring.desc[i].cnk_len = 0;
		shared_ring.desc[i].bitmap = 0;
		shared_ring.desc[i].num_locdest = 0;
		shared_ring.desc[i].watch_time.hi = 0;
		shared_ring.desc[i].watch_time.lo = 0;
		shared_ring.desc[i].buf = kmalloc(PAGE_SIZE,GFP_KERNEL);
		if (shared_ring.desc[i].buf == NULL) {
			printg("gamma_start():  kmalloc() failed\n");
			return;
		}
		shared_ring.desc[i].sender = NULL;
	}
#endif

/*  Initial credit size is max (as defined by constant MAX_CREDIT_SIZE).  */
	credit_size = MAX_CREDIT_SIZE;

/*  Compute the magnitude of sizeof(par_entry), needed to allocate GAMMA vm  */
	for (i=1,j=0;i*PAGE_SIZE<sizeof(par_entry);i<<=1,j++);
	par_entry_magnitude = j;

/*  Set the entry points for the GAMMA primitives (trap 0x81)  */
	set_system_gate(0x81,&gamma_system_call);

	printg("%s",gamma_version);
#ifdef DRIVER_DEBUG
	printg("       Debug level: High\n");
#else
#ifdef DRIVER_DEBUG_LIGHT
	printg("       Debug level: Low\n");
#else
	printg("       Debug level: Null\n");
#endif
#endif
#ifdef GAMMA_FULL_DUPLEX
	printg("       Full-duplex mode");
#else
	printg("       Half-duplex mode");
#endif
#ifdef COLL_SAFE_TX
	printg(", collision-safe tx enabled");
#endif
#ifdef SAFE_TX
	printg(", safe tx enabled");
#endif
	printg("\n");
#ifdef USE_JUMBO_FRAMES
	printg("       jumbo frame support enabled:  MTU %d\n",FRAME_SIZ);
#endif
	printg("       Size of Tx ring %d     Size of Rx ring %d\n",
		GAMMA_TX_RING_SIZE, GAMMA_RX_RING_SIZE
	);
	printg("       GAMMA protocol: ");
	printg("%s",gamma_authors);
	printg("       BIND2 layer   :\n");
	printg("       ");
	printg("%s",bind2_authors);

	return;

}


/************************************************************************
 *	Service send/receive routines					*
 ************************************************************************/


/*  CAUTION:  do not use flow ctl here!  */
void gamma_kernel_send
(
	unsigned long phys_dest,
	unsigned char par_pid,
	unsigned char kind,
	void *data,
	unsigned short len,
	int use_lock
)
{

	tx_tmp_ring_type *win;
	unsigned long entry;
	l2s *h;
	tx_common_decl_and_init;
#ifdef  COLL_SAFE_TX
	int trials = MAX_TRIALS;
#endif


#ifdef MORE_INSTANCES_ON_SAME_NODE
	if (use_lock) spin_lock(&gamma_tx_lock);
#endif
/*
 *  Pkts of the flow-ctl protocol must be kept into the
 *  Tx tmp ring (for possible retransmission, and to keep the numbering
 *  consistent); other kernel pkts need not.
 */
	if ((kind == FILL_CREDIT) || (kind == CRED_REQ))
		if (phys_dest != BROADCAST)
			win = &tx_tmp_ring[phys_dest];
		else
			win = &tx_tmp_ring[phys_whoami];
	else
		win = &kernel_tx_tmp_ring;

	entry = win->tail;
	incr_modulo(win->tail,GAMMA_TX_WIN_SIZE);
	h = (l2s *)(win->buf[entry].b);

	if (!tx_ring_is_full()) {

room_in_tx_ring:

	make_send_header(phys_dest,par_pid,kind,0,ETH_GAMMA_KERNEL,1,len,0,h);
#ifdef SEQCHECK
	if ((kind == FILL_CREDIT) || (kind == CRED_REQ))
/*
 *  Pkts of the flow-ctl and frame no. protocols must be kept numbered,
 *  as they are kept into the Tx tmp ring for possible retransmission.
 */
		set_send_header_frameno(h,win->tx_seq++);

DEBUG_F(
if (kind == FILL_CREDIT)
	printg("send FILL_CREDIT to node %ld, frame_no %d\n",
		phys_dest,read_send_header_frameno(h));
if (kind == CRED_REQ)
	printg("send CRED_REQ to node %ld, frame_no %d\n",
		phys_dest,read_send_header_frameno(h));
if (kind == ACK)
	printg("send ACK to node %ld, frame_no %d\n",
		phys_dest,*(unsigned char *)data);
if (kind == ACKB)
	printg("send ACKB to node %ld, frame_no %d\n",
		phys_dest,*(unsigned char *)data);
)

#endif
	memcpy(((unsigned char *)h)+HEADER_SIZE,(void *)data,len);

DEBUG_E(printg(
"gamma_kernel_send(): h0 %8.8lx, h1 %8.8lx, h2 %8.8lx, " \
"h3 %8.8lx, h4 %8.8lx\n",
h[0].ll, h[1].ll, h[2].ll, h[3].ll, h[4].ll
));

#ifdef COLL_SAFE_TX
back:
#endif

DEBUG(printg(
"gamma_kernel_send(): sending to node %ld, kind %d\n",
phys_dest,kind
));

	tx_enq_pkt1(h,HEADER_SIZE+len);
	tx_start();
	tx_forw_tail();
#ifdef COLL_SAFE_TX
	tx_flush_ring_1(tx_ring_tail,MAX_FLUSH_WAIT*10,i);
	if ((!i) && (--trials >= 0)) {
DEBUG_L(printg("retrying\n"));
		goto back;
	}
#endif
#ifdef SAFE_TX
	if (tx_ring_occupancy >= STEPS_THEN_STOP)
		tx_flush_ring(tx_ring_tail,MAX_FLUSH_WAIT,i);
#endif
	win->buf[entry].l = HEADER_SIZE+len;
	tx_close();
#ifdef MORE_INSTANCES_ON_SAME_NODE
	if (use_lock) spin_unlock(&gamma_tx_lock);
#endif
	return;

	}

	tx_wait_ring_room(MAX_FLUSH_WAIT,i);
	if (i)
		goto room_in_tx_ring;

DEBUG_L(printg("gamma_kernel_send(): Transmit failure before send.\n"));

	tx_reset_and_restart();
#ifdef MORE_INSTANCES_ON_SAME_NODE
	if (use_lock) spin_unlock(&gamma_tx_lock);
#endif
	return;

}


#ifdef SEQCHECK
/*
 *  Scan a given Tx tmp ring to retrieve pkts which are to be retransmitted.
 */
static int gamma_scan_tx_tmp_ring
(unsigned long phys_dest, unsigned long frameno, int is_bcast)
{

	tx_tmp_ring_type *win;
	int i;
	l2s *h;


#ifdef MORE_INSTANCES_ON_SAME_NODE
	spin_lock(&gamma_tx_lock);
#endif
	if (!is_bcast)
		win = &tx_tmp_ring[phys_dest];
	else
		win = &tx_tmp_ring[phys_whoami];

	for (
		i = (win->tail+GAMMA_TX_WIN_SIZE-1) % GAMMA_TX_WIN_SIZE;
		i != (win->tail % GAMMA_TX_WIN_SIZE);
		i = (i+GAMMA_TX_WIN_SIZE-1) % GAMMA_TX_WIN_SIZE
	) {
		h = (l2s *)win->buf[i].b;

DEBUG(printg(
"gamma_scan_tx_tmp_ring(): h0 %8.8lx, h1 %8.8lx, h2 %8.8lx, h3 %8.8lx, h4 %8.8lx\n",
h[0].ll, h[1].ll, h[2].ll, h[3].ll, h[4].ll
));

		if (read_send_header_frameno(h) != frameno)
			continue;
/*  Pkt found.  */
#ifdef MORE_INSTANCES_ON_SAME_NODE
		spin_unlock(&gamma_tx_lock);
#endif
		return i;
	}
/*  Pkt not found.  */
#ifdef MORE_INSTANCES_ON_SAME_NODE
	spin_unlock(&gamma_tx_lock);
#endif
	return -1;

}


/*  Re-enqueue pkts to be retransmitted.  */
/*  CAUTION:  do not use flow ctl here!  */
static int gamma_retx
(unsigned long phys_dest, unsigned long pos, int is_bcast, int use_lock)
{

	tx_tmp_ring_type *win;
	int ii;
	int tx_handle;
	l2s *h;
	tx_common_decl_and_init;
#ifdef  COLL_SAFE_TX
	int trials = MAX_TRIALS;
#endif


#ifdef MORE_INSTANCES_ON_SAME_NODE
	if (use_lock) spin_lock(&gamma_tx_lock);
#endif
	if (!is_bcast)
		win = &tx_tmp_ring[phys_dest];
	else
		win = &tx_tmp_ring[phys_whoami];

/*  Loop over the Tx tmp ring from "pos" to end.  */
	for (ii = pos;;) {
		h = (l2s *)win->buf[ii].b;
		tx_handle = -(win->buf[ii].l) - 1;
		if (tx_handle >= 0) {
/* peppe */
#if 0
/*
 *  Pkt is from a non-blocking send.  We must check it is still pending in
 *  the Tx ring so that we can still retransmit it
 *  (according to the non-blocking semantics of isend() routines).
 */
			if (
			(!(tx_ring_occupancy))
			||
			(tx_handle > tx_ring_tail)
			||
			((tx_ring_head < tx_ring_tail)
			&& (tx_handle <= tx_ring_head))
			) {
#endif

DEBUG_L(printg(
"Could not retransmit a flushed pkt originated from a non-blocking send()\n"
));

				goto skip;
/* peppe */
#if 0
			}
#endif
		}
/*
 *  This pkt is OK for retransmission.  But we have to change its type
 *  to "retransmitted".
 *  Rationale:  pkts in the Tx tmp ring can be bcast; if we retransmit them,
 *  we must prevent non-involved nodes to catch them twice!
 */
		if (rec_header_frmtype(h) == ETH_GAMMA)
			set_send_header_frmtype(h,ETH_GAMMA_RETX);
		else
			if (rec_header_frmtype(h) == ETH_GAMMA_KERNEL)
				set_send_header_frmtype(
					h,ETH_GAMMA_KERNEL_RETX
				);
/*  Enqueue pkt for retransmit.  */
/* peppe */
#if 0
		if (tx_handle >= 0) {
printg("ciao retx isend pkt, addr %8.8lx\n",
tx_entry_2_payloadptr((tx_handle+GAMMA_TX_RING_SIZE-1) % GAMMA_TX_RING_SIZE));
/*  Pkt from non-blocking send:  copy payload from user to kernel space.  */
			memcpy(
((unsigned char *)h) + win->buf[i].hlen,
tx_entry_2_payloadptr((tx_handle+GAMMA_TX_RING_SIZE-1) % GAMMA_TX_RING_SIZE),
win->buf[i].plen
			);
printg("ciao fatto\n");
		}
#endif
		if (!tx_ring_is_full()) {

room_in_tx_ring:

DEBUG(printg(
"gamma_retx(): h0 %8.8lx, h1 %8.8lx, h2 %8.8lx, " \
"h3 %8.8lx, h4 %8.8lx\n",
h[0].ll, h[1].ll, h[2].ll, h[3].ll, h[4].ll
));

#ifdef COLL_SAFE_TX
back:
#endif

DEBUG(printg(
"gamma_retx(): sending to node %ld, frame_no %d\n",
phys_dest,read_send_header_frameno(h)
));

		tx_enq_pkt1(h,win->buf[ii].l);
		tx_start();
		tx_forw_tail();
#ifdef COLL_SAFE_TX
		tx_flush_ring_1(tx_ring_tail,MAX_FLUSH_WAIT*10,i);
		if ((!i) && (--trials >= 0)) {
DEBUG_L(printg("retrying\n"));
			goto back;
		}
#endif
#ifdef SAFE_TX
		if (tx_ring_occupancy >= STEPS_THEN_STOP)
			tx_flush_ring(tx_ring_tail,MAX_FLUSH_WAIT,i);
#endif

skip:

		incr_modulo(ii,GAMMA_TX_WIN_SIZE);
		if (ii == win->tail)
			break;
		continue;

		}

		tx_wait_ring_room(MAX_FLUSH_WAIT,i);
		if (i)
			goto room_in_tx_ring;

DEBUG_L(printg("gamma_retx(): Transmit failure before send.\n"));

		tx_reset_and_restart();
#ifdef MORE_INSTANCES_ON_SAME_NODE
		if (use_lock) spin_unlock(&gamma_tx_lock);
#endif
		return -1;
	}
/*
 *  This is because a gamma_send_flowctl() might be in progress, and we don't
 *  want to propagate the ETH_GAMMA_RETX type to all the Tx ring entries.
 */
	tx_flush_ring(tx_ring_tail,MAX_FLUSH_WAIT,i);
	if (rec_header_frmtype(h) == ETH_GAMMA_RETX)
		set_send_header_frmtype(h,ETH_GAMMA);
#ifdef MORE_INSTANCES_ON_SAME_NODE
	if (use_lock) spin_unlock(&gamma_tx_lock);
#endif
	return 0;

}
#endif


static void gamma_kernel_rec
(
	void *addr,
	unsigned char par_pid,
	unsigned short len,
	unsigned char kind,
	unsigned long phys_sender
)
{

	par_entry *p;
	port_entry *pp;

	if (kind == FILL_CREDIT) {

DEBUG_F(printg(
"got FILL_CREDIT from node %ld, frame_no %ld\n",
phys_sender,*(unsigned long *)addr
))

#ifdef SEQCHECK
		if (!congested[phys_sender]) {
			if (
				initial_credit[phys_sender]
				>= credit_ceil[phys_sender]
			) {
				/*  Credit size raises the allowed ceiling.  */
				if (credit_ceil[phys_sender] >= credit_size) {
					/*  Credit ceiling is max.  */
					local_credit1[phys_sender] =
						initial_credit[phys_sender];
					return;
				}
				/*
				 *  Credit ceiling can grow.
				 *  We gradually (and slowly) push up the
				 *  ceiling till the max.
				 */
				if (++credit_growth_cnt[phys_sender] >= 1000) {
					credit_growth_cnt[phys_sender] = 0;
					credit_ceil[phys_sender]++;
				}
				local_credit1[phys_sender] =
					initial_credit[phys_sender];
				return;
			}
			/*
			 *  Push up the credit size till current ceiling.
			 *  Do not push up the ceiling yet.
			 */
			initial_credit[phys_sender]++;
			local_credit1[phys_sender] =
				initial_credit[phys_sender];
			return;
		}
/*  A possible congestion on node "phys_sender".  Slow down...  */
		congested[phys_sender] = 0;
/*  Shrink down the credit ceiling (just a bit)  */
		credit_ceil[phys_sender]
			= gmax(2,initial_credit[phys_sender]-1);
		credit_growth_cnt[phys_sender] = 0;
/*  Shrink down the credit size (quite a lot)  */
		if (initial_credit[phys_sender] > 2)
			initial_credit[phys_sender]
				= gmax(2,initial_credit[phys_sender]>>1);
/*  This is a less aggressive shrink rule...I could not do any meaninful
comparison with the previous one yet :(
				= gmax(2,initial_credit[phys_sender]-1);
*/
#endif
		local_credit1[phys_sender] = initial_credit[phys_sender];
		return;
	}

	if (kind == CRED_REQ) {
	 	gamma_kernel_send(phys_sender,DONT_MIND,FILL_CREDIT,NULL,0,1);
		return;
	}

#ifdef SEQCHECK
	if (kind == ACK) {
/*  Check the frame numbering of "phys_sender" (unicast).  */
		unsigned long frame_no;

		frame_no = *(unsigned long *)addr;
		if (frame_no == rx_seq[phys_sender]) {

DEBUG(printg("got ACK from node %ld, frame_no %ld\n", phys_sender,frame_no))

/*  No missing pkt was detected. Increase the time interval to next check.  */
			if (max_polls < CEILING_MAX_POLLS)
				max_polls +=
				base_max_polls/(2*(total_num_nodes-1));
			return;
		}
/*  Missing pkts from "phys_sender". See rx_gamma(), label "err_frame_no"  */

DEBUG_F(printg(
"gamma_kernel_rec(): Received ACK #%ld, expctd #%d, phys sender %ld\n",
frame_no,rx_seq[phys_sender],phys_sender
))

		max_polls = base_max_polls;
		gamma_kernel_send(BROADCAST,DONT_MIND,CONGESTION,NULL,0,1);
		frame_no = rx_seq[phys_sender];
		gamma_kernel_send(
			phys_sender,DONT_MIND,NACK,&frame_no,
			sizeof(unsigned long),1
		);
		rx_discard[phys_sender] = 1;
		return;
	}

	if (kind == ACKB) {
/*  Check the frame numbering of "phys_sender" (bcast).  */
		unsigned long frame_no;

		frame_no = *(unsigned long *)addr;
		if (frame_no == rx_seq_bcast[phys_sender]) {

DEBUG(printg("got ACKB from node %ld, frame_no %ld\n", phys_sender,frame_no))

/*  No missing pkt was detected. Increase the time interval to next check.  */
			if (max_polls < CEILING_MAX_POLLS)
				max_polls +=
				base_max_polls/(2*(total_num_nodes-1));
			return;
		}
/*  Missing pkts from "phys_sender". See rx_gamma(), label "err_frame_no"  */

DEBUG_F(printg(
"gamma_kernel_rec(): Received ACKB #%ld, expctd #%d, phys sender %ld\n",
frame_no,rx_seq[phys_sender],phys_sender
))

		frame_no = rx_seq_bcast[phys_sender];
		gamma_kernel_send(
			phys_sender,DONT_MIND,NACKB,&frame_no,
			sizeof(unsigned long),1
		);
		rx_discard[phys_sender] = 1;
		return;
	}

	if (kind == CONGESTION) {
/*
 *  A possible congestion on node "phys_sender".  Slow down at next credit
 *  renewal.
 */
		congested[phys_sender] = 1;
		return;
	}

	if (kind == NACK) {
/*  NACK for missing unicast pkts.  */
		int frame_no,pos;

                frame_no = ((unsigned long *)addr)[0];

DEBUG_L(printg(
"got NACK from node %ld:  missing frame %d\n",phys_sender,frame_no
));

		if (
			(pos = gamma_scan_tx_tmp_ring(phys_sender,frame_no,0))
			>= 0
		) {
/*  Go-back-N and retransmit.  */
			gamma_kernel_send(
				phys_sender,DONT_MIND,RETX_START,NULL,0,1
			);
			gamma_retx(phys_sender,pos,0,1);
			gamma_kernel_send(
				phys_sender,DONT_MIND,RETX_END,NULL,0,1
			);
			return;
		}

DEBUG_L(printg(
"Missing pkt %d from node %ld:  not found. Cannot retransmit!\n",
frame_no,phys_sender
));

		gamma_kernel_send(phys_sender,DONT_MIND,ERR_NO_RETX,addr,len,1);
		return;
	}

	if (kind == NACKB) {
/*  NACK for missing bcast pkts.  */
		int frame_no,pos;

                frame_no = ((unsigned long *)addr)[0];

DEBUG_L(printg(
"got NACKB from node %ld:  missing bcast frame %d\n",phys_sender,frame_no
));

		if (
			(pos = gamma_scan_tx_tmp_ring(phys_sender,frame_no,1))
			>= 0
		) {
/*  Go-back-N and retransmit.  */
			gamma_kernel_send(
				phys_sender,DONT_MIND,RETX_START,NULL,0,1
			);
			gamma_retx(phys_sender,pos,1,1);
			gamma_kernel_send(
				phys_sender,DONT_MIND,RETX_END,NULL,0,1
			);
			return;
		}

DEBUG_L(printg(
"Missing bcast pkt %d from node %ld:  not found. Cannot retransmit!\n",
frame_no,phys_sender
));

		gamma_kernel_send(phys_sender,DONT_MIND,ERR_NO_RETX,addr,len,1);
		return;
	}

	if (kind == TEST) {
/*
 *  This node is asked to spread its table of frame numbering (unicast
 *  as well as bcast);
 *  this can help recovering from a deadlock in case of pkt losses.
 */
		unsigned long frame_no, i;

		if (total_num_nodes == 0) {
/*
 *  Maybe this machine has rebooted and has not yet called
 *  sys_gamma_getconfig(), while other ones are waiting for pkts.
 *  Discard this request.
 */
			return;
		}

		for (
			i = (phys_whoami + 1) % total_num_nodes;
			i != phys_whoami;
			i = (i + 1) % total_num_nodes
		) {
			frame_no =
			(unsigned long)(tx_tmp_ring[i].tx_seq);
			gamma_kernel_send(
				i,DONT_MIND,ACK,
				&frame_no,sizeof(unsigned long),1
			);
			frame_no =
			(unsigned long)(tx_tmp_ring[phys_whoami].tx_seq);
			gamma_kernel_send(
				i,DONT_MIND,ACKB,
				&frame_no,sizeof(unsigned long),1
			);
		}
		return;
	}

	if (kind == RETX_START) {
/*
 *  A sequence of retransmitted pkts begins.
 *  Since this may occur only after detecting a missing pkt in the
 *  sequence from node "phys_sender", we should have set
 *  rx_discard[phys_sender] = 1.
 */
		if (rx_discard[phys_sender] == 1) {

DEBUG_F(printg("Start of retx sequence, sender %ld\n",phys_sender));

			rx_discard[phys_sender] = 2;
			return;
		}

DEBUG_F(printg(
"gamma_kernel_rec():  Start of double retransmission from node %ld\n",
phys_sender
));

		return;
	}

	if (kind == RETX_END) {
/*
 *  A sequence of retransmitted pkts ends.
 *  Since this may occur only after a sequence of retransmitted pkts
 *  from node "phys_sender", we should have set
 *  rx_discard[phys_sender] = 2.
 */
		if (rx_discard[phys_sender] == 2) {

DEBUG_F(printg("End of retx sequence, sender %ld\n",phys_sender));

			rx_discard[phys_sender] = 0;
			return;
		}
		if (rx_discard[phys_sender] == 1) {
/*
 *  Asked for retransmission but did not catch the RETX_START pkt;
 *  or, detected missing pkt during a retransmission sequence and
 *  therefore refusing remaining part of sequence, but did not yet get
 *  RETX_START.
 */

DEBUG_F(printg(
"gamma_kernel_rec():  unexpectedly got RETX_END (missing RETX_START?)\n"
));

			return;
		}

DEBUG_F(printg(
"gamma_kernel_rec():  End of double retransmission from node %ld\n",
phys_sender
));

		return;
	}

/* peppe : questa e' sospesa xche' e' complicato gestirla nel caso in
cui il processo ricevente non e' running */
#if 0
	if (kind == ERR_NO_RETX) {
/*
 *  Sender could not retransmit required missing pkt.
 *  Invoke error handler.
 */
		unsigned long par_pid, port, procnum;

		par_pid = ((unsigned long *)addr)[1];
		port = ((unsigned long *)addr)[2];
		procnum = ((unsigned long *)addr)[3];
/*  Check whether virtual GAMMA #par_pid exists and is assigned.  */
		if (
(par_proc_table[par_pid] != (par_entry *)NULL)
&&
(par_proc_table[par_pid]->inst_table[procnum].pid != -1)
		) {

DEBUG_F(printg(
"Got ERR_NO_RETX. Executing ERR handler dispatcher for par_pid %ld port %ld\n",
par_pid,port
));

			invoke_err_handler(
				handler_entry,par_pid,port,procnum,
				GAMMA_ERR_FRAMELOST	
			);

DEBUG(printg("ERR handler executed\n"));

		}
		return;
	}
#endif
#endif

	p = par_proc_table[par_pid];

	if (kind == VM_WISH) {
/*  A remote proc wishes to join virtual GAMMA #par_pid  */
		int i;

		if (p != (par_entry *)NULL) {
/*  Virtual GAMMA already created.  Express the wish, then return.  */
			p->vm.how_many_nodes++;
			p->vm.wish_flag[phys_sender]++;
			return;
		}
/*  Allocate a descriptor for a new virtual GAMMA  */
		par_proc_table[par_pid] =
			(par_entry *)__get_free_pages(GFP_ATOMIC,par_entry_magnitude);
		p = par_proc_table[par_pid];
		if (p == (par_entry *)NULL) {
printg("gamma_kernel_rec(): kmalloc() failed when creating virtual GAMMA #%d\n", par_pid);
			return;
		}
		for (i=0;i<MAX_NUM_NODES;i++)
			p->vm.wish_flag[i] = 0;
		for (i=0;i<MAX_NUM_LOCAL_INST;i++) {
			p->inst_table[i].pid = -1;
			p->inst_table[i].task = NULL;
			p->inst_table[i].whoami = -1;
		}
/*  Express the wish, then return.  */
		p->vm.how_many_nodes = 1;
		p->vm.wish_flag[phys_sender]++;
		return;

	}

	if (kind == VM_DESTROY) {
		int i,j;

		if (p == (par_entry *)NULL) {

DEBUG(printg(
"gamma_kernel_rec(): trying to destroy non-existant virtual GAMMA #%d\n",
par_pid
));

			return;
		}

		if (p->vm.random_key != *(typeof(p->vm.random_key) *)addr) {

DEBUG(printg(
"gamma_kernel_rec(): trying to destroy non-existant virtual GAMMA #%d\n",
par_pid
));

			return;
		}

#ifdef MORE_INSTANCES_ON_SAME_NODE
/*  clean up shared ring  */
		for (i=0;i<SHARED_RING_SIZE;i++)
			if (shared_ring.desc[i].par_pid == par_pid)
/*  par_pid #0 is reserved, rx_gamma_samenode() will eventually discard  */
				shared_ring.desc[i].par_pid = 0;
#endif

		for (i=0;i<MAX_NUM_LOCAL_INST;i++)
			if (p->inst_table[i].whoami >= 0)
				for (j=0;j<=MAX_PORT_NUMBER;j++)
				if (p->inst_table[i].port_table[j].header)
				kfree(p->inst_table[i].port_table[j].header);
		free_pages((unsigned long)p,par_entry_magnitude);
		par_proc_table[par_pid] = (par_entry *)NULL;

DEBUG(printg("gamma_kernel_rec(): par_pid %d destroyed\n", par_pid));

		return;

	}

	if (kind == VM_ABORT) {
		int i,j;

		if (p == (par_entry *)NULL) {

DEBUG(printg(
"gamma_kernel_rec(): trying to abort non-existant virtual GAMMA #%d from remote node\n",
par_pid
));

			return;
		}

		if (p->vm.random_key != *(typeof(p->vm.random_key) *)addr) {

DEBUG(printg(
"gamma_kernel_rec(): trying to abort non-existant virtual GAMMA #%d from remote node\n",
par_pid
));

			return;
		}

/*  Quit local process instances, if any  */
		for (i=0;i<MAX_NUM_LOCAL_INST;i++)
			if (p->inst_table[i].whoami >= 0)
			if ((p->inst_table[i].task) != NULL)
/*  When does p->task become NULL???  */
			if ((p->inst_table[i].task)->state != TASK_ZOMBIE)
			if (
			 (p->inst_table[i].task)->pid == p->inst_table[i].pid
			)
				force_sig(SIGTERM, p->inst_table[i].task);

#ifdef MORE_INSTANCES_ON_SAME_NODE
/*  clean up shared ring  */
		for (i=0;i<SHARED_RING_SIZE;i++)
			if (shared_ring.desc[i].par_pid == par_pid)
/*  par_pid #0 is reserved, rx_gamma_samenode() will eventually discard  */
				shared_ring.desc[i].par_pid = 0;
#endif

		for (i=0;i<MAX_NUM_LOCAL_INST;i++)
			if (p->inst_table[i].whoami >= 0)
				for (j=0;j<=MAX_PORT_NUMBER;j++)
				if (p->inst_table[i].port_table[j].header)
				kfree(p->inst_table[i].port_table[j].header);
		free_pages((unsigned long)p,par_entry_magnitude);
		par_proc_table[par_pid] = (par_entry *)NULL;

DEBUG(printg("gamma_kernel_rec(): par_pid %d aborted\n", par_pid));

		return;

	}

	if (kind == VM_CREATE) {
		int j,procnum;
		unsigned long i;

/*
 *  It is possible that an entry for virtual GAMMA with id #par_pid
 *  be already allocated due to a previous call to sys_gamma_vm_wish()
 *  (on a local as well as remote node).  So, no error if entry already
 *  allocated.
 */
		if (p == (par_entry *)NULL) {
/*  Allocate a descriptor for a new virtual GAMMA  */
			par_proc_table[par_pid] = (par_entry *)
				__get_free_pages(GFP_ATOMIC,par_entry_magnitude);
			p = par_proc_table[par_pid];
			if (p == (par_entry *)NULL) {
printg("gamma_kernel_rec(): kmalloc() failed when creating virtual GAMMA #%d\n",par_pid);
				return;
			}
			for (i=0;i<MAX_NUM_LOCAL_INST;i++) {
				p->inst_table[i].pid = -1;
				p->inst_table[i].task = NULL;
				p->inst_table[i].whoami = -1;
			}
		}

/*  Get and store the virtual GAMMA descriptor  */
                memcpy(&(p->vm),addr,len);
/*
 *  Compute the table of ethernet addresses of the new virtual GAMMA, and
 *  instance # ("whoami") of all the "procnum" processes to be run on the
 *  local node
 *  (-1 iff local node is not involved at all in the virtual GAMMA).
 */
		procnum = 0;
		for (i=0;i<p->vm.how_many_nodes;i++) {
			p->eth_table[i] = eth_table[p->vm.node_list[i]];
			if (p->vm.node_list[i] == phys_whoami) {
/*
 *  Involved node: set "whoami" to the instance # and allocate room for
 *  per-port pre-computed pkt headers
 */
			procnum++;
			if (procnum > MAX_NUM_LOCAL_INST) {
printg("gamma_kernel_rec():  too many proc instances on this node when creating virtual GAMMA %d.  Aborting...\n",par_pid);
/*  Abort virtual GAMMA locally and globally.  */
				sys_gamma_vm_abort(par_pid);
				return;
			}
			p->inst_table[procnum-1].whoami = i;
			p->inst_table[procnum-1].has_pkts_in_rx_tmp_ring = 0;
			for (j=0;j<=MAX_PORT_NUMBER;j++) {
				pp = &p->inst_table[procnum-1].port_table[j];
				pp->semaphore =
				pp->boundary_1p =
				pp->in_size =
				pp->in_room = 0;
				pp->in_buffer =
				pp->in_progress = NULL;
				pp->buffer_kind = GO_BACK;
#ifdef MORE_INSTANCES_ON_SAME_NODE
				pp->partner_is_same_node =
				pp->bcast_was_susp = 0;
#endif
				pp->header = (l2s *)kmalloc(
					HEADER_SIZE+MAX_1ST_PIECE_SIZE,
					GFP_ATOMIC
				);
				if (pp->header == NULL) {
printg("gamma_kernel_rec(): kmalloc() failed when creating virtual GAMMA #%d\n",par_pid);
					return;
				}
			}

DEBUG(
{
unsigned long k;
if (procnum == 1) {
	printg(
"gamma_kernel_rec():  Virtual GAMMA %d: how_many_nodes %ld, whoami %d,\n",
par_pid, p->vm.how_many_nodes, p->inst_table[procnum-1].whoami
	);
	printg("node list:\n");
	for (k=0;k<p->vm.how_many_nodes;k++)
		printg("%d ",p->vm.node_list[k]);
	printg("\n");
	printg("phys_whoami %ld\n",phys_whoami);
	for (k=0;k<p->vm.how_many_nodes;k++)
		printg(
			"%ld: %2.2x %2.2x %2.2x %2.2x %2.2x %2.2x\n",
			k,
			p->eth_table[k].addr[0],
			p->eth_table[k].addr[1],
			p->eth_table[k].addr[2],
			p->eth_table[k].addr[3],
			p->eth_table[k].addr[4],
			p->eth_table[k].addr[5]
		);
}
printg("Local instance %d\n",p->inst_table[procnum-1].whoami);
}
);

			}
		}
		p->num_locinst = procnum;
		return;
	}

}


#ifdef MORE_INSTANCES_ON_SAME_NODE
/************************************************************************
 *	Intra-node send/receive routines				*
 ************************************************************************/


static int gamma_send_2p_samenode
(
	unsigned long my_par_pid,
	unsigned long my_procnum,
	unsigned long out_port,
	void *data1, unsigned long len1,
	void *data2, unsigned long len2,
	unsigned long is_bcast
)
{

	unsigned long dest_par_pid;
	unsigned long dest_procnum;
	unsigned long dest_port;
	unsigned long len;
	unsigned long dirty;
	unsigned long cnk_len;
	task_t *receiver;
	unsigned long num_locdest = 1;
	unsigned long i;
	unsigned long ring_room;


	if (len1 > MAX_1ST_PIECE_SIZE)
		goto first_piece_too_long;

	dest_par_pid =
		par_proc_table[my_par_pid]->inst_table[my_procnum]
		.port_table[out_port].partner_par_pid;
	dest_procnum =
		par_proc_table[my_par_pid]->inst_table[my_procnum]
		.port_table[out_port].partner_procnum,
	dest_port =
		par_proc_table[my_par_pid]->inst_table[my_procnum]
		.port_table[out_port].partner_port;

	if (is_bcast)
		goto send_bcast;

DEBUG(printg("gamma_send_2p_samenode: enter: len1 %ld, len2 %ld\n",len1,len2));

send_bcast_back:

	receiver =
		par_proc_table[dest_par_pid]->inst_table[dest_procnum]
		.port_table[dest_port].task;
	if (receiver == NULL)
		goto no_partner;

	len = len1+len2;

DEBUG(printg(
"gamma_send_2p_samenode: my_procnum %ld, num_locdest %ld, dest_procnum %ld\n",
my_procnum,num_locdest,dest_procnum
));

	ring_room = SHARED_RING_SIZE-(shared_ring.dirty-shared_ring.head);
	if (ring_room <= 4)
		goto ring_full;

	dirty = shared_ring.dirty;  /* this is the next avail pos */
	incr_modulo(shared_ring.dirty,SHARED_RING_SIZE);

	shared_ring.desc[dirty].par_pid = dest_par_pid;
	shared_ring.desc[dirty].procnum = dest_procnum;
	shared_ring.desc[dirty].port = dest_port;
	cnk_len = gmin(len,PAGE_SIZE);
	shared_ring.desc[dirty].cnk_len = cnk_len;
	if (cnk_len < len)
		shared_ring.desc[dirty].bitmap = IS_1ST_CNK;
	else
		shared_ring.desc[dirty].bitmap = IS_1ST_CNK | IS_LAST_CNK;
	shared_ring.desc[dirty].num_locdest = num_locdest;
	shared_ring.desc[dirty].watch_time.hi = 0;
	shared_ring.desc[dirty].watch_time.lo = 0;
	shared_ring.desc[dirty].sender = current;
	copy_from_user(shared_ring.desc[dirty].buf,data1,len1);

DEBUG(printg(
"gamma_send_2p_samenode(): par_pid %ld, procnum %ld, port %ld, len %ld\n",
shared_ring.desc[dirty].par_pid,shared_ring.desc[dirty].procnum,
shared_ring.desc[dirty].port,len
));

	for (;;) {

DEBUG(printg(
"gamma_send_2p_samenode(): cnk_len %ld, head %ld, tail %ld, dirty %ld\n",
shared_ring.desc[dirty].cnk_len,shared_ring.head,shared_ring.tail,dirty
));

		copy_from_user(
			(void *)((char *)
			(shared_ring.desc[dirty].buf)+len1),
			data2,
			cnk_len-len1
		);

/*  commit this chunk  */
		incr_modulo(shared_ring.tail,SHARED_RING_SIZE);

		len -= cnk_len;
		if (!len)
			break;
/*  Next chunk...  */
		data2 = (void *)((char *)data2+cnk_len-len1);
		len1 = 0;
		ring_room =
			SHARED_RING_SIZE-(shared_ring.dirty-shared_ring.head);
		if (ring_room <= 4)
			goto ring_full;

		dirty = shared_ring.dirty;  /* this is the next avail pos */
		incr_modulo(shared_ring.dirty,SHARED_RING_SIZE);

		shared_ring.desc[dirty].par_pid = dest_par_pid;
		shared_ring.desc[dirty].procnum = dest_procnum;
		shared_ring.desc[dirty].port = dest_port;
		cnk_len = gmin(len,PAGE_SIZE);
		shared_ring.desc[dirty].cnk_len = cnk_len;
		if (cnk_len == len)
			shared_ring.desc[dirty].bitmap = IS_LAST_CNK;
		shared_ring.desc[dirty].num_locdest = num_locdest;
		shared_ring.desc[dirty].watch_time.hi = 0;
		shared_ring.desc[dirty].watch_time.lo = 0;
		shared_ring.desc[dirty].sender = current;
	}

#if 0
	if (!atomic_read(&gamma_susp_irq_count))
/*
 *  IRQ-driven receive is enabled.
 *  Issue a SIGURG to the receiver process, to force it to eventually
 *  poll the shared buffer (the signal simulates an IRQ).
 *  NOTE: in case of bcast, only the receiver with smallest "procnum" will
 *  get the signal (it will propagate it to the others, however; see
 *  rx_gamma_samenode()).
 */
		force_sig(SIGURG,receiver);
#endif

/*  Exit  */

DEBUG(printg(
"gamma_send_2p_samenode():  Done! head %ld, tail %ld\n",
shared_ring.head,shared_ring.tail
));

	return 0;

send_bcast:

/*  How many receivers on this node?  */
	if (
		par_proc_table[my_par_pid]->inst_table[my_procnum]
		.port_table[out_port].partner_par_pid
		==
		my_par_pid
	)
/*  Sending to same group:  all members but sender count.  */
		num_locdest = par_proc_table[my_par_pid]->num_locinst-1;
	else
/*  Sending to different group:  all members of dest group count.  */
		num_locdest = par_proc_table[
			par_proc_table[my_par_pid]->inst_table[my_procnum]
			.port_table[out_port].partner_par_pid
		]->num_locinst;
/*
 *  In case of bcast on local node, the receiver with smallest
 *  "procnum" is appointed to be the first to operate the receive
 *  (all entries produced in the shared ring are marked with its "procnum").
 *  A "propagation" mechanism is implemented in rx_gamma_samenode().
 *  Take note of the smallest "procnum" among receivers.
 */
	for (i=0;i<MAX_NUM_LOCAL_INST;i++) {
		if (
			par_proc_table[
			par_proc_table[my_par_pid]
			->inst_table[my_procnum].port_table[out_port]
			.partner_par_pid
			]->inst_table[i].whoami
			< 0
		)
			/*  No more local receivers  */
			break;
		if (
			par_proc_table[my_par_pid]->inst_table[my_procnum]
			.port_table[out_port].partner_par_pid
			==
			my_par_pid
		) {
/*  Sending to same group:  sender does not count as receiver.  */
			if (i != my_procnum) {
				dest_procnum = i;
				break;
			}
		}
		else {
			dest_procnum = i;
			break;
		}
	}
/*  Do we have local receivers actually?  */
	if (num_locdest) {

DEBUG(printg("gamma_send_2p_samenode: enter: len1 %ld, len2 %ld\n",len1,len2));
DEBUG(printg(
"gamma_send_2p_samenode: bcast! my_procnum %ld, num_locdest %ld, dest_procnum %ld, len1 %ld, len2 %ld\n",
my_procnum,num_locdest,dest_procnum,len1,len2
));

		goto send_bcast_back;
	}
/*  No local receivers:  exit. */
	return 0;

ring_full:

	return -GAMMA_ERR_GENERIC;

first_piece_too_long:
no_partner:

	return -GAMMA_ERR_GENERIC;

}


static int gamma_send_2p_flowctl_samenode
(
	unsigned long my_par_pid,
	unsigned long my_procnum,
	unsigned long out_port,
	void *data1, unsigned long len1,
	void *data2, unsigned long len2,
	unsigned long is_bcast,
	unsigned long was_susp
)
{

	unsigned long dest_par_pid;
	unsigned long dest_procnum;
	unsigned long dest_port;
	unsigned long ring_room;
	unsigned long dirty;
	unsigned long len;
	unsigned long cnk_len;
	task_t *receiver;
	unsigned long num_locdest = 1;
	unsigned long i;


	if (len1 > MAX_1ST_PIECE_SIZE)
		goto first_piece_too_long;

	dest_par_pid =
		par_proc_table[my_par_pid]->inst_table[my_procnum]
		.port_table[out_port].partner_par_pid;
	dest_procnum =
		par_proc_table[my_par_pid]->inst_table[my_procnum]
		.port_table[out_port].partner_procnum,
	dest_port =
		par_proc_table[my_par_pid]->inst_table[my_procnum]
		.port_table[out_port].partner_port;

	if (is_bcast)
/*
 *  Compute who is the first receiver in the bcast group
 *  (will possibly change the value of "dest_procnum").
 */
		goto send_bcast;

DEBUG(printg(
"gamma_send_2p_flowctl_samenode: enter: len1 %ld, len2 %ld\n",len1,len2
));

send_bcast_back:

	receiver =
		par_proc_table[dest_par_pid]->inst_table[dest_procnum]
		.port_table[dest_port].task;
	if (receiver == NULL)
		goto no_partner;

	ring_room = SHARED_RING_SIZE-(shared_ring.dirty-shared_ring.head);
	if (ring_room <= 4)
		goto ring_full;

	dirty = shared_ring.dirty;  /* this is the next avail pos */
	incr_modulo(shared_ring.dirty,SHARED_RING_SIZE);

	len = len1+len2;

DEBUG(printg(
"gamma_send_2p_flowctl_samenode: my_procnum %ld, num_locdest %ld, dest_procnum %ld\n",
my_procnum,num_locdest,dest_procnum
));

	shared_ring.desc[dirty].par_pid = dest_par_pid;
	shared_ring.desc[dirty].procnum = dest_procnum;
	shared_ring.desc[dirty].port = dest_port;
	cnk_len = gmin(len,PAGE_SIZE);
	shared_ring.desc[dirty].cnk_len = cnk_len;
	if (was_susp) {
		if (cnk_len >= len)
			shared_ring.desc[dirty].bitmap = IS_LAST_CNK;
	}
	else
		if (cnk_len < len)
			shared_ring.desc[dirty].bitmap = IS_1ST_CNK;
		else
			shared_ring.desc[dirty].bitmap =
				IS_1ST_CNK | IS_LAST_CNK;
	shared_ring.desc[dirty].num_locdest = num_locdest;
	shared_ring.desc[dirty].watch_time.hi = 0;
	shared_ring.desc[dirty].watch_time.lo = 0;
	shared_ring.desc[dirty].sender = current;
	copy_from_user(shared_ring.desc[dirty].buf,data1,len1);

DEBUG(printg(
"gamma_send_2p_flowctl_samenode(): par_pid %ld, procnum %ld, port %ld, len %ld\n",
shared_ring.desc[dirty].par_pid,shared_ring.desc[dirty].procnum,
shared_ring.desc[dirty].port,len
));

	for (;;) {

DEBUG(printg(
"gamma_send_2p_flowctl_samenode(): cnk_len %ld, head %ld, tail %ld, dirty %ld\n",
shared_ring.desc[dirty].cnk_len,shared_ring.head,shared_ring.tail,dirty
));

		copy_from_user(
			(void *)((char *)
			(shared_ring.desc[dirty].buf)+len1),
			data2,
			cnk_len-len1
		);

/*  commit this chunk  */
		incr_modulo(shared_ring.tail,SHARED_RING_SIZE);

		len -= cnk_len;
		if (!len)
			break;
/*  Next chunk...  */
		data2 = (void *)((char *)data2+cnk_len-len1);
		len1 = 0;
check_room:
		ring_room =
			SHARED_RING_SIZE-(shared_ring.dirty-shared_ring.head);
		if (ring_room <= 4)
			goto wait_room;

		dirty = shared_ring.dirty;  /* this is the next avail pos */
		incr_modulo(shared_ring.dirty,SHARED_RING_SIZE);

		shared_ring.desc[dirty].par_pid = dest_par_pid;
		shared_ring.desc[dirty].procnum = dest_procnum;
		shared_ring.desc[dirty].port = dest_port;
		cnk_len = gmin(len,PAGE_SIZE);
		shared_ring.desc[dirty].cnk_len = cnk_len;
		if (cnk_len == len)
			shared_ring.desc[dirty].bitmap = IS_LAST_CNK;
		shared_ring.desc[dirty].num_locdest = num_locdest;
		shared_ring.desc[dirty].watch_time.hi = 0;
		shared_ring.desc[dirty].watch_time.lo = 0;
		shared_ring.desc[dirty].sender = current;
		continue;
wait_room:
		wait_room_in_shared_ring(i,my_par_pid,my_procnum);
		if (i)
/*  try to resume transmission  */
			goto check_room;
/*  still no room after waiting time.  suspend...  */
		break;
	}

	if (!len) {
/*  Transmission complete.  Exit  */

#if 0
		if (!atomic_read(&gamma_susp_irq_count))
/*
 *  IRQ-driven receive is enabled.
 *  Issue a SIGURG to the receiver process, to force it to eventually
 *  poll the shared buffer (the signal simulates an IRQ).
 *  NOTE: in case of bcast, only the current receiver ("dest_procnum",
 *  which also is the one with smallest "procnum") will get the signal
 *  (it will propagate it to the others, however; see rx_gamma_samenode()).
 */
			force_sig(SIGURG,receiver);
#endif

DEBUG(printg(
"gamma_send_2p_flowctl_samenode():  Done! head %ld, tail %ld\n",
shared_ring.head,shared_ring.tail
));

		return 0;
	}
/*
 *  Transmission suspended due to ring full.
 */
	if (is_bcast) {
/*
 *  Take note that the operation was suspended while bcasting in the
 *  same node.
 */
		par_proc_table[my_par_pid]->inst_table[my_procnum]
		.port_table[out_port].bcast_was_susp = 1;

DEBUG(printg("gamma_send_2p_flowctl_samenode(): suspend bcast.\n"));

	}

#if 0
	if (!atomic_read(&gamma_susp_irq_count))
/*
 *  IRQ-driven receive is enabled.
 *  Issue a SIGURG to the next receiver process, to force it to eventually
 *  poll the shared buffer (the signal simulates an IRQ).
 *  NOTE: in case of bcast, only the current receiver ("dest_procnum",
 *  which also is the one with smallest "procnum") will get the signal
 *  (it will propagate it to the others, however; see rx_gamma_samenode()).
 */
		force_sig(SIGURG,receiver);
#endif

DEBUG(printg(
"gamma_send_2p_flowctl_samenode(): head %ld,tail %ld: No room.  Suspend (residual len %ld)\n",
shared_ring.head,shared_ring.tail,len
));

/*  Return how many bytes are still to be transmitted.  */
	return len;

send_bcast:

	if (
		par_proc_table[my_par_pid]->inst_table[my_procnum]
		.port_table[out_port].bcast_was_susp
	) {
/*  Bcast in the same node was suspended due to ring full.  */
		par_proc_table[my_par_pid]->inst_table[my_procnum]
		.port_table[out_port].bcast_was_susp = 0;

DEBUG(printg("gamma_send_2p_flowctl_samenode(): resume bcast.\n"));

	}
/*  How many receivers on this node?  */
	if (
		par_proc_table[my_par_pid]->inst_table[my_procnum]
		.port_table[out_port].partner_par_pid
		==
		my_par_pid
	)
/*  Sending to same group:  all members but sender count.  */
		num_locdest = par_proc_table[my_par_pid]->num_locinst-1;
	else
/*  Sending to different group:  all members of dest group count.  */
		num_locdest = par_proc_table[
			par_proc_table[my_par_pid]->inst_table[my_procnum]
			.port_table[out_port].partner_par_pid
		]->num_locinst;
/*
 *  In case of bcast on local node, the receiver with smallest
 *  "procnum" is appointed to be the first to operate the receive
 *  (all entries produced in the shared ring are marked with its "procnum").
 *  A "propagation" mechanism is implemented in rx_gamma_samenode().
 *  Take note of the smallest "procnum" among receivers.
 */
	for (i=0;i<MAX_NUM_LOCAL_INST;i++) {
		if (
			par_proc_table[
			par_proc_table[my_par_pid]
			->inst_table[my_procnum].port_table[out_port]
			.partner_par_pid
			]->inst_table[i].whoami
			< 0
		)
			/*  No more local receivers  */
			break;
		if (
			par_proc_table[my_par_pid]->inst_table[my_procnum]
			.port_table[out_port].partner_par_pid
			==
			my_par_pid
		) {
/*  Sending to same group:  sender does not count as receiver.  */
			if (i != my_procnum) {
				dest_procnum = i;
				break;
			}
		}
		else {
			dest_procnum = i;
			break;
		}
	}
/*  Do we have local receivers actually?  */
	if (num_locdest) {

DEBUG(printg(
"gamma_send_2p_flowctl_samenode: enter: len1 %ld, len2 %ld\n",len1,len2
));
DEBUG(printg(
"gamma_send_2p_flowctl_samenode: bcast! my_procnum %ld, num_locdest %ld, dest_procnum %ld, len1 %ld, len2 %ld\n",
my_procnum,num_locdest,dest_procnum,len1,len2
));

		goto send_bcast_back;
	}
/*  No local receivers:  exit. */
	return 0;

ring_full:

	wait_room_in_shared_ring(i,my_par_pid,my_procnum);
	if (i)
		goto send_bcast_back;  /*  check again for available room  */
	if (is_bcast) {
/*
 *  Take note that the operation was suspended while bcasting in the
 *  same node.
 */
		par_proc_table[my_par_pid]->inst_table[my_procnum]
		.port_table[out_port].bcast_was_susp = 1;

DEBUG(printg("gamma_send_2p_flowctl_samenode(): suspend bcast.\n"));

	}

DEBUG_E(printg(
"gamma_send_2p_flowctl_samenode(): head %ld,tail %ld: No room.  Try again!\n",
shared_ring.head,shared_ring.tail
));

	return -GAMMA_ERR_TRYAGAIN;

first_piece_too_long:
no_partner:

	return -GAMMA_ERR_GENERIC;

}


/*
 *  Returns:
 *
 *  P | IS_SAME_NODE (0 <= P <= MAX_PORT_NUMBER) if msg complete for port P
 *  (user-level caller will perform notification on sem #P and run rec handler
 *  if present; msg lenght is in user-level variable "gamma_msglen")
 *
 *  P | IS_FIRST_CHUNK | IS_SAME_NODE if msg not complete for port P but
 *  rec handler must run to manage the arrival of 1st piece of msg
 *  (user-level caller will run the handler, 1st piece lenght is in user-level
 *  variable "gamma_msglen")
 *
 *  -GAMMA_RX_INCOMPLETE if no complete msg and no handler to run
 *
 *  -GAMMA_ERR_MUSTYIELD if there is a pending pkt for a process other than the
 *  invoker
 *
 *  -GAMMA_ERR_BUSY in some other cases.
 */
static int rx_gamma_samenode
(unsigned long flag)
{

int res;
unsigned long bcast_sequence;
unsigned long head;
unsigned long par_pid;
unsigned long procnum;
unsigned long port;
unsigned long cnk_len;
unsigned long bitmap;
unsigned long num_locdest;
unsigned long first_piece_size;
unsigned long rx_buffer;
unsigned long next_procn = 0;
task_t *receiver;


if (!spin_trylock(&gamma_rxsamenode_lock))
 return -GAMMA_ERR_BUSY;

if (gamma_must_yield_samenode != NULL)
 goto check_yield;

ok_yield:

if (gamma_run_handler_samenode != NULL)
 goto rec_reentering;

DEBUG_E(printg("rx_gamma_samenode(), pid %u: entering...\n",current->pid));

/* shared ring might have been consumed by another CPU meanwhile...? */
if (shared_ring.head == shared_ring.tail)
 goto bogus_inv;

res = -GAMMA_RX_INCOMPLETE;
bcast_sequence = 0;
head = shared_ring.head;

for (;;) {

DEBUG_E(printg(
"rx_gamma_samenode(): loop: head %ld, tail %ld, bitmap %ld\n",
head,shared_ring.tail,shared_ring.desc[head].bitmap
));

 par_pid = shared_ring.desc[head].par_pid;
 procnum = shared_ring.desc[head].procnum;
 port = shared_ring.desc[head].port;
 cnk_len = shared_ring.desc[head].cnk_len;
 bitmap = shared_ring.desc[head].bitmap;
 num_locdest = shared_ring.desc[head].num_locdest;

DEBUG(printg(
"rx_gamma_samenode(), pid %u: par_pid %ld, procnum %ld, port %ld, cnk_len %ld\n",
current->pid,par_pid,procnum,port,cnk_len
));

/*  Check whether virtual GAMMA #par_pid exists and is assigned.  */
 if (
  (par_proc_table[par_pid] == (par_entry *)NULL)
  ||
  (par_proc_table[par_pid]->inst_table[procnum].pid == -1)
 )
  goto discard_pkt;

/*  Check whether port is in use as an input one.  */
 receiver =
  par_proc_table[par_pid]->inst_table[procnum].port_table[port].task;
 if (receiver == NULL)
  goto discard_pkt;

/*  Check whether receiver is the caller.  */
 if (receiver->mm != current->mm)
  goto cannot_deliver_yet;

#if 0
/*
 *  If caller is on send() and this is an end-of-msg, suspend and leave pkt
 *  in rx queue; receiver will take it on receive() and perform notification
 */
 if (unlikely(flag == FROM_SEND_ROUTINE)) {
  if (bitmap & IS_LAST_CNK) {
   spin_unlock(&gamma_rxsamenode_lock);
   return -GAMMA_RX_SUSP;
  }
 }
#endif

/* Check whether a seq of bcast pkts is beginning or has ended */
 if ((num_locdest > 1) && (!bcast_sequence))
  goto found_bcast_chunk;
 else
  if ((num_locdest == 1) && bcast_sequence) {
/*
 *  A seq. of bcast pkts has ended.
 *  Quit the routine, to allow other receivers to get the bcast data
 *  before considering the current pkt.
 *
 *  NOTE: variable "gamma_must_yield", indicating which proc instance should
 *  run next in the bcast recv, has been set up when the bcast seq begun.
 */

DEBUG(printg(
"rx_gamma_samenode(): exiting (end of bcast seq), head %ld (%ld), tail %ld\n",
shared_ring.head,head,shared_ring.tail
));

   spin_unlock(&gamma_rxsamenode_lock);
   return -GAMMA_ERR_MUSTYIELD;
  }

found_bcast_chunk_back:

/*  Empty messages:  no buffer, no frame numbering.  */
 if (!cnk_len)
  goto end_recv_samenode;

/*  Begin of payload transfer.  */

 par_proc_table[par_pid]->inst_table[procnum].port_table[port]
 .current_rx_msglen
  += cnk_len;

/*
 *  Is this packet to be scattered into two pieces?
 *  If present, 1st piece is carried by 1st packet of the msg.
 */

 first_piece_size = 0;

 if (bitmap & IS_1ST_CNK) {
  first_piece_size =
   par_proc_table[par_pid]->inst_table[procnum].port_table[port].boundary_1p;
  if ((first_piece_size > 0) && (first_piece_size < cnk_len)) {
/*  Payload is to be scattered into two pieces.  Download 1st piece.  */

#if 0
/*
 *  Download 1st piece if caller is on receive; suspend and leave pkt in queue
 *  if caller in on send (flow ctl credit waiting)
 */
   if (unlikely(flag == FROM_SEND_ROUTINE)) {
    spin_unlock(&gamma_rxsamenode_lock);
    return -GAMMA_RX_SUSP;
   }
#endif

DEBUG(printg(
"rx_gamma_samenode():  Msg to be scattered, size of first piece is %ld\n",
first_piece_size
));

   get_ptr_to_user_recv_buffer(par_pid,port,procnum,rx_buffer);
   copy_to_user((void *)rx_buffer,shared_ring.desc[head].buf,first_piece_size);
   cnk_len -= first_piece_size;

/*
 *  Prepare to quit, to run the receiver handler of the port.
 *  Since this message is going to be scattered, we can assume
 *  that there is a handler to run.
 */

/*
 *  Tell the caller which handler to run, and len of 1st chunk of data:
 *  "intra-node msg not complete but must run handler"
 */
   res = port | IS_FIRST_CHUNK | IS_SAME_NODE;

   set_msglen(par_pid,procnum,first_piece_size);

/*
 *  Current process only is allowed to reenter later,
 *  to download remaining part of payload.
 *  Note that the user-level destination of 2nd piece may have been
 *  set/reset/changed by the receiver handler invoked above.
 *  If not, 2nd piece will overwrite 1st one at the same user-level
 *  destination address, since the macro
 *  update_ptr_to_user_recv_buffer() (see below)
 *  has not been invoked.
 */
   gamma_run_handler_samenode = current;

DEBUG(printg("Quit to run rec handler for port %ld (1st piece)\n",port));

   spin_unlock(&gamma_rxsamenode_lock);
   return res;

  }
  else
   first_piece_size = 0;
 }

get_remaining_payload:

/*
 *  Shift user-space rec buffer's current ptr forward.
 *  Check for room availability.
 *  If not enough room in rec buffer, truncate the msg chunk
 *  by setting cnk_len equal to the available room.
 *  Decrement the available room by cnk_len.
 *  The current msg chunk will be stored starting
 *  from address "rx_buffer".
 */
 update_ptr_to_user_recv_buffer(par_pid,port,procnum,cnk_len,rx_buffer);

/*  Copy the payload (2nd piece) into user space  */
 copy_to_user(
  (void *)rx_buffer,
  (void *)((char *)(shared_ring.desc[head].buf)+first_piece_size),
  cnk_len
 );

end_recv_samenode:

/*  Empty messages:  no buffer restore, no payload transfer.  */
 if (!shared_ring.desc[head].cnk_len)
  goto notif;

 if (!(bitmap & IS_LAST_CNK))
  goto msg_not_complete;

/*
 *  Do one of the following, depending on the kind of buffer attached
 *  to the recv port:
 *  case GO_BACK:
 *  reset the current position ptr of the recv buffer to the begin of
 *  the recv buffer itself.  Next msg will be stored at the begin
 *  of the buffer.
 *  case GO_AHEAD:
 *  If there is still room available in the attached buffer,
 *  leave the current position ptr untouched.
 *  Otherwise, reset the current position ptr of the recv buffer to
 *  the begin of the recv buffer itself; in this case, next msg will
 *  be stored at the begin of the buffer.
 */
 switch (
  par_proc_table[par_pid]->inst_table[procnum].port_table[port].buffer_kind
 ) {
  case GO_BACK:
   reset_current_ptr(par_pid,port,procnum);
   break;
  case GO_AHEAD:
   check_available_room_and_possibly_reset_current_ptr(par_pid,port,procnum);
   break;
 }

notif:

/*
 *  Prepare to quit, to notify msg arrival and run the receiver handler
 *  (if a handler is present and it did not run already)
 */
 res = port;

DEBUG_E(printg(
"Setting \"gamma_msglen\" to %ld\n",
par_proc_table[par_pid]->inst_table[procnum].port_table[port].current_rx_msglen
));

 set_msglen(
  par_pid,procnum,
  par_proc_table[par_pid]->inst_table[procnum].port_table[port]
   .current_rx_msglen
 );

/*  Next msg starts from len = 0  */
 par_proc_table[par_pid]->inst_table[procnum]
 .port_table[port].current_rx_msglen = 0;

 if (!bcast_sequence) {
/*  Not a bcast:  discard pkt as usual.  */
  shared_ring.desc[head].bitmap = 0;
  incr_modulo(shared_ring.head,SHARED_RING_SIZE);
  break;
 }

/*
 *  It is a bcast pkt:
 *  do not discard, set next receiver, decrease no. of receivers.
 */
 shared_ring.desc[head].procnum = next_procn;
 shared_ring.desc[head].num_locdest--;
 break;

msg_not_complete:

 if (bcast_sequence) {
/*
 *  It is a bcast pkt:
 *  do not discard, set next receiver, decrease no. of receivers,
 *  continue with next pkt if any, but don't increment the head of shared ring;
 *  just increment my own local copy of head, so ring remains unchanged for the
 *  next receiver in the bcast group.
 */
  shared_ring.desc[head].procnum = next_procn;
  shared_ring.desc[head].num_locdest--;
  incr_modulo(head,SHARED_RING_SIZE);
  if (head == shared_ring.tail) {
/*
 *  No more pkts available:  quit rx loop but let next receiver to run.
 *
 *  NOTE: variable "gamma_must_yield", indicating which proc instance should
 *  run next in the bcast recv, has been set up when the bcast seq begun.
 */
   break;
  }
  continue;
 }
/*  Not a bcast:  discard pkt as usual.  */

discard_pkt:

DEBUG_E(printg(
"rx_gamma_samenode(): discarding pkt, head %ld (%ld), tail %ld\n",
shared_ring.head,head,shared_ring.tail
));

 shared_ring.desc[head].bitmap = 0;
 incr_modulo(shared_ring.head,SHARED_RING_SIZE);
 if (shared_ring.head == shared_ring.tail)
/*  No more pkts available:  quit rx loop.  */
  break;

 head = shared_ring.head;
  
DEBUG(printg("rx_gamma_samenode(): continuing loop...\n"));

}

DEBUG(printg(
"rx_gamma_samenode(), pid %u: exiting, head %ld (%ld), tail %ld, res %d\n",
current->pid,shared_ring.head,head,shared_ring.tail,res
));

/*  Quit.  Possibly yield the CPU.  */
if ((gamma_must_yield_samenode != NULL) && (res < 0))
 res = -GAMMA_ERR_MUSTYIELD;
spin_unlock(&gamma_rxsamenode_lock);
return res;

/*  Here all the unusual branches.  */

found_bcast_chunk:

/*
 *  Found a bcast pkt.  We suppose it is the first of a sequence.
 *  Such pkt will not be removed from the ring until all receivers
 *  have copied them into user space.  Instead, remove all the other
 *  pkts received so far (by forwarding the ring tail to the current value).
 *  Also, pre-compute the next local receiver and propagate SIGURG to it
 *  if required.
 */
bcast_sequence = 1;
/*  Search for next receiver.  */
for (next_procn = procnum+1; next_procn < MAX_NUM_LOCAL_INST; next_procn++)
/*  We must exclude the sender from the search.  */
 if (
  (par_proc_table[par_pid]->inst_table[next_procn].task
  != shared_ring.desc[head].sender)
  && (par_proc_table[par_pid]->inst_table[next_procn].whoami >= 0)
 )
  break;
if (next_procn == MAX_NUM_LOCAL_INST) {
/*
 *  Next receiver not found.  Should not have happened: the sender had
 *  indicated more than one local receiver is to get this chunk of data
 *  (shared_ring.desc[head].num_locdest > 1), so there is something bad...
 */

DEBUG_L(printg(
"rx_gamma_samenode(): Failed to find next receiver (bcast)!!  Discard...\n"
));

 goto discard_pkt;
}

#if 0
if (!atomic_read(&gamma_susp_irq_count)) {
/*
 *  IRQ-driven receive is enabled.
 *  Issue a SIGURG to the next receiver process, to force it to eventually
 *  poll the shared buffer (the signal simulates an IRQ).
 */

DEBUG(printg(
"rx_gamma_samenode(): Propagating SIGURG to procnum %ld (bcast)\n", next_procn
));

 force_sig(SIGURG,par_proc_table[par_pid]->inst_table[next_procn].task);
}
#endif

/*  Prepare to yield CPU to next receiver in bcast group. */
gamma_must_yield_samenode = par_proc_table[par_pid]->inst_table[next_procn].task;
goto found_bcast_chunk_back;

cannot_deliver_yet:

{
 task_t *p;
 time_586 cur_time;

 gamma_time(cur_time);

 if (
  !shared_ring.desc[head].watch_time.hi &&
  !shared_ring.desc[head].watch_time.lo
 ) {
  shared_ring.desc[head].watch_time = cur_time;
/*  Quit.  Yield the CPU.  */
  gamma_must_yield_samenode = receiver;
  spin_unlock(&gamma_rxsamenode_lock);
  return -GAMMA_ERR_MUSTYIELD;
 }

 if (
  gamma_time_diff_ticks(cur_time,shared_ring.desc[head].watch_time) < 100*CLOCK
 ) {
/*  Quit.  Yield the CPU.  */
  gamma_must_yield_samenode = receiver;
  spin_unlock(&gamma_rxsamenode_lock);
  return -GAMMA_ERR_MUSTYIELD;
 }

/*
 *  The pkt is stored in the shared ring since more than 100 usec ago.
 *  Is the receiver process alive?  NOTE: we prefer wait 100 usec then
 *  perform the liveness check on receiver, for the check itself is
 *  time-consuming and the frequent case is when the receiver is alive.
 */
 p = find_task_by_pid(receiver->pid);
 if (p == NULL)
  goto dead_process;

 if (receiver->state & (TASK_ZOMBIE | TASK_DEAD))
  goto dead_process;

DEBUG(printg(
"rx_gamma_samenode(): delaying recv, receiver is %u, state is %ld\n",
receiver->pid, receiver->state
));

#if 0
/*
 *  If IRQ-driven receive is disabled, the gamma_send_2p_samenode() that
 *  originated this pkt did not issue any SIGURG to the receiver process at end
 *  of transmission.
 *  So the receiver would not read this pkt before it really needs it.
 *  But in this case we need the pkt to be consumed as soon as possible!
 *
 *  Issue a SIGURG to the receiver process, to force it to poll the shared
 *  buffer.  This simulates an IRQ.
 */
 if (!atomic_read(&gamma_susp_irq_count))
  force_sig(SIGURG,receiver);
#endif

/*  Quit.  Yield the CPU.  */
 gamma_must_yield_samenode = receiver;
 spin_unlock(&gamma_rxsamenode_lock);
 return -GAMMA_ERR_MUSTYIELD;

dead_process:

/*  Receiver is no longer existing.  Discard this packet.  */

DEBUG(printg(
"rx_gamma_samenode(): recv proc does not exist.  Discarding packet...\n"
));

 goto discard_pkt;
}

check_yield:

if (likely(gamma_must_yield_samenode == current)) {
 gamma_must_yield_samenode = NULL;
 goto ok_yield;
}
spin_unlock(&gamma_rxsamenode_lock);
return -GAMMA_ERR_BUSY;

bogus_inv:

DEBUG_E(printg(
"rx_gamma_samenode(): bogus invocation (race on shared ring tail)\n"
));

spin_unlock(&gamma_rxsamenode_lock);
return -GAMMA_ERR_BUSY;

rec_reentering:

if (gamma_run_handler_samenode != current) {
/*
 *  A receiver has temporarily quit to run the rec handler, but will
 *  reenter soon.
 *  Prevent other processes to enter now.
 */
 spin_unlock(&gamma_rxsamenode_lock);
 return -GAMMA_ERR_BUSY;
}
/*  The receiver is reentering after running a rec handler (1sh piece).  */

DEBUG_E(printg("rx_gamma_samenode(), pid %u: reentering...\n",current->pid));

gamma_run_handler_samenode = NULL;
res = -GAMMA_RX_INCOMPLETE;
head = shared_ring.head;

DEBUG_E(printg(
"rx_gamma_samenode(): loop: head %ld, tail %ld, bitmap %ld\n",
head,shared_ring.tail,shared_ring.desc[head].bitmap
));

par_pid = shared_ring.desc[head].par_pid;
procnum = shared_ring.desc[head].procnum;
port = shared_ring.desc[head].port;
cnk_len = shared_ring.desc[head].cnk_len;
bitmap = shared_ring.desc[head].bitmap;
num_locdest = shared_ring.desc[head].num_locdest;

receiver =
 par_proc_table[par_pid]->inst_table[procnum].port_table[port].task;

bcast_sequence = (num_locdest > 1);

DEBUG(printg(
"rx_gamma_samenode(), pid %u, reentering: par_pid %ld, procnum %ld, port %ld, cnk_len %ld\n",
current->pid,par_pid,procnum,port,cnk_len
));

first_piece_size =
 par_proc_table[par_pid]->inst_table[procnum].port_table[port].boundary_1p;

cnk_len -= first_piece_size;

goto get_remaining_payload;

}
#endif


/************************************************************************
 *	Blocking send routines, best-effort				*
 ************************************************************************/


int sys_gamma_send
(
	unsigned long out_port,
	unsigned long my_par_pid,
	unsigned long procnum,
	void *data,
	unsigned long len
)
{

	unsigned long phys_dest;
	tx_tmp_ring_type *win;
#ifdef MORE_INSTANCES_ON_SAME_NODE
	void *ddata;
	unsigned long llen;
#endif
	l2s *h;
	unsigned long entry;
	unsigned long cnk_len;
	unsigned long frame_size;
	tx_common_decl_and_init;
#ifdef ALLOW_IRQS
	time_586 t_start, t_elaps;
	unsigned long flags;
#endif
#ifdef  COLL_SAFE_TX
	int trials = MAX_TRIALS;
#endif
#ifdef MORE_INSTANCES_ON_SAME_NODE
	int res;
#endif
#ifdef MEASURE_OUT_TH_MEM
time_586 ima,imb;
#endif


#ifdef MORE_INSTANCES_ON_SAME_NODE
	if (!spin_trylock(&gamma_tx_lock))
		goto dont_reenter;
#endif

	phys_dest =
		par_proc_table[my_par_pid]->inst_table[procnum]
		.port_table[out_port].phys_partner_node;

	if (phys_dest != BROADCAST)
		win = &tx_tmp_ring[phys_dest];
	else
		win = &tx_tmp_ring[phys_whoami];

#ifdef MORE_INSTANCES_ON_SAME_NODE
	if (
		par_proc_table[my_par_pid]->inst_table[procnum]
		.port_table[out_port].partner_is_same_node
	)
		goto same_node;
#endif

DEBUG_E(printg("sys_gamma_send(): Entering\n"));

#ifdef MORE_INSTANCES_ON_SAME_NODE
	ddata = data;
	llen = len;
#endif

	h =	par_proc_table[my_par_pid]->inst_table[procnum]
		.port_table[out_port].header;
	entry = win->tail;
	incr_modulo(win->tail,GAMMA_TX_WIN_SIZE);
	memcpy(win->buf[entry].b,h,HEADER_SIZE);
	h = (l2s *)(win->buf[entry].b);

	if (!len) {

/*  Fast path for zero-byte messages.  */
	fill_send_header_cnklen(
		h, MSG_1ST_FRAME_MASK | MSG_COMPLETE_MASK
	);

	if (!tx_ring_is_full()) {

room_in_tx_ring1:

#ifdef SEQCHECK
	set_send_header_frameno(h,win->tx_seq++);
#endif
#ifdef COLL_SAFE_TX
ro1:
#endif

DEBUG(printg(
"sending through port %ld, frame_no %d\n",out_port,read_send_header_frameno(h)
));

	tx_enq_pkt1(h,HEADER_SIZE);
	tx_start();
	tx_forw_tail();
#ifdef COLL_SAFE_TX
	tx_flush_ring_1(tx_ring_tail,MAX_FLUSH_WAIT*10,i);
	if ((!i) && (--trials >= 0)) {
DEBUG_L(printg("retrying, frame_no %d\n",read_send_header_frameno(h)));
		goto ro1;
	}
#endif
#ifdef SAFE_TX
	if (tx_ring_occupancy >= STEPS_THEN_STOP)
		tx_flush_ring(tx_ring_tail,MAX_FLUSH_WAIT,i);
#endif
	win->buf[entry].l = HEADER_SIZE;
	goto end_tx;

	}

	tx_wait_ring_room(MAX_FLUSH_WAIT,i);
	if (!i)
		goto tx_failure;
	goto room_in_tx_ring1;

	}

	frame_size = tx_optimal_frag(len);

	cnk_len = gmin(len,frame_size-HEADER_SIZE);

	if (cnk_len < len)
		fill_send_header_cnklen(h, cnk_len | MSG_1ST_FRAME_MASK);
	else
		fill_send_header_cnklen(
			h, cnk_len | MSG_1ST_FRAME_MASK | MSG_COMPLETE_MASK
		);

#ifdef MEASURE_OUT_TH_MEM
gamma_time(ima);
#endif
	copy_from_user(((unsigned char *)h)+HEADER_SIZE,(void *)data,cnk_len);
#ifdef MEASURE_OUT_TH_MEM
gamma_time(imb);
out_counter_mem++;
out_total_amount_mem_delay += (unsigned long)gamma_time_diff_ticks(imb,ima);
out_total_amount_mem_bytes += cnk_len;
#endif

#ifdef ALLOW_IRQS
	gamma_time(t_start);
#endif

	for (;;) {

		if (!tx_ring_is_full()) {

room_in_tx_ring:

#ifdef SEQCHECK
		set_send_header_frameno(h,win->tx_seq++);
#endif
#ifdef COLL_SAFE_TX
ro:
#endif

DEBUG(printg(
"sending through port %ld, frame_no %d\n",out_port,read_send_header_frameno(h)
));

		tx_enq_pkt1(h,HEADER_SIZE+cnk_len);
		tx_start();
		tx_forw_tail();
#ifdef COLL_SAFE_TX
		tx_flush_ring_1(tx_ring_tail,MAX_FLUSH_WAIT*10,i);
		if ((!i) && (--trials >= 0)) {
DEBUG_L(printg("retrying, frame_no %d\n",read_send_header_frameno(h)));
			goto ro;
		}
#endif
#ifdef SAFE_TX
		if (tx_ring_occupancy >= STEPS_THEN_STOP)
			tx_flush_ring(tx_ring_tail,MAX_FLUSH_WAIT,i);
#endif
		win->buf[entry].l = HEADER_SIZE+cnk_len;

		len -= cnk_len;
		if (!len)
			break;

#ifdef ALLOW_IRQS
		gamma_time(t_elaps);
		IRQ_enable_for_5usec(t_elaps,t_start);
#endif

/*  Prepare next frame  */
		entry = win->tail;
		incr_modulo(win->tail,GAMMA_TX_WIN_SIZE);
		memcpy(win->buf[entry].b,h,HEADER_SIZE);
		h = (l2s *)(win->buf[entry].b);
		data = (void *)((char *)data+cnk_len);
		cnk_len = gmin(len,frame_size - HEADER_SIZE);
		if (cnk_len < len)
			fill_send_header_cnklen(h,cnk_len);
		else
			fill_send_header_cnklen(h,cnk_len | MSG_COMPLETE_MASK);

#ifdef MEASURE_OUT_TH_MEM
gamma_time(ima);
#endif
		copy_from_user(
			((unsigned char *)h)+HEADER_SIZE,
			(void *)data,
			cnk_len
		);
#ifdef MEASURE_OUT_TH_MEM
gamma_time(imb);
out_counter_mem++;
out_total_amount_mem_delay += (unsigned long)gamma_time_diff_ticks(imb,ima);
out_total_amount_mem_bytes += cnk_len;
#endif
		continue;

		}

		tx_wait_ring_room(MAX_FLUSH_WAIT,i);
		if (!i)
			goto tx_failure;
		goto room_in_tx_ring;

	}

end_tx:

	tx_close();

DEBUG(printg("sys_gamma_send(): Done!\n"));

#ifdef MORE_INSTANCES_ON_SAME_NODE
	if (phys_dest != BROADCAST) {
		spin_unlock(&gamma_tx_lock);
		return 0;
	}
	res =	gamma_send_2p_samenode(
			my_par_pid,procnum,out_port,(void *)NULL,0,ddata,llen,1
		);
	spin_unlock(&gamma_tx_lock);
	return res;
#else
	return 0;
#endif

tx_failure:

DEBUG_L(printg(
"sys_gamma_send(): par_pid %ld, out_port %ld, frame_no %d: Failure (ring pos %d)\n",
my_par_pid,out_port,rec_header_frame_no(h),tx_ring_head
));

	tx_reset_and_restart();
#ifdef MORE_INSTANCES_ON_SAME_NODE
	spin_unlock(&gamma_tx_lock);
#endif
	return -GAMMA_ERR_SENDFAILURE;

#ifdef MORE_INSTANCES_ON_SAME_NODE
same_node:

	if (phys_dest != BROADCAST) {
		res =	gamma_send_2p_samenode(
				my_par_pid,procnum,out_port,
				(void *)NULL,0,data,len,
				0
			);
		spin_unlock(&gamma_tx_lock);
		return res;
	}
	res =	gamma_send_2p_samenode(
			my_par_pid,procnum,out_port,(void *)NULL,0,data,len,
			1
		);
	spin_unlock(&gamma_tx_lock);
	return res;

dont_reenter:

	return -GAMMA_ERR_TRYAGAIN;
#endif

}


int sys_gamma_send_flowctl
(
	unsigned long out_port_and_was_susp,
	unsigned long my_par_pid,
	unsigned long procnum,
	void *data,
	unsigned long len
)
{

	unsigned long out_port;
	unsigned long was_susp;
	unsigned long phys_dest;
	tx_tmp_ring_type *win;
	unsigned long credit;
	unsigned long curr_credit;
	l2s *h;
	unsigned long entry;
	unsigned long cnk_len;
	unsigned long frame_size;
	tx_common_decl_and_init;
#ifdef ALLOW_IRQS
	time_586 t_start, t_elaps;
	unsigned long flags;
#endif
#ifdef COLL_SAFE_TX
	int trials = MAX_TRIALS;
#endif
#ifdef MORE_INSTANCES_ON_SAME_NODE
	int res;
#endif
#ifdef MEASURE_OUT_TH_MEM
time_586 ima,imb;
#endif


#ifdef MORE_INSTANCES_ON_SAME_NODE
	if (!spin_trylock(&gamma_tx_lock))
		goto dont_reenter;
#endif

	out_port = out_port_and_was_susp & ~0x80000000;
	was_susp = out_port_and_was_susp & 0x80000000;

	phys_dest =
		par_proc_table[my_par_pid]->inst_table[procnum]
		.port_table[out_port].phys_partner_node;

	if (phys_dest != BROADCAST)
		win = &tx_tmp_ring[phys_dest];
	else
		win = &tx_tmp_ring[phys_whoami];

#ifdef MORE_INSTANCES_ON_SAME_NODE
	if (
		par_proc_table[my_par_pid]->inst_table[procnum]
		.port_table[out_port].partner_is_same_node
	)
		goto same_node;
	if (
		par_proc_table[my_par_pid]->inst_table[procnum]
		.port_table[out_port].bcast_was_susp
	)
		goto same_node_bcast_susp;
#endif

DEBUG_E(printg("sys_gamma_send_flowctl(): Entering\n"));

	if ((phys_dest == BROADCAST) && (!was_susp)) {
/*
 *  Take note of addr and size of message, for resumption after possible
 *  suspension (shortage of credit).
 */
		par_proc_table[my_par_pid]->inst_table[procnum]
		.port_table[out_port].current_tx_data1 = data;
		par_proc_table[my_par_pid]->inst_table[procnum]
		.port_table[out_port].current_tx_len1 = len;
	}

	read_credit(phys_dest,credit);
	curr_credit = credit;

	if (curr_credit) {

ok_credit:

	h =	par_proc_table[my_par_pid]->inst_table[procnum]
		.port_table[out_port].header;
	entry = win->tail;
	incr_modulo(win->tail,GAMMA_TX_WIN_SIZE);
	memcpy(win->buf[entry].b,h,HEADER_SIZE);
	h = (l2s *)(win->buf[entry].b);

	if (!len) {

/*  Fast path for zero-byte messages.  */
	if (!is_low_watermark(phys_dest,credit,curr_credit)) {
		fill_send_header_cnklen(
			h, MSG_1ST_FRAME_MASK | MSG_COMPLETE_MASK
		);
	}
	else {
		fill_send_header_cnklen(
			h,
			MSG_1ST_FRAME_MASK | MSG_COMPLETE_MASK | CREQ_MASK
		);

DEBUG_F({
 unsigned long i;
 printg("SEND frame_no %d: curr credit %ld: ",win->tx_seq,curr_credit);
 for (i=0;i<total_num_nodes;i++)
  printg("%ld ",local_credit[i]);
 printg("\n");
})

	}

	if (!tx_ring_is_full()) {

room_in_tx_ring1:

#ifdef SEQCHECK
	set_send_header_frameno(h,win->tx_seq++);
#endif
#ifdef COLL_SAFE_TX
ro1:
#endif

DEBUG(printg(
"sending through port %ld, frame_no %d\n",out_port,read_send_header_frameno(h)
));

	tx_enq_pkt1(h,HEADER_SIZE);
	tx_start();
	tx_forw_tail();
#ifdef COLL_SAFE_TX
	tx_flush_ring_1(tx_ring_tail,MAX_FLUSH_WAIT*10,i);
	if ((!i) && (--trials >= 0)) {
DEBUG_L(printg("retrying, frame_no %d\n",read_send_header_frameno(h)));
		goto ro1;
	}
#endif
#ifdef SAFE_TX
	if (tx_ring_occupancy >= STEPS_THEN_STOP)
		tx_flush_ring(tx_ring_tail,MAX_FLUSH_WAIT,i);
#endif
	win->buf[entry].l = HEADER_SIZE;
	curr_credit--;
	goto end_tx;

	}

	tx_wait_ring_room(MAX_FLUSH_WAIT,i);
	if (!i)
		goto tx_failure;
	goto room_in_tx_ring1;

	}

	frame_size = tx_optimal_frag(len);

	cnk_len = gmin(len,frame_size-HEADER_SIZE);

	if (!was_susp) {
		if (cnk_len < len)
			fill_send_header_cnklen(
				h, cnk_len | MSG_1ST_FRAME_MASK
			);
		else
			fill_send_header_cnklen(
				h,
				cnk_len | MSG_1ST_FRAME_MASK
				| MSG_COMPLETE_MASK
			);
	}
	else {
		if (cnk_len < len)
			fill_send_header_cnklen(h, cnk_len);
		else {
			fill_send_header_cnklen(
				h, cnk_len | MSG_COMPLETE_MASK
			);
		}
	}

#ifdef MEASURE_OUT_TH_MEM
gamma_time(ima);
#endif
	copy_from_user(((unsigned char *)h)+HEADER_SIZE,(void *)data,cnk_len);
#ifdef MEASURE_OUT_TH_MEM
gamma_time(imb);
out_counter_mem++;
out_total_amount_mem_delay += (unsigned long)gamma_time_diff_ticks(imb,ima);
out_total_amount_mem_bytes += cnk_len;
#endif

#ifdef ALLOW_IRQS
	gamma_time(t_start);
#endif

	for (;;) {

		if (is_low_watermark(phys_dest,credit,curr_credit)) {
			set_send_header_creq(h);

DEBUG_F({
 unsigned long i;
 printg("SEND frame_no %d with CREQ: curr credit %ld: ",
  win->tx_seq,curr_credit);
 for (i=0;i<total_num_nodes;i++)
  printg("%ld ",local_credit[i]);
 printg("\n");
})

		}

		if (!tx_ring_is_full()) {

room_in_tx_ring:

#ifdef SEQCHECK
		set_send_header_frameno(h,win->tx_seq++);
#endif
#ifdef COLL_SAFE_TX
ro:
#endif

DEBUG(printg(
"sending through port %ld, frame_no %d\n",out_port,read_send_header_frameno(h)
));

		tx_enq_pkt1(h,HEADER_SIZE+cnk_len);
		tx_start();
		tx_forw_tail();
#ifdef COLL_SAFE_TX
		tx_flush_ring_1(tx_ring_tail,MAX_FLUSH_WAIT*10,i);
		if ((!i) && (--trials >= 0)) {
DEBUG_L(printg("retrying, frame_no %d\n",read_send_header_frameno(h)));
			goto ro;
		}
#endif
#ifdef SAFE_TX
		if (tx_ring_occupancy >= STEPS_THEN_STOP)
			tx_flush_ring(tx_ring_tail,MAX_FLUSH_WAIT,i);
#endif
		win->buf[entry].l = HEADER_SIZE+cnk_len;

		curr_credit--;
		len -= cnk_len;
		if (!len)
			break;

		if (!curr_credit)
			goto credit_exhausted;

credit_back:

#ifdef ALLOW_IRQS
		gamma_time(t_elaps);
		IRQ_enable_for_5usec(t_elaps,t_start);
#endif

/*  Prepare next frame  */
		entry = win->tail;
		incr_modulo(win->tail,GAMMA_TX_WIN_SIZE);
		memcpy(win->buf[entry].b,h,HEADER_SIZE);
		h = (l2s *)(win->buf[entry].b);
		data = (void *)((char *)data+cnk_len);
		cnk_len = gmin(len,frame_size - HEADER_SIZE);
		if (cnk_len < len)
			fill_send_header_cnklen(h,cnk_len);
		else {
			fill_send_header_cnklen(h,cnk_len | MSG_COMPLETE_MASK);
		}
#ifdef MEASURE_OUT_TH_MEM
gamma_time(ima);
#endif
		copy_from_user(
			((unsigned char *)h)+HEADER_SIZE,
			(void *)data,
			cnk_len
		);
#ifdef MEASURE_OUT_TH_MEM
gamma_time(imb);
out_counter_mem++;
out_total_amount_mem_delay += (unsigned long)gamma_time_diff_ticks(imb,ima);
out_total_amount_mem_bytes += cnk_len;
#endif

		continue;

credit_exhausted:

		dec_credit(phys_dest,credit);

		wait_credit_refill(i,phys_dest,my_par_pid,procnum);

		if (unlikely(par_proc_table[my_par_pid] == NULL)) {
			/*  VM_ABORT or VM_DESTROY pkt catched  */

DEBUG_L(printg(
"Virtual GAMMA %lu aborted/destroyed.  quitting send()...\n",my_par_pid
));

			tx_close();
#ifdef MORE_INSTANCES_ON_SAME_NODE
			spin_unlock(&gamma_tx_lock);
#endif
			return -GAMMA_ERR_GENERIC;
		}

		if (i)
			goto credit_back;

DEBUG_F(printg(
"sys_gamma_send_flowctl():  No credit.  Suspend (residual len %ld)\n",len
));

		tx_close();
#ifdef MORE_INSTANCES_ON_SAME_NODE
		spin_unlock(&gamma_tx_lock);
#endif
		if (!len)
			return -GAMMA_ERR_TRYAGAIN;
		return len;

		}

		tx_wait_ring_room(MAX_FLUSH_WAIT,i);
		if (!i)
			goto tx_failure;
		goto room_in_tx_ring;

	}

end_tx:

	tx_close();

	dec_credit(phys_dest,credit-curr_credit);

DEBUG(printg("sys_gamma_send_flowctl(): Done!\n"));

#ifdef MORE_INSTANCES_ON_SAME_NODE
	if (phys_dest != BROADCAST) {
		spin_unlock(&gamma_tx_lock);
		return 0;
	}

	res =	gamma_send_2p_flowctl_samenode(
			my_par_pid,procnum,out_port,(void *)NULL,0,
			par_proc_table[my_par_pid]->inst_table[procnum]
			.port_table[out_port].current_tx_data1,
			par_proc_table[my_par_pid]->inst_table[procnum]
			.port_table[out_port].current_tx_len1,
			1,0
		);
	spin_unlock(&gamma_tx_lock);
	return res;
#else
	return 0;
#endif

	}

	wait_credit_refill(i,phys_dest,my_par_pid,procnum);

	if (unlikely(par_proc_table[my_par_pid] == NULL)) {
		/*  VM_ABORT or VM_DESTROY pkt catched  */

DEBUG_L(printg(
"Virtual GAMMA %lu aborted/destroyed.  quitting send()...\n",my_par_pid
));

		tx_close();
#ifdef MORE_INSTANCES_ON_SAME_NODE
		spin_unlock(&gamma_tx_lock);
#endif
		return -GAMMA_ERR_GENERIC;
	}

	if (i)
		goto ok_credit;

DEBUG_E(printg("sys_gamma_send_flowctl():  No credit.  Try again!\n"));

#ifdef MORE_INSTANCES_ON_SAME_NODE
	spin_unlock(&gamma_tx_lock);
#endif
	return -GAMMA_ERR_TRYAGAIN;

tx_failure:

DEBUG_L(printg(
"sys_gamma_send_flowctl(): par_pid %ld, out_port %ld, frame_no %d: Failure (ring pos %d)\n",
my_par_pid,out_port,rec_header_frame_no(h),tx_ring_head
));

	tx_reset_and_restart();
	dec_credit(phys_dest,credit-curr_credit);

#ifdef MORE_INSTANCES_ON_SAME_NODE
	spin_unlock(&gamma_tx_lock);
#endif
	return -GAMMA_ERR_SENDFAILURE;

#ifdef MORE_INSTANCES_ON_SAME_NODE
same_node:

	if (phys_dest != BROADCAST) {
		res =	gamma_send_2p_flowctl_samenode(
				my_par_pid,procnum,out_port,
				(void *)NULL,0,data,len,
				0,was_susp
			);
		spin_unlock(&gamma_tx_lock);
		return res;
	}

same_node_bcast_susp:

	res =	gamma_send_2p_flowctl_samenode(
			my_par_pid,procnum,out_port,(void *)NULL,0,data,len,
			1,was_susp
		);
	spin_unlock(&gamma_tx_lock);
	return res;

dont_reenter:

	return -GAMMA_ERR_TRYAGAIN;
#endif

}


int sys_gamma_send_2p
(
	unsigned long out_port_and_len1,
	unsigned long my_par_pid_and_procnum,
	void *data1,
	void *data2,
	unsigned long len2
)
{

	unsigned long my_par_pid;
	unsigned long procnum;
	unsigned long out_port;
	unsigned long len1;
	unsigned long phys_dest;
	tx_tmp_ring_type *win;
#ifdef MORE_INSTANCES_ON_SAME_NODE
	void *ddata2;
	unsigned long llen1;
#endif
	l2s *h;
	unsigned long entry;
	unsigned long len;
	unsigned long frame_size;
	unsigned long cnk_len;
	tx_common_decl_and_init;
#ifdef ALLOW_IRQS
	time_586 t_start, t_elaps;
	unsigned long flags;
#endif
#ifdef COLL_SAFE_TX
	int trials = MAX_TRIALS;
#endif
#ifdef MORE_INSTANCES_ON_SAME_NODE
	int res;
#endif


#ifdef MORE_INSTANCES_ON_SAME_NODE
	if (!spin_trylock(&gamma_tx_lock))
		goto dont_reenter;
#endif

	my_par_pid = (unsigned short)my_par_pid_and_procnum;
	procnum = my_par_pid_and_procnum >> 16;
	out_port = out_port_and_len1>>8;
	len1 = (unsigned char)out_port_and_len1;
	phys_dest =
		par_proc_table[my_par_pid]->inst_table[procnum]
		.port_table[out_port].phys_partner_node;

	if (phys_dest != BROADCAST)
		win = &tx_tmp_ring[phys_dest];
	else
		win = &tx_tmp_ring[phys_whoami];

#ifdef MORE_INSTANCES_ON_SAME_NODE
	if (
		par_proc_table[my_par_pid]->inst_table[procnum]
		.port_table[out_port].partner_is_same_node
	)
		goto same_node;
#endif

DEBUG_E(printg("sys_gamma_send_2p(): Entering\n"));

#ifdef MORE_INSTANCES_ON_SAME_NODE
	ddata2 = data2;
	llen1 = len1;
#endif

	h =	par_proc_table[my_par_pid]->inst_table[procnum]
		.port_table[out_port].header;
	entry = win->tail;
	incr_modulo(win->tail,GAMMA_TX_WIN_SIZE);
	memcpy(win->buf[entry].b,h,HEADER_SIZE);
	h = (l2s *)(win->buf[entry].b);

	if (len1 <= MAX_1ST_PIECE_SIZE) {

	copy_from_user(&(((unsigned char *)h)[HEADER_SIZE]),data1,len1);

	if (!len2) {

/*  Fast path for zero-byte messages.  */
	fill_send_header_cnklen(
		h, len1 | MSG_1ST_FRAME_MASK | MSG_COMPLETE_MASK
	);

	if (!tx_ring_is_full()) {

room_in_tx_ring1:

#ifdef SEQCHECK
	set_send_header_frameno(h,win->tx_seq++);
#endif
#ifdef COLL_SAFE_TX
ro1:
#endif

DEBUG(printg(
"sending through port %ld, frame_no %d\n",out_port,read_send_header_frameno(h)
));

	tx_enq_pkt1(h,HEADER_SIZE+len1);
	tx_start();
	tx_forw_tail();
#ifdef COLL_SAFE_TX
	tx_flush_ring_1(tx_ring_tail,MAX_FLUSH_WAIT*10,i);
	if ((!i) && (--trials >= 0)) {
DEBUG_L(printg("retrying, frame_no %d\n",read_send_header_frameno(h)));
		goto ro1;
	}
#endif
#ifdef SAFE_TX
	if (tx_ring_occupancy >= STEPS_THEN_STOP)
		tx_flush_ring(tx_ring_tail,MAX_FLUSH_WAIT,i);
#endif
	win->buf[entry].l = HEADER_SIZE+len1;
	goto end_tx;

	}

	tx_wait_ring_room(MAX_FLUSH_WAIT,i);
	if (!i)
		goto tx_failure;
	goto room_in_tx_ring1;

	}	

	len = len1 + len2;
	frame_size = tx_optimal_frag(len);

	cnk_len = gmin(len,frame_size-HEADER_SIZE);

	if (cnk_len < len)
		fill_send_header_cnklen(h, cnk_len | MSG_1ST_FRAME_MASK);
	else
		fill_send_header_cnklen(
			h, cnk_len | MSG_1ST_FRAME_MASK | MSG_COMPLETE_MASK
		);

	copy_from_user(
		((unsigned char *)h)+HEADER_SIZE+len1,
		(void *)data2,
		cnk_len-len1
	);

#ifdef ALLOW_IRQS
	gamma_time(t_start);
#endif

	for (;;) {

		if (!tx_ring_is_full()) {

room_in_tx_ring:

#ifdef SEQCHECK
		set_send_header_frameno(h,win->tx_seq++);
#endif
#ifdef COLL_SAFE_TX
ro:
#endif

DEBUG(printg(
"sending through port %ld, frame_no %d\n",out_port,read_send_header_frameno(h)
));

		tx_enq_pkt1(h,HEADER_SIZE+cnk_len);
		tx_start();
		tx_forw_tail();
#ifdef COLL_SAFE_TX
		tx_flush_ring_1(tx_ring_tail,MAX_FLUSH_WAIT*10,i);
		if ((!i) && (--trials >= 0)) {
DEBUG_L(printg("retrying, frame_no %d\n",read_send_header_frameno(h)));
			goto ro;
		}
#endif
#ifdef SAFE_TX
		if (tx_ring_occupancy >= STEPS_THEN_STOP)
			tx_flush_ring(tx_ring_tail,MAX_FLUSH_WAIT,i);
#endif
		win->buf[entry].l = HEADER_SIZE+cnk_len;

		len -= cnk_len;
		if (!len)
			break;

#ifdef ALLOW_IRQS
		gamma_time(t_elaps);
		IRQ_enable_for_5usec(t_elaps,t_start);
#endif

/*  Prepare next frame  */
		entry = win->tail;
		incr_modulo(win->tail,GAMMA_TX_WIN_SIZE);
		memcpy(win->buf[entry].b,h,HEADER_SIZE);
		h = (l2s *)(win->buf[entry].b);
		data2 = (void *)((char *)data2+cnk_len-len1);
		len1 = 0;
		cnk_len = gmin(len,frame_size - HEADER_SIZE);
		if (cnk_len < len)
			fill_send_header_cnklen(h,cnk_len);
		else
			fill_send_header_cnklen(h,cnk_len | MSG_COMPLETE_MASK);

		copy_from_user(
			((unsigned char *)h)+HEADER_SIZE,
			(void *)data2,
			cnk_len
		);
		continue;

		}

		tx_wait_ring_room(MAX_FLUSH_WAIT,i);
		if (!i)
			goto tx_failure;
		goto room_in_tx_ring;

	}

end_tx:

	tx_close();

DEBUG(printg("sys_gamma_send_2p(): Done!\n"));

#ifdef MORE_INSTANCES_ON_SAME_NODE
	if (phys_dest != BROADCAST) {
		spin_unlock(&gamma_tx_lock);
		return 0;
	}
	res =	gamma_send_2p_samenode(
			my_par_pid,procnum,out_port,data1,llen1,ddata2,len2,1
		);
	spin_unlock(&gamma_tx_lock);
	return res;
#else
	return 0;
#endif

tx_failure:

DEBUG_L(printg(
"sys_gamma_send_2p(): par_pid %ld, out_port %ld, frame_no %d: Failure (ring pos %d)\n",
my_par_pid,out_port,rec_header_frame_no(h),tx_ring_head
));

	tx_reset_and_restart();
#ifdef MORE_INSTANCES_ON_SAME_NODE
	spin_unlock(&gamma_tx_lock);
#endif
	return -GAMMA_ERR_SENDFAILURE;

	}

#ifdef MORE_INSTANCES_ON_SAME_NODE
	spin_unlock(&gamma_tx_lock);
#endif
	return -GAMMA_ERR_GENERIC;

#ifdef MORE_INSTANCES_ON_SAME_NODE
same_node:

	if (phys_dest != BROADCAST) {
		res =	gamma_send_2p_samenode(
				my_par_pid,procnum,out_port,
				data1,len1,data2,len2,
				0
			);
		spin_unlock(&gamma_tx_lock);
		return res;
	}
	res =	gamma_send_2p_samenode(
			my_par_pid,procnum,out_port,data1,len1,data2,len2,
			1
		);
	spin_unlock(&gamma_tx_lock);
	return res;

dont_reenter:

	return -GAMMA_ERR_TRYAGAIN;
#endif

}


int sys_gamma_send_2p_flowctl
(
	unsigned long out_port_and_len1_and_was_susp,
	unsigned long my_par_pid_and_procnum,
	void *data2,
	unsigned long len2,
	void *data1
)
{

	unsigned long my_par_pid;
	unsigned long procnum;
	unsigned long out_port;
	unsigned long len1;
	unsigned long was_susp;
	unsigned long phys_dest;
	tx_tmp_ring_type *win;
	unsigned long credit;
	unsigned long curr_credit;
	l2s *h;
	unsigned long entry;
	unsigned long len;
	unsigned long cnk_len;
	unsigned long frame_size;
	tx_common_decl_and_init;
#ifdef ALLOW_IRQS
	time_586 t_start, t_elaps;
	unsigned long flags;
#endif
#ifdef COLL_SAFE_TX
	int trials = MAX_TRIALS;
#endif
#ifdef MORE_INSTANCES_ON_SAME_NODE
	int res;
#endif


#ifdef MORE_INSTANCES_ON_SAME_NODE
	if (!spin_trylock(&gamma_tx_lock))
		goto dont_reenter;
#endif

	my_par_pid = (unsigned short)my_par_pid_and_procnum;
	procnum = my_par_pid_and_procnum >> 16;
	out_port = (out_port_and_len1_and_was_susp & ~0x80000000) >> 8;
	len1 = (unsigned char)out_port_and_len1_and_was_susp;
	was_susp = out_port_and_len1_and_was_susp & 0x80000000;

	phys_dest =
		par_proc_table[my_par_pid]->inst_table[procnum]
		.port_table[out_port].phys_partner_node;

	if (phys_dest != BROADCAST)
		win = &tx_tmp_ring[phys_dest];
	else
		win = &tx_tmp_ring[phys_whoami];

#ifdef MORE_INSTANCES_ON_SAME_NODE
	if (
		par_proc_table[my_par_pid]->inst_table[procnum]
		.port_table[out_port].partner_is_same_node
	)
		goto same_node;
	if (
		par_proc_table[my_par_pid]->inst_table[procnum]
		.port_table[out_port].bcast_was_susp
	)
		goto same_node_bcast_susp;
#endif

DEBUG_E(printg("sys_gamma_send_2p_flowctl(): Entering\n"));

	if ((phys_dest == BROADCAST) && (!was_susp)) {
/*
 *  Take note of addr and size of message, for resumption after possible
 *  suspension (shortage of credit).
 */
		par_proc_table[my_par_pid]->inst_table[procnum]
		.port_table[out_port].current_tx_data1 = data1;
		par_proc_table[my_par_pid]->inst_table[procnum]
		.port_table[out_port].current_tx_len1 = len1;
		par_proc_table[my_par_pid]->inst_table[procnum]
		.port_table[out_port].current_tx_data2 = data2;
		par_proc_table[my_par_pid]->inst_table[procnum]
		.port_table[out_port].current_tx_len2 = len2;
	}

	read_credit(phys_dest,credit);
	curr_credit = credit;

	if (len1 <= MAX_1ST_PIECE_SIZE) {

	if (curr_credit) {

ok_credit:

	h =	par_proc_table[my_par_pid]->inst_table[procnum]
		.port_table[out_port].header;
	entry = win->tail;
	incr_modulo(win->tail,GAMMA_TX_WIN_SIZE);
	memcpy(win->buf[entry].b,h,HEADER_SIZE);
	h = (l2s *)(win->buf[entry].b);

	copy_from_user(&(((unsigned char *)h)[HEADER_SIZE]),data1,len1);

	if (!len2) {

/*  Fast path for zero-byte messages.  */
	if (!is_low_watermark(phys_dest,credit,curr_credit)) {
		fill_send_header_cnklen(
			h,
			len1 | MSG_1ST_FRAME_MASK | MSG_COMPLETE_MASK
		);
	}
	else {
		fill_send_header_cnklen(
			h,
			len1 | MSG_1ST_FRAME_MASK | MSG_COMPLETE_MASK
			| CREQ_MASK
		);

DEBUG_F({
 unsigned long i;
 printg("SEND frame_no %d: curr credit %ld: ",win->tx_seq,curr_credit);
 for (i=0;i<total_num_nodes;i++)
  printg("%ld ",local_credit[i]);
 printg("\n");
})

	}

	if (!tx_ring_is_full()) {

room_in_tx_ring1:

#ifdef SEQCHECK
	set_send_header_frameno(h,win->tx_seq++);
#endif
#ifdef COLL_SAFE_TX
ro1:
#endif

DEBUG(printg(
"sending through port %ld, frame_no %d\n",out_port,read_send_header_frameno(h)
));

	tx_enq_pkt1(h,HEADER_SIZE+len1);
	tx_start();
	tx_forw_tail();
#ifdef COLL_SAFE_TX
	tx_flush_ring_1(tx_ring_tail,MAX_FLUSH_WAIT*10,i);
	if ((!i) && (--trials >= 0)) {
DEBUG_L(printg("retrying, frame_no %d\n",read_send_header_frameno(h)));
		goto ro1;
	}
#endif
#ifdef SAFE_TX
	if (tx_ring_occupancy >= STEPS_THEN_STOP)
		tx_flush_ring(tx_ring_tail,MAX_FLUSH_WAIT,i);
#endif
	win->buf[entry].l = HEADER_SIZE+len1;
	curr_credit--;
	goto end_tx;

	}

	tx_wait_ring_room(MAX_FLUSH_WAIT,i);
	if (!i)
		goto tx_failure;
	goto room_in_tx_ring1;

	}

	len = len1 + len2;
	frame_size = tx_optimal_frag(len);

	cnk_len = gmin(len,frame_size-HEADER_SIZE);

	if (!was_susp) {
		if (cnk_len < len)
			fill_send_header_cnklen(
				h, cnk_len | MSG_1ST_FRAME_MASK
			);
		else
			fill_send_header_cnklen(
				h,
				cnk_len | MSG_1ST_FRAME_MASK
				| MSG_COMPLETE_MASK
			);
	}
	else {
		if (cnk_len < len)
			fill_send_header_cnklen(h, cnk_len);
		else {
			fill_send_header_cnklen(
				h, cnk_len | MSG_COMPLETE_MASK
			);
		}
	}

	copy_from_user(
		((unsigned char *)h)+HEADER_SIZE+len1,
		(void *)data2,
		cnk_len-len1
	);

#ifdef ALLOW_IRQS
	gamma_time(t_start);
#endif

	for (;;) {

		if (is_low_watermark(phys_dest,credit,curr_credit)) {
			set_send_header_creq(h);

DEBUG_F({
 unsigned long i;
 printg("SEND frame_no %d: curr credit %ld: ",win->tx_seq,curr_credit);
 for (i=0;i<total_num_nodes;i++)
  printg("%ld ",local_credit[i]);
 printg("\n");
})

		}

		if (!tx_ring_is_full()) {

room_in_tx_ring:

#ifdef SEQCHECK
		set_send_header_frameno(h,win->tx_seq++);
#endif
#ifdef COLL_SAFE_TX
ro:
#endif

DEBUG(printg(
"sending through port %ld, frame_no %d\n",out_port,read_send_header_frameno(h)
));

		tx_enq_pkt1(h,HEADER_SIZE+cnk_len);
		tx_start();
		tx_forw_tail();
#ifdef COLL_SAFE_TX
		tx_flush_ring_1(tx_ring_tail,MAX_FLUSH_WAIT*10,i);
		if ((!i) && (--trials >= 0)) {
DEBUG_L(printg("retrying, frame_no %d\n",read_send_header_frameno(h)));
			goto ro;
		}
#endif
#ifdef SAFE_TX
		if (tx_ring_occupancy >= STEPS_THEN_STOP)
			tx_flush_ring(tx_ring_tail,MAX_FLUSH_WAIT,i);
#endif
		win->buf[entry].l = HEADER_SIZE+cnk_len;

		curr_credit--;
		len -= cnk_len;
		if (!len)
			break;

		if (!curr_credit)
			goto credit_exhausted;
credit_back:

#ifdef ALLOW_IRQS
		gamma_time(t_elaps);
		IRQ_enable_for_5usec(t_elaps,t_start);
#endif

/*  Prepare next frame  */
		entry = win->tail;
		incr_modulo(win->tail,GAMMA_TX_WIN_SIZE);
		memcpy(win->buf[entry].b,h,HEADER_SIZE);
		h = (l2s *)(win->buf[entry].b);
		data2 = (void *)((char *)data2+cnk_len-len1);
		len1 = 0;
		cnk_len = gmin(len,frame_size - HEADER_SIZE);
		if (cnk_len < len)
			fill_send_header_cnklen(h,cnk_len);
		else {
			fill_send_header_cnklen(h,cnk_len | MSG_COMPLETE_MASK);
		}
		copy_from_user(
			((unsigned char *)h)+HEADER_SIZE,
			(void *)data2,
			cnk_len
		);

		continue;

credit_exhausted:

		dec_credit(phys_dest,credit);

		wait_credit_refill(i,phys_dest,my_par_pid,procnum);

		if (unlikely(par_proc_table[my_par_pid] == NULL)) {
			/*  VM_ABORT or VM_DESTROY pkt catched  */

DEBUG_L(printg(
"Virtual GAMMA %lu aborted/destroyed.  quitting send()...\n",my_par_pid
));

			tx_close();
#ifdef MORE_INSTANCES_ON_SAME_NODE
			spin_unlock(&gamma_tx_lock);
#endif
			return -GAMMA_ERR_GENERIC;
		}

		if (i)
			goto credit_back;

DEBUG_F(printg(
"sys_gamma_send_2p_flowctl():  No credit.  Suspend (residual len %ld)\n",len
));

		tx_close();
#ifdef MORE_INSTANCES_ON_SAME_NODE
		spin_unlock(&gamma_tx_lock);
#endif
		if (!len)
			return -GAMMA_ERR_TRYAGAIN;
		return len;

		}

		tx_wait_ring_room(MAX_FLUSH_WAIT,i);
		if (!i)
			goto tx_failure;
		goto room_in_tx_ring;

	}

end_tx:

	tx_close();

	dec_credit(phys_dest,credit-curr_credit);

DEBUG(printg("sys_gamma_send_2p_flowctl(): Done!\n"));

#ifdef MORE_INSTANCES_ON_SAME_NODE
	if (phys_dest != BROADCAST) {
		spin_unlock(&gamma_tx_lock);
		return 0;
	}
	res =	gamma_send_2p_flowctl_samenode(
			my_par_pid,procnum,out_port,
			par_proc_table[my_par_pid]->inst_table[procnum]
			.port_table[out_port].current_tx_data1,
			par_proc_table[my_par_pid]->inst_table[procnum]
			.port_table[out_port].current_tx_len1,
			par_proc_table[my_par_pid]->inst_table[procnum]
			.port_table[out_port].current_tx_data2,
			par_proc_table[my_par_pid]->inst_table[procnum]
			.port_table[out_port].current_tx_len2,
			1,0
		);
	spin_unlock(&gamma_tx_lock);
	return res;
#else
	return 0;
#endif

	}

	wait_credit_refill(i,phys_dest,my_par_pid,procnum);

	if (unlikely(par_proc_table[my_par_pid] == NULL)) {
		/*  VM_ABORT or VM_DESTROY pkt catched  */

DEBUG_L(printg(
"Virtual GAMMA %lu aborted/destroyed.  quitting send()...\n",my_par_pid
));

		tx_close();
#ifdef MORE_INSTANCES_ON_SAME_NODE
		spin_unlock(&gamma_tx_lock);
#endif
		return -GAMMA_ERR_GENERIC;
	}

	if (i)
		goto ok_credit;

DEBUG_E(printg("sys_gamma_send_2p_flowctl():  No credit.  Try again!\n"));

#ifdef MORE_INSTANCES_ON_SAME_NODE
	spin_unlock(&gamma_tx_lock);
#endif
	return -GAMMA_ERR_TRYAGAIN;

tx_failure:

DEBUG_L(printg(
"sys_gamma_send_2p_flowctl(): par_pid %ld, out_port %ld, frame_no %d: Failure (ring pos %d)\n",
my_par_pid,out_port,rec_header_frame_no(h),tx_ring_head
));

	tx_reset_and_restart();
	dec_credit(phys_dest,credit-curr_credit);

#ifdef MORE_INSTANCES_ON_SAME_NODE
	spin_unlock(&gamma_tx_lock);
#endif
	return -GAMMA_ERR_SENDFAILURE;

	}

#ifdef MORE_INSTANCES_ON_SAME_NODE
	spin_unlock(&gamma_tx_lock);
#endif
	return -GAMMA_ERR_GENERIC;

#ifdef MORE_INSTANCES_ON_SAME_NODE
same_node:

	if (phys_dest != BROADCAST) {
		res = 	gamma_send_2p_flowctl_samenode(
				my_par_pid,procnum,out_port,
				data1,len1,data2,len2,
				0,was_susp
			);
		spin_unlock(&gamma_tx_lock);
		return res;
	}

same_node_bcast_susp:

	res =	gamma_send_2p_flowctl_samenode(
			my_par_pid,procnum,out_port,data1,len1,data2,len2,
			1,was_susp
		);
	spin_unlock(&gamma_tx_lock);
	return res;

dont_reenter:

	return -GAMMA_ERR_TRYAGAIN;
#endif

}


/************************************************************************
 *	Non-blocking send routines, best-effort				*
 ************************************************************************/


int sys_gamma_isend
(
	unsigned long out_port,
	unsigned long my_par_pid,
	unsigned long procnum,
	void *data,
	unsigned long len
)
{

	unsigned long phys_dest;
	tx_tmp_ring_type *win;
#ifdef MORE_INSTANCES_ON_SAME_NODE
	void *ddata;
	unsigned long llen;
#endif
	l2s *h;
	unsigned long entry;
	unsigned long frame_size;
	unsigned long cnk_len;
	unsigned long paddr;
	unsigned long next_paddr;
	tx_common_decl_and_init;
#ifdef ALLOW_IRQS
	time_586 t_start, t_elaps;
	unsigned long flags;
#endif
#ifdef COLL_SAFE_TX
	int trials = MAX_TRIALS;
#endif
#ifdef MORE_INSTANCES_ON_SAME_NODE
	int res;
#endif


#ifdef MORE_INSTANCES_ON_SAME_NODE
	if (!spin_trylock(&gamma_tx_lock))
		goto dont_reenter;
#endif

	phys_dest =
		par_proc_table[my_par_pid]->inst_table[procnum]
		.port_table[out_port].phys_partner_node;

	if (phys_dest != BROADCAST)
		win = &tx_tmp_ring[phys_dest];
	else
		win = &tx_tmp_ring[phys_whoami];

#ifdef MORE_INSTANCES_ON_SAME_NODE
	if (
		par_proc_table[my_par_pid]->inst_table[procnum]
		.port_table[out_port].partner_is_same_node
	)
		goto same_node;
#endif

DEBUG_E(printg("sys_gamma_isend(): Entering\n"));

#ifdef MORE_INSTANCES_ON_SAME_NODE
	ddata = data;
	llen = len;
#endif

	h =	par_proc_table[my_par_pid]->inst_table[procnum]
		.port_table[out_port].header;
	entry = win->tail;
	incr_modulo(win->tail,GAMMA_TX_WIN_SIZE);
	memcpy(win->buf[entry].b,h,HEADER_SIZE);
	h = (l2s *)(win->buf[entry].b);

	if (!len) {

/*  Fast path for zero-length messages  */
	fill_send_header_cnklen(
		h, MSG_1ST_FRAME_MASK | MSG_COMPLETE_MASK
	);

	if (!tx_ring_is_full()) {

room_in_tx_ring1:

#ifdef SEQCHECK
	set_send_header_frameno(h,win->tx_seq++);
#endif
#ifdef COLL_SAFE_TX
ro1:
#endif

DEBUG(printg(
"sending through port %ld, frame_no %d\n",out_port,read_send_header_frameno(h)
));

	tx_enq_pkt1(h,HEADER_SIZE);
	tx_start();
	tx_forw_tail();
#ifdef COLL_SAFE_TX
	tx_flush_ring_1(tx_ring_tail,MAX_FLUSH_WAIT*10,i);
	if ((!i) && (--trials >= 0)) {
DEBUG_L(printg("retrying, frame_no %d\n",read_send_header_frameno(h)));
		goto ro1;
	}
#endif
#ifdef SAFE_TX
	if (tx_ring_occupancy >= STEPS_THEN_STOP)
		tx_flush_ring(tx_ring_tail,MAX_FLUSH_WAIT,i);
#endif
	win->buf[entry].l = -(1 + tx_ring_tail);
	win->buf[entry].hlen = HEADER_SIZE;
	goto end_tx;

	}

	tx_wait_ring_room(MAX_FLUSH_WAIT,i);
	if (!i)
		goto tx_failure;
	goto room_in_tx_ring1;

	}

	frame_size = tx_optimal_frag(len);

	cnk_len = gmin(len,frame_size-HEADER_SIZE);

	paddr = user_space_to_phys(current,data);
	if (cnk_len >= (PAGE_SIZE-(paddr & 0x00000fff)))
		goto next_page1;
	next_paddr = paddr+cnk_len;

next_page_back1:

	if (cnk_len < len)
		fill_send_header_cnklen(h, cnk_len | MSG_1ST_FRAME_MASK);
	else
		fill_send_header_cnklen(
			h, cnk_len | MSG_1ST_FRAME_MASK | MSG_COMPLETE_MASK
		);

#ifdef ALLOW_IRQS
	gamma_time(t_start);
#endif

	for (;;) {

		if (!tx_ring_is_full()) {

room_in_tx_ring:

#ifdef SEQCHECK
		set_send_header_frameno(h,win->tx_seq++);
#endif
#ifdef COLL_SAFE_TX
ro:
#endif

DEBUG(printg(
"sending through port %ld, frame_no %d\n",out_port,read_send_header_frameno(h)
));

		tx_enq_pkt(h,HEADER_SIZE,(void *)paddr,cnk_len);
		tx_start();
		tx_forw_tail();
#ifdef COLL_SAFE_TX
		tx_flush_ring_1(tx_ring_tail,MAX_FLUSH_WAIT*10,i);
		if ((!i) && (--trials >= 0)) {
DEBUG_L(printg("retrying, frame_no %d\n",read_send_header_frameno(h)));
			goto ro;
		}
#endif
#ifdef SAFE_TX
		if (tx_ring_occupancy >= STEPS_THEN_STOP)
			tx_flush_ring(tx_ring_tail,MAX_FLUSH_WAIT,i);
#endif
		win->buf[entry].l = -(1 + tx_ring_tail);
		win->buf[entry].hlen = HEADER_SIZE;

		len -= cnk_len;
		if (!len)
			break;

#ifdef ALLOW_IRQS
		gamma_time(t_elaps);
		IRQ_enable_for_5usec(t_elaps,t_start);
#endif

/*  Prepare next frame  */
		entry = win->tail;
		incr_modulo(win->tail,GAMMA_TX_WIN_SIZE);
		memcpy(win->buf[entry].b,h,HEADER_SIZE);
		h = (l2s *)(win->buf[entry].b);
		data = (void *)((char *)data+cnk_len);
		cnk_len = gmin(len,frame_size - HEADER_SIZE);
		paddr = next_paddr;
		if (cnk_len >= (PAGE_SIZE-(paddr & 0x00000fff)))
			goto next_page;
		next_paddr += cnk_len;

next_page_back:

		if (cnk_len < len)
			fill_send_header_cnklen(h,cnk_len);
		else
			fill_send_header_cnklen(h,cnk_len | MSG_COMPLETE_MASK);

		continue;

next_page:

		cnk_len = (unsigned short)(PAGE_SIZE-(paddr&0x00000fff));
		next_paddr = user_space_to_phys(current,(char *)data+cnk_len);
		goto next_page_back;

		}

		tx_wait_ring_room(MAX_FLUSH_WAIT,i);
		if (!i)
	        	goto tx_failure;
		goto room_in_tx_ring;

	}

end_tx:

	tx_close();

DEBUG(printg("sys_gamma_isend(): Done!  isend_handle %d\n",tx_ring_tail));

#ifdef MORE_INSTANCES_ON_SAME_NODE
	if (phys_dest != BROADCAST) {
		spin_unlock(&gamma_tx_lock);
		return tx_ring_tail;
	}

	res = gamma_send_2p_samenode(
		my_par_pid,procnum,out_port,(void *)NULL,0,ddata,llen,1
	);

	spin_unlock(&gamma_tx_lock);

	if (!res)
		return tx_ring_tail;
	return res;
#else
	return tx_ring_tail;
#endif

next_page1:

	cnk_len = (unsigned short)(PAGE_SIZE-(paddr & 0x00000fff));
	next_paddr = user_space_to_phys(current,(char *)data+cnk_len);
	goto next_page_back1;

tx_failure:

DEBUG_L(printg(
"sys_gamma_isend(): par_pid %ld, out_port %ld, frame_no %d: Failure (ring pos %d)\n",
my_par_pid,out_port,rec_header_frame_no(h),tx_ring_head
));

	tx_reset_and_restart();
#ifdef MORE_INSTANCES_ON_SAME_NODE
	spin_unlock(&gamma_tx_lock);
#endif
	return -GAMMA_ERR_SENDFAILURE;

#ifdef MORE_INSTANCES_ON_SAME_NODE
same_node:

	if (phys_dest != BROADCAST) {
		res = gamma_send_2p_samenode(
			my_par_pid,procnum,out_port,(void *)NULL,0,data,len,
			0
		);
		spin_unlock(&gamma_tx_lock);
		if (!res)
			return -GAMMA_ISEND_IS_DONE;
		return res;
	}
	res = gamma_send_2p_samenode(
		my_par_pid,procnum,out_port,(void *)NULL,0,data,len,
		1
	);
	spin_unlock(&gamma_tx_lock);
	if (!res)
		return -GAMMA_ISEND_IS_DONE;
	return res;

dont_reenter:

	return -GAMMA_ERR_TRYAGAIN;
#endif

}


int sys_gamma_isend_flowctl
(
	unsigned long out_port_and_was_susp,
	unsigned long my_par_pid,
	unsigned long procnum,
	void *data,
	unsigned long len
)
{

	unsigned long out_port;
	unsigned long was_susp;
	unsigned long phys_dest;
	tx_tmp_ring_type *win;
	unsigned long credit;
	unsigned long curr_credit;
	l2s *h;
	unsigned long entry;
	unsigned long frame_size;
	unsigned long cnk_len;
	unsigned long paddr;
	unsigned long next_paddr;
	tx_common_decl_and_init;
#ifdef ALLOW_IRQS
	time_586 t_start, t_elaps;
	unsigned long flags;
#endif
#ifdef COLL_SAFE_TX
	int trials = MAX_TRIALS;
#endif
#ifdef MORE_INSTANCES_ON_SAME_NODE
	int res;
#endif


#ifdef MORE_INSTANCES_ON_SAME_NODE
	if (!spin_trylock(&gamma_tx_lock))
		goto dont_reenter;
#endif

	out_port = out_port_and_was_susp & ~0x80000000;
	was_susp = out_port_and_was_susp & 0x80000000;

	phys_dest =
		par_proc_table[my_par_pid]->inst_table[procnum]
		.port_table[out_port].phys_partner_node;

	if (phys_dest != BROADCAST)
		win = &tx_tmp_ring[phys_dest];
	else
		win = &tx_tmp_ring[phys_whoami];

#ifdef MORE_INSTANCES_ON_SAME_NODE
	if (
		par_proc_table[my_par_pid]->inst_table[procnum]
		.port_table[out_port].partner_is_same_node
	)
		goto same_node;
	if (
		par_proc_table[my_par_pid]->inst_table[procnum]
		.port_table[out_port].bcast_was_susp
	)
		goto same_node_bcast_susp;
#endif

DEBUG_E(printg("sys_gamma_isend_flowctl(): Entering\n"));

	if ((phys_dest == BROADCAST) && (!was_susp)) {
/*
 *  Take note of addr and size of message, for resumption after possible
 *  suspension (shortage of credit).
 */
		par_proc_table[my_par_pid]->inst_table[procnum]
		.port_table[out_port].current_tx_data1 = data;
		par_proc_table[my_par_pid]->inst_table[procnum]
		.port_table[out_port].current_tx_len1 = len;
	}

	read_credit(phys_dest,credit);
	curr_credit = credit;

	if (curr_credit) {

ok_credit:

	h =	par_proc_table[my_par_pid]->inst_table[procnum]
		.port_table[out_port].header;
	entry = win->tail;
	incr_modulo(win->tail,GAMMA_TX_WIN_SIZE);
	memcpy(win->buf[entry].b,h,HEADER_SIZE);
	h = (l2s *)(win->buf[entry].b);

	if (!len) {

	if (!is_low_watermark(phys_dest,credit,curr_credit)) {
		fill_send_header_cnklen(
			h, MSG_1ST_FRAME_MASK | MSG_COMPLETE_MASK
		);
	}
	else {
		fill_send_header_cnklen(
			h,
			MSG_1ST_FRAME_MASK | MSG_COMPLETE_MASK
			| CREQ_MASK
		);
	}

	if (!tx_ring_is_full()) {

room_in_tx_ring1:

#ifdef SEQCHECK
	set_send_header_frameno(h,win->tx_seq++);
#endif
#ifdef COLL_SAFE_TX
ro1:
#endif

DEBUG(printg(
"sending through port %ld, frame_no %d\n",out_port,read_send_header_frameno(h)
));

	tx_enq_pkt1(h,HEADER_SIZE);
	tx_start();
	tx_forw_tail();
#ifdef COLL_SAFE_TX
	tx_flush_ring_1(tx_ring_tail,MAX_FLUSH_WAIT*10,i);
	if ((!i) && (--trials >= 0)) {
DEBUG_L(printg("retrying, frame_no %d\n",read_send_header_frameno(h)));
		goto ro1;
	}
#endif
#ifdef SAFE_TX
	if (tx_ring_occupancy >= STEPS_THEN_STOP)
		tx_flush_ring(tx_ring_tail,MAX_FLUSH_WAIT,i);
#endif
	win->buf[entry].l = -(1 + tx_ring_tail);
	win->buf[entry].hlen = HEADER_SIZE;
	curr_credit--;
	goto end_tx;

	}

	tx_wait_ring_room(MAX_FLUSH_WAIT,i);
	if (!i)
		goto tx_failure;
	goto room_in_tx_ring1;

	}

	frame_size = tx_optimal_frag(len);

	cnk_len = gmin(len,frame_size-HEADER_SIZE);

	paddr = user_space_to_phys(current,data);
	if (cnk_len >= (PAGE_SIZE-(paddr & 0x00000fff)))
		goto next_page1;
	next_paddr = paddr+cnk_len;

next_page_back1:

	if (!was_susp) {
		if (cnk_len < len)
			fill_send_header_cnklen(
				h, cnk_len | MSG_1ST_FRAME_MASK
			);
		else
			fill_send_header_cnklen(
				h,
				cnk_len | MSG_1ST_FRAME_MASK
				| MSG_COMPLETE_MASK
			);
	}
	else {
		if (cnk_len < len)
			fill_send_header_cnklen(h, cnk_len);
		else {
			fill_send_header_cnklen(
				h, cnk_len | MSG_COMPLETE_MASK
			);
		}
	}

#ifdef ALLOW_IRQS
	gamma_time(t_start);
#endif

	for (;;) {

		if (is_low_watermark(phys_dest,credit,curr_credit))
			set_send_header_creq(h);

		if (!tx_ring_is_full()) {

room_in_tx_ring:

#ifdef SEQCHECK
		set_send_header_frameno(h,win->tx_seq++);
#endif
#ifdef COLL_SAFE_TX
ro:
#endif

DEBUG(printg(
"sending through port %ld, frame_no %d\n",out_port,read_send_header_frameno(h)
));

		tx_enq_pkt(h,HEADER_SIZE,(void *)paddr,cnk_len);
		tx_start();
		tx_forw_tail();
#ifdef COLL_SAFE_TX
		tx_flush_ring_1(tx_ring_tail,MAX_FLUSH_WAIT*10,i);
		if ((!i) && (--trials >= 0)) {
DEBUG_L(printg("retrying, frame_no %d\n",read_send_header_frameno(h)));
			goto ro;
		}
#endif
#ifdef SAFE_TX
		if (tx_ring_occupancy >= STEPS_THEN_STOP)
			tx_flush_ring(tx_ring_tail,MAX_FLUSH_WAIT,i);
#endif
		win->buf[entry].l = -(1 + tx_ring_tail);
		win->buf[entry].hlen = HEADER_SIZE;

		curr_credit--;
		len -= cnk_len;
		if (!len)
			break;

		if (!curr_credit)
			goto credit_exhausted;

credit_back:

#ifdef ALLOW_IRQS
		gamma_time(t_elaps);
		IRQ_enable_for_5usec(t_elaps,t_start);
#endif

/*  Prepare next frame  */
		entry = win->tail;
		incr_modulo(win->tail,GAMMA_TX_WIN_SIZE);
		memcpy(win->buf[entry].b,h,HEADER_SIZE);
		h = (l2s *)(win->buf[entry].b);
		data = (void *)((char *)data+cnk_len);
		cnk_len = gmin(len,frame_size - HEADER_SIZE);
		paddr = next_paddr;
		if (cnk_len >= (PAGE_SIZE-(paddr & 0x00000fff)))
			goto next_page;
		next_paddr += cnk_len;

next_page_back:

		if (cnk_len < len)
			fill_send_header_cnklen(h,cnk_len);
		else {
			fill_send_header_cnklen(h,cnk_len | MSG_COMPLETE_MASK);
		}

		continue;

next_page:

		cnk_len = (unsigned short)(PAGE_SIZE-(paddr&0x00000fff));
		next_paddr = user_space_to_phys(current,(char *)data+cnk_len);
		goto next_page_back;

credit_exhausted:

		dec_credit(phys_dest,credit);

		wait_credit_refill(i,phys_dest,my_par_pid,procnum);

		if (unlikely(par_proc_table[my_par_pid] == NULL)) {
			/*  VM_ABORT or VM_DESTROY pkt catched  */

DEBUG_L(printg(
"Virtual GAMMA %lu aborted/destroyed.  quitting send()...\n",my_par_pid
));

			tx_close();
#ifdef MORE_INSTANCES_ON_SAME_NODE
			spin_unlock(&gamma_tx_lock);
#endif
			return -GAMMA_ERR_GENERIC;
		}

		if (i)
			goto credit_back;

DEBUG_F(printg(
"sys_gamma_isend_flowctl():  No credit.  Suspend (residual len %ld)\n",len
));

		tx_close();
/*  Return the total lenght of the not yet sent part of msg  */
#ifdef MORE_INSTANCES_ON_SAME_NODE
		spin_unlock(&gamma_tx_lock);
#endif
		return -(GAMMA_ERR_MAX+len);

		}

		tx_wait_ring_room(MAX_FLUSH_WAIT,i);
		if (!i)
			goto tx_failure;
		goto room_in_tx_ring;

	}

end_tx:

	tx_close();

	dec_credit(phys_dest,credit-curr_credit);

DEBUG(printg(
"sys_gamma_isend_flowctl(): Done!  isend_handle %d\n",tx_ring_tail
));

#ifdef MORE_INSTANCES_ON_SAME_NODE
	if (phys_dest != BROADCAST) {
		spin_unlock(&gamma_tx_lock);
		return tx_ring_tail;
	}
/*
 *  Here I would like a non-blocking intra-node send(), but this would need
 *  additional hardware...
 */
	res = gamma_send_2p_flowctl_samenode(
		my_par_pid,procnum,out_port,(void *)NULL,0,
		par_proc_table[my_par_pid]->inst_table[procnum]
		.port_table[out_port].current_tx_data1,
		par_proc_table[my_par_pid]->inst_table[procnum]
		.port_table[out_port].current_tx_len1,
		1,0
	);
	if (!res) {
		spin_unlock(&gamma_tx_lock);
		return tx_ring_tail;
	}
/*
 *  Store the inter-node isend handle: will be returned to the caller as soon
 *  as the intra-node Tx terminates.
 */
	par_proc_table[my_par_pid]->inst_table[procnum]
	.port_table[out_port].suspended_isend_handle = tx_ring_tail;

	spin_unlock(&gamma_tx_lock);

	if (res == -GAMMA_ERR_TRYAGAIN)
/*  Return the total lenght of the not yet sent part of msg  */
		return -(GAMMA_ERR_MAX+len);
	if (res > 0)
/*  Return the total lenght of the not yet sent part of msg  */
		return -(GAMMA_ERR_MAX+res);
	return res;
#else
	return tx_ring_tail;
#endif

next_page1:

	cnk_len = (unsigned short)(PAGE_SIZE-(paddr & 0x00000fff));
	next_paddr = user_space_to_phys(current,(char *)data+cnk_len);
	goto next_page_back1;

	}

	wait_credit_refill(i,phys_dest,my_par_pid,procnum);

	if (unlikely(par_proc_table[my_par_pid] == NULL)) {
		/*  VM_ABORT or VM_DESTROY pkt catched  */

DEBUG_L(printg(
"Virtual GAMMA %lu aborted/destroyed.  quitting send()...\n",my_par_pid
));

		tx_close();
#ifdef MORE_INSTANCES_ON_SAME_NODE
		spin_unlock(&gamma_tx_lock);
#endif
		return -GAMMA_ERR_GENERIC;
	}

	if (i)
		goto ok_credit;

DEBUG_E(printg("sys_gamma_isend_flowctl():  No credit.  Try again!\n"));

/*  Return the total lenght of the not yet sent part of msg  */
#ifdef MORE_INSTANCES_ON_SAME_NODE
	spin_unlock(&gamma_tx_lock);
#endif
	return -(GAMMA_ERR_MAX+len);

tx_failure:

DEBUG_L(printg(
"sys_gamma_isend_flowctl(): par_pid %ld, out_port %ld, frame_no %d: Failure (ring pos %d)\n",
my_par_pid,out_port,rec_header_frame_no(h),tx_ring_head
));

	tx_reset_and_restart();
	dec_credit(phys_dest,credit-curr_credit);

#ifdef MORE_INSTANCES_ON_SAME_NODE
	spin_unlock(&gamma_tx_lock);
#endif
	return -GAMMA_ERR_SENDFAILURE;

#ifdef MORE_INSTANCES_ON_SAME_NODE
same_node:

	if (phys_dest != BROADCAST) {
/*
 *  Here I would like a non-blocking intra-node send(), but this would need
 *  additional hardware...
 */
		res = gamma_send_2p_flowctl_samenode(
			my_par_pid,procnum,out_port,(void *)NULL,0,data,len,
			0,was_susp
		);
		spin_unlock(&gamma_tx_lock);
		if (!res)
			return -GAMMA_ISEND_IS_DONE;
		if (res == -GAMMA_ERR_TRYAGAIN)
/*  Return the total lenght of the not yet sent part of msg  */
			return -(GAMMA_ERR_MAX+len);
		if (res > 0)
/*  Return the total lenght of the not yet sent part of msg  */
			return -(GAMMA_ERR_MAX+res);
		return res;
	}

same_node_bcast_susp:

/*
 *  Here I would like a non-blocking intra-node send(), but this would need
 *  additional hardware...
 */
	res = gamma_send_2p_flowctl_samenode(
		my_par_pid,procnum,out_port,(void *)NULL,0,data,len,
		1,was_susp
	);
	if (!res) {
/*
 *  The intra-node Tx has finished.
 *  Recover the inter-node isend handle, return it to the caller.
 */
		res =	par_proc_table[my_par_pid]->inst_table[procnum]
			.port_table[out_port].suspended_isend_handle;
		par_proc_table[my_par_pid]->inst_table[procnum]
		.port_table[out_port].suspended_isend_handle =
			-GAMMA_ISEND_IS_DONE;
		spin_unlock(&gamma_tx_lock);
		return res;
	}
	spin_unlock(&gamma_tx_lock);
	if (res == -GAMMA_ERR_TRYAGAIN)
/*  Return the total lenght of the not yet sent part of msg  */
		return -(GAMMA_ERR_MAX+len);
	if (res > 0)
/*  Return the total lenght of the not yet sent part of msg  */
		return -(GAMMA_ERR_MAX+res);
	return res;

dont_reenter:

	return -GAMMA_ERR_TRYAGAIN;
#endif

}


int sys_gamma_isend_2p
(
	unsigned long out_port_and_len1,
	unsigned long my_par_pid_and_procnum,
	void *data1,
	void *data2,
	unsigned long len2
)
{

	unsigned long my_par_pid;
	unsigned long procnum;
	unsigned long out_port;
	unsigned long len1;
	unsigned long phys_dest;
	tx_tmp_ring_type *win;
#ifdef MORE_INSTANCES_ON_SAME_NODE
	void *ddata2;
	unsigned long llen1;
#endif
	l2s *h;
	unsigned long entry;
	unsigned long len;
	unsigned long frame_size;
	unsigned long cnk_len;
	unsigned long paddr;
	unsigned long next_paddr;
	tx_common_decl_and_init;
#ifdef ALLOW_IRQS
	time_586 t_start, t_elaps;
	unsigned long flags;
#endif
#ifdef COLL_SAFE_TX
	int trials = MAX_TRIALS;
#endif
#ifdef MORE_INSTANCES_ON_SAME_NODE
	int res;
#endif


#ifdef MORE_INSTANCES_ON_SAME_NODE
	if (!spin_trylock(&gamma_tx_lock))
		goto dont_reenter;
#endif

	my_par_pid = (unsigned short)my_par_pid_and_procnum;
	procnum = my_par_pid_and_procnum >> 16;
	out_port = out_port_and_len1 >> 8;
	len1 = (unsigned char)out_port_and_len1;
	phys_dest =
		par_proc_table[my_par_pid]->inst_table[procnum]
		.port_table[out_port].phys_partner_node;

	if (phys_dest != BROADCAST)
		win = &tx_tmp_ring[phys_dest];
	else
		win = &tx_tmp_ring[phys_whoami];

#ifdef MORE_INSTANCES_ON_SAME_NODE
	if (
		par_proc_table[my_par_pid]->inst_table[procnum]
		.port_table[out_port].partner_is_same_node
	)
		goto same_node;
#endif

DEBUG_E(printg("sys_gamma_isend_2p(): Entering\n"));

#ifdef MORE_INSTANCES_ON_SAME_NODE
	ddata2 = data2;
	llen1 = len1;
#endif

	h =	par_proc_table[my_par_pid]->inst_table[procnum]
		.port_table[out_port].header;
	entry = win->tail;
	incr_modulo(win->tail,GAMMA_TX_WIN_SIZE);
	memcpy(win->buf[entry].b,h,HEADER_SIZE);
	h = (l2s *)(win->buf[entry].b);

	if (len1 <= MAX_1ST_PIECE_SIZE) {

	copy_from_user(&(((unsigned char *)h)[HEADER_SIZE]),data1,len1);

	if (!len2) {

	fill_send_header_cnklen(
		h, len1 | MSG_1ST_FRAME_MASK | MSG_COMPLETE_MASK
	);

	if (!tx_ring_is_full()) {

room_in_tx_ring1:

#ifdef SEQCHECK
	set_send_header_frameno(h,win->tx_seq++);
#endif
#ifdef COLL_SAFE_TX
ro1:
#endif

DEBUG(printg(
"sending through port %ld, frame_no %d\n",out_port,read_send_header_frameno(h)
));

	tx_enq_pkt1(h,HEADER_SIZE+len1);
	tx_start();
	tx_forw_tail();
#ifdef COLL_SAFE_TX
	tx_flush_ring_1(tx_ring_tail,MAX_FLUSH_WAIT*10,i);
	if ((!i) && (--trials >= 0)) {
DEBUG_L(printg("retrying, frame_no %d\n",read_send_header_frameno(h)));
		goto ro1;
	}
#endif
#ifdef SAFE_TX
	if (tx_ring_occupancy >= STEPS_THEN_STOP)
		tx_flush_ring(tx_ring_tail,MAX_FLUSH_WAIT,i);
#endif
	win->buf[entry].l = -(1 + tx_ring_tail);
	win->buf[entry].hlen = HEADER_SIZE+len1;
	goto end_tx;

	}

	tx_wait_ring_room(MAX_FLUSH_WAIT,i);
	if (!i)
		goto tx_failure;
	goto room_in_tx_ring1;

	}

	len = len1+len2;
	frame_size = tx_optimal_frag(len1+len2);

	cnk_len = gmin(len,frame_size-HEADER_SIZE);

	paddr = user_space_to_phys(current,data2);
	if (cnk_len >= len1+(PAGE_SIZE-(paddr & 0x00000fff)))
		goto next_page1;
	next_paddr = paddr+cnk_len-len1;

next_page_back1:

	if (cnk_len < len)
		fill_send_header_cnklen(h, cnk_len | MSG_1ST_FRAME_MASK);
	else
		fill_send_header_cnklen(
			h, cnk_len | MSG_1ST_FRAME_MASK | MSG_COMPLETE_MASK
		);

#ifdef ALLOW_IRQS
	gamma_time(t_start);
#endif

	for (;;) {

		if (!tx_ring_is_full()) {

room_in_tx_ring:

#ifdef SEQCHECK
		set_send_header_frameno(h,win->tx_seq++);
#endif
#ifdef COLL_SAFE_TX
ro:
#endif

DEBUG(printg(
"sending through port %ld, frame_no %d\n",out_port,read_send_header_frameno(h)
));

		tx_enq_pkt(h,HEADER_SIZE+len1,(void *)paddr,cnk_len-len1);
		tx_start();
		tx_forw_tail();
#ifdef COLL_SAFE_TX
		tx_flush_ring_1(tx_ring_tail,MAX_FLUSH_WAIT*10,i);
		if ((!i) && (--trials >= 0)) {
DEBUG_L(printg("retrying, frame_no %d\n",read_send_header_frameno(h)));
			goto ro;
		}
#endif
#ifdef SAFE_TX
		if (tx_ring_occupancy >= STEPS_THEN_STOP)
			tx_flush_ring(tx_ring_tail,MAX_FLUSH_WAIT,i);
#endif
		win->buf[entry].l = -(1 + tx_ring_tail);
		win->buf[entry].hlen = HEADER_SIZE+len1;

		len -= cnk_len;
		if (!len)
			break;

#ifdef ALLOW_IRQS
		gamma_time(t_elaps);
		IRQ_enable_for_5usec(t_elaps,t_start);
#endif

/*  Prepare next frame  */
		entry = win->tail;
		incr_modulo(win->tail,GAMMA_TX_WIN_SIZE);
		memcpy(win->buf[entry].b,h,HEADER_SIZE);
		h = (l2s *)(win->buf[entry].b);
		data2 = (void *)((char *)data2+cnk_len-len1);
		len1 = 0;
		cnk_len = gmin(len,frame_size - HEADER_SIZE);
		paddr = next_paddr;
		if (cnk_len >= (PAGE_SIZE-(paddr & 0x00000fff)))
			goto next_page;
		next_paddr += cnk_len;

next_page_back:

		if (cnk_len < len)
			fill_send_header_cnklen(h,cnk_len);
		else
			fill_send_header_cnklen(h,cnk_len | MSG_COMPLETE_MASK);

		continue;

next_page:

		cnk_len = (unsigned short)(PAGE_SIZE-(paddr&0x00000fff));
		next_paddr = user_space_to_phys(current,(char *)data2+cnk_len);
		goto next_page_back;

		}

		tx_wait_ring_room(MAX_FLUSH_WAIT,i);
		if (!i)
			goto tx_failure;
		goto room_in_tx_ring;

	}

end_tx:

	tx_close();

DEBUG(printg("sys_gamma_isend_2p(): Done!  isend_handle %d\n",tx_ring_tail));

#ifdef MORE_INSTANCES_ON_SAME_NODE
	if (phys_dest != BROADCAST) {
		spin_unlock(&gamma_tx_lock);
		return tx_ring_tail;
	}
	res = gamma_send_2p_samenode(
		my_par_pid,procnum,out_port,data1,llen1,ddata2,len2,1
	);

	spin_unlock(&gamma_tx_lock);

	if (!res)
		return tx_ring_tail;
	return res;
#else
	return tx_ring_tail;
#endif

next_page1:

	cnk_len = len1+(unsigned short)(PAGE_SIZE-(paddr & 0x00000fff));
	next_paddr = user_space_to_phys(current,(char *)data2+cnk_len-len1);
	goto next_page_back1;

tx_failure:

DEBUG_L(printg(
"sys_gamma_isend_2p(): par_pid %ld, out_port %ld, frame_no %d: Failure (ring pos %d)\n",
my_par_pid,out_port,rec_header_frame_no(h),tx_ring_head
));

	tx_reset_and_restart();
#ifdef MORE_INSTANCES_ON_SAME_NODE
	spin_unlock(&gamma_tx_lock);
#endif
	return -GAMMA_ERR_SENDFAILURE;

	}

#ifdef MORE_INSTANCES_ON_SAME_NODE
	spin_unlock(&gamma_tx_lock);
#endif
	return -GAMMA_ERR_GENERIC;

#ifdef MORE_INSTANCES_ON_SAME_NODE
same_node:

	if (phys_dest != BROADCAST) {
		res = gamma_send_2p_samenode(
			my_par_pid,procnum,out_port,data1,len1,data2,len2,0
		);
		spin_unlock(&gamma_tx_lock);
		if (!res)
			return -GAMMA_ISEND_IS_DONE;
		return res;
	}
	res = gamma_send_2p_samenode(
		my_par_pid,procnum,out_port,data1,len1,data2,len2,1
	);
	spin_unlock(&gamma_tx_lock);
	if (!res)
		return -GAMMA_ISEND_IS_DONE;
	return res;

dont_reenter:

	return -GAMMA_ERR_TRYAGAIN;
#endif

}


int sys_gamma_isend_2p_flowctl
(
	unsigned long out_port_and_len1_and_was_susp,
	unsigned long my_par_pid_and_procnum,
	void *data2,
	unsigned long len2,
	void *data1
)
{

	unsigned long my_par_pid;
	unsigned long procnum;
	unsigned long out_port;
	unsigned long len1;
	unsigned long len;
	unsigned long was_susp;
	unsigned long phys_dest;
	tx_tmp_ring_type *win;
	unsigned long credit;
	unsigned long curr_credit;
	l2s *h;
	unsigned long entry;
	unsigned long frame_size;
	unsigned long cnk_len;
	unsigned long paddr;
	unsigned long next_paddr;
	tx_common_decl_and_init;
#ifdef ALLOW_IRQS
	time_586 t_start, t_elaps;
	unsigned long flags;
#endif
#ifdef COLL_SAFE_TX
	int trials = MAX_TRIALS;
#endif
#ifdef MORE_INSTANCES_ON_SAME_NODE
	int res;
#endif


#ifdef MORE_INSTANCES_ON_SAME_NODE
	if (!spin_trylock(&gamma_tx_lock))
		goto dont_reenter;
#endif

	my_par_pid = (unsigned short)my_par_pid_and_procnum;
	procnum = my_par_pid_and_procnum >> 16;
	out_port = (out_port_and_len1_and_was_susp & ~0x80000000) >> 8;
	len1 = (unsigned char)out_port_and_len1_and_was_susp;
	len = len1+len2;
	was_susp = out_port_and_len1_and_was_susp & 0x80000000;

	phys_dest =
		par_proc_table[my_par_pid]->inst_table[procnum]
		.port_table[out_port].phys_partner_node;

	if (phys_dest != BROADCAST)
		win = &tx_tmp_ring[phys_dest];
	else
		win = &tx_tmp_ring[phys_whoami];

#ifdef MORE_INSTANCES_ON_SAME_NODE
	if (
		par_proc_table[my_par_pid]->inst_table[procnum]
		.port_table[out_port].partner_is_same_node
	)
		goto same_node;
	if (
		par_proc_table[my_par_pid]->inst_table[procnum]
		.port_table[out_port].bcast_was_susp
	)
		goto same_node_bcast_susp;
#endif

DEBUG_E(printg("sys_gamma_isend_2p_flowctl(): Entering\n"));

	if ((phys_dest == BROADCAST) && (!was_susp)) {
/*
 *  Take note of addr and size of message, for resumption after possible
 *  suspension (shortage of credit).
 */
		par_proc_table[my_par_pid]->inst_table[procnum]
		.port_table[out_port].current_tx_data1 = data1;
		par_proc_table[my_par_pid]->inst_table[procnum]
		.port_table[out_port].current_tx_len1 = len1;
		par_proc_table[my_par_pid]->inst_table[procnum]
		.port_table[out_port].current_tx_data2 = data2;
		par_proc_table[my_par_pid]->inst_table[procnum]
		.port_table[out_port].current_tx_len2 = len2;
	}

	read_credit(phys_dest,credit);
	curr_credit = credit;

	if (len1 <= MAX_1ST_PIECE_SIZE) {

	len = len1+len2;

	if (curr_credit) {

ok_credit:

	h =	par_proc_table[my_par_pid]->inst_table[procnum]
		.port_table[out_port].header;
	entry = win->tail;
	incr_modulo(win->tail,GAMMA_TX_WIN_SIZE);
	memcpy(win->buf[entry].b,h,HEADER_SIZE);
	h = (l2s *)(win->buf[entry].b);

	copy_from_user(&(((unsigned char *)h)[HEADER_SIZE]),data1,len1);

	if (!len2) {

	if (!is_low_watermark(phys_dest,credit,curr_credit)) {
		fill_send_header_cnklen(
			h,
			len1 | MSG_1ST_FRAME_MASK | MSG_COMPLETE_MASK
		);
	}
	else {
		fill_send_header_cnklen(
			h,
			len1 | MSG_1ST_FRAME_MASK | MSG_COMPLETE_MASK
			| CREQ_MASK
		);
	}

	if (!tx_ring_is_full()) {

room_in_tx_ring1:

#ifdef SEQCHECK
	set_send_header_frameno(h,win->tx_seq++);
#endif
#ifdef COLL_SAFE_TX
ro1:
#endif

DEBUG(printg(
"sending through port %ld, frame_no %d\n",out_port,read_send_header_frameno(h)
));

	tx_enq_pkt1(h,HEADER_SIZE+len1);
	tx_start();
	tx_forw_tail();
#ifdef COLL_SAFE_TX
	tx_flush_ring_1(tx_ring_tail,MAX_FLUSH_WAIT*10,i);
	if ((!i) && (--trials >= 0)) {
DEBUG_L(printg("retrying, frame_no %d\n",read_send_header_frameno(h)));
		goto ro1;
	}
#endif
#ifdef SAFE_TX
	if (tx_ring_occupancy >= STEPS_THEN_STOP)
		tx_flush_ring(tx_ring_tail,MAX_FLUSH_WAIT,i);
#endif
	win->buf[entry].l = -(1 + tx_ring_tail);
	win->buf[entry].hlen = HEADER_SIZE+len1;
	curr_credit--;
	goto end_tx;

	}

	tx_wait_ring_room(MAX_FLUSH_WAIT,i);
	if (!i)
		goto tx_failure;
	goto room_in_tx_ring1;

	}

	frame_size = tx_optimal_frag(len);

	cnk_len = gmin(len,frame_size-HEADER_SIZE);

	paddr = user_space_to_phys(current,data2);
	if (cnk_len >= len1+(PAGE_SIZE-(paddr & 0x00000fff)))
		goto next_page1;
	next_paddr = paddr+cnk_len-len1;

next_page_back1:

	if (!was_susp) {
		if (cnk_len < len)
			fill_send_header_cnklen(
				h, cnk_len | MSG_1ST_FRAME_MASK
			);
		else
			fill_send_header_cnklen(
				h,
				cnk_len | MSG_1ST_FRAME_MASK
				| MSG_COMPLETE_MASK
			);
	}
	else {
		if (cnk_len < len)
			fill_send_header_cnklen(h, cnk_len);
		else {
			fill_send_header_cnklen(
				h, cnk_len | MSG_COMPLETE_MASK
			);
		}
	}

#ifdef ALLOW_IRQS
	gamma_time(t_start);
#endif

	for (;;) {

		if (is_low_watermark(phys_dest,credit,curr_credit))
			set_send_header_creq(h);

		if (!tx_ring_is_full()) {

room_in_tx_ring:

#ifdef SEQCHECK
		set_send_header_frameno(h,win->tx_seq++);
#endif
#ifdef COLL_SAFE_TX
ro:
#endif

DEBUG(printg(
"sending through port %ld, frame_no %d\n",out_port,read_send_header_frameno(h)
));

		tx_enq_pkt(h,HEADER_SIZE+len1,(void *)paddr,cnk_len-len1);
		tx_start();
		tx_forw_tail();
#ifdef COLL_SAFE_TX
		tx_flush_ring_1(tx_ring_tail,MAX_FLUSH_WAIT*10,i);
		if ((!i) && (--trials >= 0)) {
DEBUG_L(printg("retrying, frame_no %d\n",read_send_header_frameno(h)));
			goto ro;
		}
#endif
#ifdef SAFE_TX
		if (tx_ring_occupancy >= STEPS_THEN_STOP)
			tx_flush_ring(tx_ring_tail,MAX_FLUSH_WAIT,i);
#endif
		win->buf[entry].l = -(1 + tx_ring_tail);
		win->buf[entry].hlen = HEADER_SIZE+len1;

		curr_credit--;
		len -= cnk_len;
		if (!len)
			break;

		if (!curr_credit)
			goto credit_exhausted;
credit_back:

#ifdef ALLOW_IRQS
		gamma_time(t_elaps);
		IRQ_enable_for_5usec(t_elaps,t_start);
#endif

/*  Prepare next frame  */
		entry = win->tail;
		incr_modulo(win->tail,GAMMA_TX_WIN_SIZE);
		memcpy(win->buf[entry].b,h,HEADER_SIZE);
		h = (l2s *)(win->buf[entry].b);
		data2 = (void *)((char *)data2+cnk_len-len1);
		len1 = 0;
		cnk_len = gmin(len,frame_size - HEADER_SIZE);
		paddr = next_paddr;
		if (cnk_len >= (PAGE_SIZE-(paddr & 0x00000fff)))
			goto next_page;
		next_paddr += cnk_len;

next_page_back:

		if (cnk_len < len)
			fill_send_header_cnklen(h,cnk_len);
		else {
			fill_send_header_cnklen(h,cnk_len | MSG_COMPLETE_MASK);
		}

		continue;

next_page:

		cnk_len = (unsigned short)(PAGE_SIZE-(paddr&0x00000fff));
		next_paddr = user_space_to_phys(current,(char *)data2+cnk_len);
		goto next_page_back;

credit_exhausted:

		dec_credit(phys_dest,credit);

		wait_credit_refill(i,phys_dest,my_par_pid,procnum);

		if (unlikely(par_proc_table[my_par_pid] == NULL)) {
			/*  VM_ABORT or VM_DESTROY pkt catched  */

DEBUG_L(printg(
"Virtual GAMMA %lu aborted/destroyed.  quitting send()...\n",my_par_pid
));

			tx_close();
#ifdef MORE_INSTANCES_ON_SAME_NODE
			spin_unlock(&gamma_tx_lock);
#endif
			return -GAMMA_ERR_GENERIC;
		}

		if (i)
			goto credit_back;

DEBUG_F(printg(
"sys_gamma_isend_2p_flowctl():  No credit.  Suspend (residual len %ld)\n",len
));

		tx_close();
/*  Return the total lenght of the not yet sent part of msg  */
#ifdef MORE_INSTANCES_ON_SAME_NODE
		spin_unlock(&gamma_tx_lock);
#endif
		return -(GAMMA_ERR_MAX+len);

		}

		tx_wait_ring_room(MAX_FLUSH_WAIT,i);
		if (!i)
			goto tx_failure;
		goto room_in_tx_ring;

	}

end_tx:

	tx_close();

	dec_credit(phys_dest,credit-curr_credit);

DEBUG(printg(
"sys_gamma_isend_2p_flowctl(): Done!  isend_handle %d\n",tx_ring_tail
));

#ifdef MORE_INSTANCES_ON_SAME_NODE
	if (phys_dest != BROADCAST) {
		spin_unlock(&gamma_tx_lock);
		return tx_ring_tail;
	}
/*
 *  Here I would like a non-blocking intra-node send(), but this would need
 *  additional hardware...
 */
	res = gamma_send_2p_flowctl_samenode(
		my_par_pid,procnum,out_port,
		par_proc_table[my_par_pid]->inst_table[procnum]
		.port_table[out_port].current_tx_data1,
		par_proc_table[my_par_pid]->inst_table[procnum]
		.port_table[out_port].current_tx_len1,
		par_proc_table[my_par_pid]->inst_table[procnum]
		.port_table[out_port].current_tx_data2,
		par_proc_table[my_par_pid]->inst_table[procnum]
		.port_table[out_port].current_tx_len2,
		1,0
	);
	if (!res) {
		spin_unlock(&gamma_tx_lock);
		return tx_ring_tail;
	}
/*
 *  Store the inter-node isend handle: will be returned to the caller as soon
 *  as the intra-node Tx terminates.
 */
	par_proc_table[my_par_pid]->inst_table[procnum]
	.port_table[out_port].suspended_isend_handle = tx_ring_tail;

	spin_unlock(&gamma_tx_lock);

	if (res == -GAMMA_ERR_TRYAGAIN)
/*  Return the total lenght of the not yet sent part of msg  */
		return -(GAMMA_ERR_MAX+len);
	if (res > 0)
/*  Return the total lenght of the not yet sent part of msg  */
		return -(GAMMA_ERR_MAX+res);
	return res;
#else
	return tx_ring_tail;
#endif

next_page1:

	cnk_len = len1+(unsigned short)(PAGE_SIZE-(paddr & 0x00000fff));
	next_paddr = user_space_to_phys(current,(char *)data2+cnk_len-len1);
	goto next_page_back1;

	}

	wait_credit_refill(i,phys_dest,my_par_pid,procnum);

	if (unlikely(par_proc_table[my_par_pid] == NULL)) {
		/*  VM_ABORT or VM_DESTROY pkt catched  */

DEBUG_L(printg(
"Virtual GAMMA %lu aborted/destroyed.  quitting send()...\n",my_par_pid
));

		tx_close();
#ifdef MORE_INSTANCES_ON_SAME_NODE
		spin_unlock(&gamma_tx_lock);
#endif
		return -GAMMA_ERR_GENERIC;
	}

	if (i)
		goto ok_credit;

DEBUG_E(printg("sys_gamma_isend_2p_flowctl():  No credit.  Try again!\n"));

/*  Return the total lenght of the not yet sent part of msg  */
#ifdef MORE_INSTANCES_ON_SAME_NODE
	spin_unlock(&gamma_tx_lock);
#endif
	return -(GAMMA_ERR_MAX+len);

tx_failure:

DEBUG_L(printg(
"sys_gamma_isend_2p_flowctl(): par_pid %ld, out_port %ld, frame_no %d: Failure (ring pos %d)\n",
my_par_pid,out_port,rec_header_frame_no(h),tx_ring_head
));

	tx_reset_and_restart();
	dec_credit(phys_dest,credit-curr_credit);

#ifdef MORE_INSTANCES_ON_SAME_NODE
	spin_unlock(&gamma_tx_lock);
#endif
	return -GAMMA_ERR_SENDFAILURE;

	}

#ifdef MORE_INSTANCES_ON_SAME_NODE
	spin_unlock(&gamma_tx_lock);
#endif
	return -GAMMA_ERR_GENERIC;

#ifdef MORE_INSTANCES_ON_SAME_NODE
same_node:

	if (phys_dest != BROADCAST) {
/*
 *  Here I would like a non-blocking intra-node send(), but this would need
 *  additional hardware...
 */
		res = gamma_send_2p_flowctl_samenode(
			my_par_pid,procnum,out_port,data1,len1,data2,len2,0,
			was_susp
		);
		spin_unlock(&gamma_tx_lock);
		if (!res)
			return -GAMMA_ISEND_IS_DONE;
		if (res == -GAMMA_ERR_TRYAGAIN)
/*  Return the total lenght of the not yet sent part of msg  */
			return -(GAMMA_ERR_MAX+len);
		if (res > 0)
/*  Return the total lenght of the not yet sent part of msg  */
			return -(GAMMA_ERR_MAX+res);
		return res;
	}

same_node_bcast_susp:

/*
 *  Here I would like a non-blocking intra-node send(), but this would need
 *  additional hardware...
 */
	res = gamma_send_2p_flowctl_samenode(
		my_par_pid,procnum,out_port,data1,len1,data2,len2,1,
		was_susp
	);
	if (!res) {
/*
 *  The intra-node Tx has finished.
 *  Recover the inter-node isend handle, return it to the caller.
 */
		res =	par_proc_table[my_par_pid]->inst_table[procnum]
			.port_table[out_port].suspended_isend_handle;
		par_proc_table[my_par_pid]->inst_table[procnum]
		.port_table[out_port].suspended_isend_handle =
			-GAMMA_ISEND_IS_DONE;
		spin_unlock(&gamma_tx_lock);
		return res;
	}
	spin_unlock(&gamma_tx_lock);
	if (res == -GAMMA_ERR_TRYAGAIN)
/*  Return the total lenght of the not yet sent part of msg  */
		return -(GAMMA_ERR_MAX+len);
	if (res > 0)
/*  Return the total lenght of the not yet sent part of msg  */
		return -(GAMMA_ERR_MAX+res);
	return res;

dont_reenter:

	return -GAMMA_ERR_TRYAGAIN;
#endif

}


int sys_gamma_wsend
(unsigned long isend_handle)
{

	tx_common_decl_and_init;


DEBUG(printg("sys_gamma_wsend(): Entering\n"));

#ifdef MORE_INSTANCES_ON_SAME_NODE
	spin_lock(&gamma_tx_lock);
#endif

	if (isend_handle > tx_ring_tail)
		goto ring_flushed;
	if ((tx_ring_head < tx_ring_tail) && (isend_handle <= tx_ring_head))
		goto ring_flushed;

	tx_flush_ring(isend_handle,MAX_FLUSH_WAIT,i)

	if (i) {

DEBUG(printg("sys_gamma_wsend(): Done\n"));

#ifdef MORE_INSTANCES_ON_SAME_NODE
		spin_unlock(&gamma_tx_lock);
#endif
		return 0;
	}

	tx_start();

	tx_flush_ring(isend_handle,MAX_FLUSH_WAIT,i)

	if (i) {

DEBUG_L(printg("sys_gamma_wsend(): Done, but had to wake up the NIC\n"));

#ifdef MORE_INSTANCES_ON_SAME_NODE
		spin_unlock(&gamma_tx_lock);
#endif
		return 0;
	}

DEBUG_L(printg(
"sys_gamma_wsend(): Failure (isend_handle %ld, entry %d)\n",
isend_handle,tx_ring_head
));

	tx_reset_and_restart();

#ifdef MORE_INSTANCES_ON_SAME_NODE
	spin_unlock(&gamma_tx_lock);
#endif
	return -GAMMA_ERR_SENDFAILURE;

ring_flushed:

DEBUG(printg("sys_gamma_wsend(): Done (tx ring flushed)\n"));

#ifdef MORE_INSTANCES_ON_SAME_NODE
	spin_unlock(&gamma_tx_lock);
#endif
	return 0;

}


int sys_gamma_tsend
(unsigned long isend_handle)
{

	tx_common_decl_and_init;


DEBUG(printg("sys_gamma_tsend(): Entering\n"));

#ifdef MORE_INSTANCES_ON_SAME_NODE
	spin_lock(&gamma_tx_lock);
#endif

	if (isend_handle > tx_ring_tail)
		goto ring_flushed;
	if ((tx_ring_head < tx_ring_tail) && (isend_handle <= tx_ring_head))
		goto ring_flushed;

/* peppe da fare (test invece di flush) */
	tx_flush_ring(isend_handle,MAX_FLUSH_WAIT,i)

	if (i) {

DEBUG(printg("sys_gamma_tsend(): Done\n"));

#ifdef MORE_INSTANCES_ON_SAME_NODE
		spin_unlock(&gamma_tx_lock);
#endif
		return 1;
	}

	tx_start();

	tx_flush_ring(isend_handle,MAX_FLUSH_WAIT,i)

	if (i) {

DEBUG_L(printg("sys_gamma_tsend(): Done, but had to wake up the NIC\n"));

#ifdef MORE_INSTANCES_ON_SAME_NODE
		spin_unlock(&gamma_tx_lock);
#endif
		return 1;
	}

DEBUG_L(printg(
"sys_gamma_tsend(): Failure (isend_handle %ld, entry %d)\n",
isend_handle,tx_ring_head
));

	tx_reset_and_restart();

#ifdef MORE_INSTANCES_ON_SAME_NODE
	spin_unlock(&gamma_tx_lock);
#endif
	return -GAMMA_ERR_SENDFAILURE;

ring_flushed:

DEBUG(printg("sys_gamma_tsend(): Done (tx ring flushed)\n"));

#ifdef MORE_INSTANCES_ON_SAME_NODE
	spin_unlock(&gamma_tx_lock);
#endif
	return 1;

}


/************************************************************************
 *	Receive routines						*
 ************************************************************************/


/*
 *  Returns:
 *  P (0 <= P <= MAX_PORT_NUMBER) if msg complete for port P
 *  (user-level caller will perform notification on sem #P and run rec handler
 *  if present; msg lenght is in user-level variable "gamma_msglen")
 *
 *  P | IS_FIRST_CHUNK if msg not complete for port P but
 *  rec handler must run to manage the arrival of 1st piece of msg
 *  (user-level caller will run the handler, 1st piece lenght is in user-level
 *  variable "gamma_msglen")
 *
 *  -GAMMA_RX_INCOMPLETE if no complete msg and no handler to run
 *
 *  -GAMMA_ERR_MUSTYIELD if there is a pending pkt for a process other than the
 *  invoker
 *
 *  -GAMMA_ERR_BUSY in some other cases.
 */
/*
 *  Allowed vals for "flag":
 *  0:  invoked by a recv(), check for inter-node msgs first
 *  1:  invoked by a recv(), check for intra-node msgs first
 *  2:  invoked by a send() while waiting for flow ctl credit renewal
 *      (see wait_credit_refill() in gamma_macro.h)
 */
int rx_gamma
(struct net_device *dev, unsigned long flag)
{

l2s *h;
unsigned long cnk_len;
unsigned long frame_no;
unsigned long phys_sender;
unsigned long par_pid;
unsigned long port;
unsigned long procnum;
task_t *receiver;
int res;
#ifdef MORE_INSTANCES_ON_SAME_NODE
unsigned long bcast_sequence;
#endif
unsigned long first_piece_size;
unsigned long rx_buffer = 0;
#ifdef MORE_INSTANCES_ON_SAME_NODE
unsigned long head = 0;
unsigned long bcast_frame_no = 0;
#endif
unsigned long err_msg[4];
rx_common_decl_and_init;
#ifdef ALLOW_IRQS
time_586 t_start, t_elaps;
unsigned long flags;
#endif
#ifdef MEASURE_IN_TH
unsigned long cumul_len;
time_586 a,b;
#endif
#ifdef MEASURE_IN_TH_MEM
time_586 ima,imb;
#endif


#ifdef MEASURE_IN_TH
cumul_len = 0;
gamma_time(a);
#endif

if (!spin_trylock(&gamma_rx_lock))
 return -GAMMA_ERR_BUSY;

if (gamma_must_yield != NULL)
 goto check_yield;

ok_yield:

if (gamma_run_handler != NULL)
 goto rec_reentering;

DEBUG_E(printg("rx_gamma(), pid %u: entering...\n",current->pid));

rx_read_pkt_info();

if (rx_err_pktnotavail())
 goto bogus_inv;

#ifdef ALLOW_IRQS
gamma_time(t_start);
#endif

res = -GAMMA_RX_INCOMPLETE;
#ifdef MORE_INSTANCES_ON_SAME_NODE
bcast_sequence = 0;
#endif

for (;;) {

DEBUG_E(printg(
"rx_gamma(): loop: rx_ring_head %d, rx_ring_tail %d, rx_status %4.4x\n",
rx_ring_head,rx_ring_tail,rx_status()
));

 rx_header(h);

DEBUG_E(printg(
"rx_gamma(): h0 %8.8lx, h1 %8.8lx, h2 %8.8lx, h3 %8.8lx, h4 %8.8lx\n",
h[0].ll, h[1].ll, h[2].ll, h[3].ll, h[4].ll
));

/*  Interpret the header  */

 cnk_len = rec_header_cnk_len(h);
 frame_no = rec_header_frame_no(h);
 phys_sender = rec_header_phys_sender(h);
 rec_header_parpid_port(h,par_pid,port,procnum);

 if (rec_header_frmtype(h) != ETH_GAMMA)
  goto pkt_is_not_gamma;

DEBUG(printg(
"rx_gamma(), pid %u: complete %d, par_pid %ld, procnum %ld, port %ld, cnk_len %ld, frame_no %ld, phys_sender %ld, creq %d\n",
current->pid,
rec_header_msg_complete(h), par_pid, procnum, port, cnk_len, frame_no,
phys_sender, rec_header_creq(h)
));

#ifdef SEQCHECK
/****************************************************************************/
/*  Are we refusing regular GAMMA pkts from phys_sender?  */
 if (rx_discard[phys_sender])
  goto end_rx;
/****************************************************************************/
#endif

accept_pkt:

#ifdef MEASURE_IN_TH
 cumul_len += cnk_len;
#endif

#ifdef SEQCHECK
/****************************************************************************/
/*  Check for correct frame sequence numbering.  */
 if (rec_is_bcast(h)) {
/*  Bcast pkts have distinct numbering.  */
  if (frame_no != rx_seq_bcast[phys_sender])
   goto err_frame_no;
  rx_seq_bcast[phys_sender]++;
 }
 else {
  if (frame_no != rx_seq[phys_sender])
   goto err_frame_no;
  rx_seq[phys_sender]++;
 }
/****************************************************************************/
#endif

/*  Check whether virtual GAMMA #par_pid exists and is assigned.  */
 if (
  (par_proc_table[par_pid] == (par_entry *)NULL)
  ||
  (par_proc_table[par_pid]->inst_table[procnum].pid == -1)
 )
  goto not_assigned;

/*  Check whether port is in use as an input one.  */
 receiver =
  par_proc_table[par_pid]->inst_table[procnum].port_table[port].task;
 if (receiver == NULL)
  goto not_assigned;

/*
 *  If there are other pkts in the Rx tmp ring for this process,
 *  all subsequent pkts for the same process must also be stored there;
 *  otherwise, the msg ordering could be violated.
 *  This is unneeded actually, but saves us a liveness check on the receiver
 *  process...
 */
 if (par_proc_table[par_pid]->inst_table[procnum].has_pkts_in_rx_tmp_ring) {

DEBUG(printg(
"rx_gamma(): pkt is diverged to Rx tmp buf to preserve msg ordering\n"
));

  goto diverge_to_rx_tmp_ring;
 }

/*  Check whether receiver is the caller.  */
 if (receiver->mm != current->mm)
  goto cannot_deliver_yet;

#if 0
/*
 *  If caller is on send() and this is an end-of-msg, suspend and leave pkt
 *  in rx queue; receiver will take it on receive() and perform notification
 */
 if (unlikely(flag == FROM_SEND_ROUTINE)) {
  if (rec_header_msg_complete(h)) {
#ifdef SEQCHECK
/*  Roll back frame numbering  */
   if (rec_is_bcast(h))
    rx_seq_bcast[phys_sender]--;
   else
    rx_seq[phys_sender]--;
#endif
/*  Quit  */
   spin_unlock(&gamma_rx_lock);
   return -GAMMA_RX_SUSP;
  }
 }
#endif

/*  Flow Control:  manage the request for credit renewal, if present.  */
 if (rec_header_creq(h))
  goto credit_request;

credit_request_back:

#ifdef MORE_INSTANCES_ON_SAME_NODE
/* Check whether a seq of bcast pkts is beginning or has ended */
 if (rec_is_bcast(h) && (!bcast_sequence))
  goto found_bcast_pkt;
 else
  if (!(rec_is_bcast(h)) && bcast_sequence) {
/*
 *  A seq. of bcast pkts has ended.
 *  Quit the routine, to allow other receivers to get the bcast data
 *  before considering the current pkt.
 */

DEBUG(printg("rx_gamma(): end of bcast pkt seq\n"));

/*
 *  NOTE: variable "gamma_must_yield", indicating which proc instance should
 *  run next in the bcast recv, has been set up when the bcast seq begun.
 */
   goto exit;
  }

found_bcast_pkt_back:
#endif

/*  Frame is OK and we get it now.  */

/*  Empty messages:  nothing to download.  */
 if (!cnk_len)
  goto notif;

/*  Download the payload.  */

 par_proc_table[par_pid]->inst_table[procnum]
 .port_table[port].current_rx_msglen
  += cnk_len;

/*
 *  Is this packet to be scattered into two pieces?
 *  If present, 1st piece is carried by 1st packet of the msg.
 */

 first_piece_size = 0;

 if (rec_header_1st_frame(h)) {
  first_piece_size =
   par_proc_table[par_pid]->inst_table[procnum].port_table[port].boundary_1p;
  if ((first_piece_size > 0) && (first_piece_size < cnk_len)) {
/*  Payload is to be scattered into two pieces.  */

#if 0
/*
 *  Download 1st piece if caller is on receive; suspend and leave pkt in queue
 *  if caller in on send (flow ctl credit waiting)
 */
   if (unlikely(flag == FROM_SEND_ROUTINE)) {
#ifdef SEQCHECK
/*  Roll back frame numbering  */
    if (rec_is_bcast(h))
     rx_seq_bcast[phys_sender]--;
    else
     rx_seq[phys_sender]--;
#endif
/*  Quit  */
    spin_unlock(&gamma_rx_lock);
    return -GAMMA_RX_SUSP;
   }
#endif

DEBUG(printg(
"rx_gamma():  Msg to be scattered, size of first piece is %ld\n",
first_piece_size
));

   get_ptr_to_user_recv_buffer(par_pid,port,procnum,rx_buffer);

DEBUG_E(printg(
"rx_gamma():  Copying 1st chunk to addr %8.8lx\n",(unsigned long)rx_buffer
));

   recv_payload_1st_chunk(rx_buffer,HEADER_SIZE,first_piece_size);
   cnk_len -= first_piece_size;

/*
 *  Prepare to quit, to run the receiver handler of the port.
 *  Since this message is going to be scattered, we can assume
 *  that there is a handler to run.
 */

/*
 *  Tell the caller which handler to run, and len of 1st chunk of data:
 *  "inter-node msg not complete but must run handler"
 */
   res = port | IS_FIRST_CHUNK;

   set_msglen(par_pid,procnum,first_piece_size);

/*
 *  Current process only is allowed to reenter later,
 *  to download remaining part of payload.
 *  Note that the user-level destination of 2nd piece may have been
 *  set/reset/changed by the receiver handler invoked above.
 *  If not, 2nd piece will overwrite 1st one at the same user-level
 *  destination address, since the macro
 *  update_ptr_to_user_recv_buffer() (see below)
 *  has not been invoked.
 */
   gamma_run_handler = current;

DEBUG(printg("Quit to run rec handler for port %ld (1st piece)\n",port));

   spin_unlock(&gamma_rx_lock);
   return res;

  }
  else first_piece_size = 0;
 }

get_remaining_payload:

/*
 *  Shift user-space rec buffer's current ptr forward.
 *  Check for room availability.
 *  If not enough room in rec buffer, truncate the msg chunk
 *  by setting cnk_len equal to the available room.
 *  Decrement the available room by cnk_len.
 *  The current msg chunk will be stored starting
 *  from address "rx_buffer".
 */
 update_ptr_to_user_recv_buffer(par_pid,port,procnum,cnk_len,rx_buffer);

#ifdef MEASURE_IN_TH_MEM
gamma_time(ima);
#endif

/*  Copy the payload (2nd piece) into user space  */
#ifdef USE_DAMNICK

DEBUG_E(printg(
"rx_gamma():  Start copying payload to addr %8.8lx through DAMNICK\n",
(unsigned long)rx_buffer)
);

 if (!rec_header_msg_complete(h))
  damnick_icopy_k2u(
   (void *)rx_buffer,
   (void *)((u32)h+HEADER_SIZE+first_piece_size),
   cnk_len
  );
 else {
  unsigned long pre_notif_val;
  pre_notify_arrival(par_pid,port,procnum,pre_notif_val);
  damnick_icopy_k2u_nu(
   (void *)rx_buffer,
   (void *)((u32)h+HEADER_SIZE+first_piece_size),
   cnk_len,
   (unsigned long *)(par_proc_table[par_pid]->inst_table[procnum].sem1_table)
   + par_proc_table[par_pid]->inst_table[procnum].port_table[port].semaphore,
   pre_notif_val
  );
 }
#else

DEBUG_E(printg(
"rx_gamma():  Copying payload to addr %8.8lx\n",(unsigned long)rx_buffer
));

 recv_payload_2nd_chunk(rx_buffer,HEADER_SIZE,first_piece_size,cnk_len);
#endif

#ifdef MEASURE_IN_TH_MEM
gamma_time(imb);
in_counter_mem++;
in_total_amount_mem_delay += (unsigned long)gamma_time_diff_ticks(imb,ima);
in_total_amount_mem_bytes += cnk_len;
#endif

 if (!rec_header_msg_complete(h))
  goto end_rx;

DEBUG(printg("Message complete\n"));

/*
 *  Do one of the following, depending on the kind of buffer attached
 *  to the recv port:
 *  case GO_BACK:
 *  reset the current position ptr of the recv buffer to the begin of
 *  the recv buffer itself.  Next msg will be stored at the begin
 *  of the buffer.
 *  case GO_AHEAD:
 *  If there is still room available in the attached buffer,
 *  leave the current position ptr untouched.
 *  Otherwise, reset the current position ptr of the recv buffer to
 *  the begin of the recv buffer itself; in this case, next msg will
 *  be stored at the begin of the buffer.
 */
 switch (
  par_proc_table[par_pid]->inst_table[procnum].port_table[port].buffer_kind
 ) {
  case GO_BACK:
   reset_current_ptr(par_pid,port,procnum);
   break;
  case GO_AHEAD:
   check_available_room_and_possibly_reset_current_ptr(par_pid,port,procnum);
   break;
 }

#ifdef USE_DAMNICK
/*
 *  Skip CPU-operated notification of arrival: it will be carried out by the
 *  DAMNICK device, because of the invocation of damnick_icopy_k2u_nu() .
 */
 goto set_msglen;
#endif

notif:

#ifdef USE_DAMNICK

DEBUG(printg("Notifying arrival at port %ld\n",port));

 {
 unsigned long pre_notif_val;
 pre_notify_arrival(par_pid,port,procnum,pre_notif_val);
 damnick_writel_u(
  (unsigned long *)
  (par_proc_table[par_pid]->inst_table[procnum].sem1_table)+
  par_proc_table[par_pid]->inst_table[procnum].port_table[port].semaphore,
  pre_notif_val
 );
 }

set_msglen:

#endif

/*
 *  Prepare to quit, to notify msg arrival and run the receiver handler
 *  (if a handler is present and it did not run already)
 */
 res = port;

DEBUG_E(printg(
"Setting \"gamma_msglen\" to %ld\n",
par_proc_table[par_pid]->inst_table[procnum].port_table[port].current_rx_msglen
));

 set_msglen(
  par_pid,procnum,
  par_proc_table[par_pid]->inst_table[procnum].port_table[port]
   .current_rx_msglen
 );

/*  Next msg starts from len = 0  */
 par_proc_table[par_pid]->inst_table[procnum]
 .port_table[port].current_rx_msglen = 0;

/*
 *  Tell the NIC this packet has been consumed, then quit rx_gamma()
 *  because a new message is now complete.
 *  Bcast pkts cannot be removed from Rx ring until all other receivers have
 *  read them.
 */
#ifdef MORE_INSTANCES_ON_SAME_NODE
 if (bcast_sequence)
  goto quit_rx_loop_bcastseq;
#endif
 discard_pkt();
 break;

#ifdef MORE_INSTANCES_ON_SAME_NODE
quit_rx_loop_bcastseq:

 rec_header_set_procnum(h,procnum+1);
 forward_rx_ring_head();
/*
 *  NOTE: variable "gamma_must_yield", indicating which proc instance should
 *  run next in the bcast recv, has been set up when the bcast seq begun.
 */
 break;
#endif

end_rx:

/*
 *  Tell the NIC this packet has been consumed.
 *  Bcast pkts cannot be removed from Rx ring until all other receivers have
 *  read them.
 */
#ifdef MORE_INSTANCES_ON_SAME_NODE
 if (!bcast_sequence) {
  discard_pkt();
 }
 else {
  rec_header_set_procnum(h,procnum+1);
  forward_rx_ring_head();
 }
#else
 discard_pkt();
#endif

#ifdef ALLOW_IRQS
gamma_time(t_elaps);
IRQ_enable_for_5usec(t_elaps,t_start);
#endif

end_rx_with_updated_ptrs:

 rx_read_pkt_info();

/*  If no further pending frames, quit rx_gamma()  */
 if (rx_err_pktnotavail())
  break;

/*  There is at least another pending frame.  */

DEBUG_E(printg("rx_gamma(): continuing loop...\n"));

}

exit:

#ifdef MEASURE_IN_TH
gamma_time(b);
#ifdef MEASURE_IN_TH_PRINTALL
printg("in: %ld usec, %ld bytes, %ld MByte/s\n",
	((unsigned long)gamma_time_diff_ticks(b,a))/CLOCK,cumul_len,
	(cumul_len*CLOCK)/(unsigned long)gamma_time_diff_ticks(b,a)
);
#endif
in_counter++;
in_total_amount += (cumul_len*CLOCK)/(unsigned long)gamma_time_diff_ticks(b,a);
#endif

#ifdef MORE_INSTANCES_ON_SAME_NODE
if (bcast_sequence)
 goto quit_rx_gamma_bcastseq;
#endif

/*  Quit.  Possibly yield the CPU.  */
if ((gamma_must_yield != NULL) && (res < 0))
 res = -GAMMA_ERR_MUSTYIELD;

#ifdef MORE_INSTANCES_ON_SAME_NODE
DEBUG_E(printg(
"rx_gamma(): exiting: rx_ring_head %d (%ld), rx_ring_tail %d, res %d\n",
 rx_ring_head,head,rx_ring_tail,res
));
#else
DEBUG_E(printg(
"rx_gamma(): exiting: rx_ring_head %d, rx_ring_tail %d, res %d\n",
 rx_ring_head,rx_ring_tail,res
));
#endif

spin_unlock(&gamma_rx_lock);
return res;

#ifdef MORE_INSTANCES_ON_SAME_NODE
quit_rx_gamma_bcastseq:

/*
 *  Next local receiver must re-read bcast pkts.
 *  Set ring head back, and roll-back frame numbering sequence.
 */
write_rx_ring_head(head);
rx_seq_bcast[phys_sender] = bcast_frame_no;
/*
 *  Quit.  Yield the CPU.
 *  "gamma_must_yield" already set to next bcast receiver.
 */
if (res < 0)
 res = -GAMMA_ERR_MUSTYIELD;

#ifdef MORE_INSTANCES_ON_SAME_NODE
DEBUG_E(printg(
"rx_gamma(): exiting: rx_ring_head %d (%ld), rx_ring_tail %d, res %d\n",
 rx_ring_head,head,rx_ring_tail,res
));
#else
DEBUG_E(printg(
"rx_gamma(): exiting: rx_ring_head %d, rx_ring_tail %d, res %d\n",
 rx_ring_head,rx_ring_tail,res
));
#endif

spin_unlock(&gamma_rx_lock);
return res;
#endif


/*  Here all the unusual branches.  */

#ifdef SEQCHECK
err_frame_no:

/*
 *  There are two possibilities, namely:
 *  a)  A previous frame was lost.
 *      In this case, a retransmission must be asked to sender.
 *  b)  This frame is a useless duplicate of a previous one,
 *      maybe caused by a late collision on a shared LAN.
 *      In this case the previous one should have been discarded
 *      and replaced by this one, but this is no longer possible
 *      (a receiver handler could have been issued).  Since no
 *      error was detected on the previous frame, we hope that
 *      everything is OK and this duplicate may be discarded.
 *      However late collisions should never occur if the LAN is
 *      properly cabled.
 */

if (!rec_is_bcast(h)) {
 if (frame_no == (unsigned char)(rx_seq[phys_sender]-1)) {
/*  Case b).  Discard...  */

DEBUG_L(printg(
"rx_gamma():  par_pid %ld, port %ld:  discarding duplicated frame #%ld, rx_status %8.8x\n",
par_pid,port,frame_no,rx_status()
));

  goto end_rx;
 }
}
else
 if (frame_no == (unsigned char)(rx_seq_bcast[phys_sender]-1)) {
/*  Case b).  Discard...  */

DEBUG_L(printg(
"rx_gamma():  par_pid %ld, port %ld:  discarding duplicated bcast frame #%ld, rx_status %8.8x\n",
par_pid,port,frame_no,rx_status()
));

  goto end_rx;
 }

/*  Case a).  */

DEBUG_L(
if (!rec_is_bcast(h)) {
 printg("rx_gamma(): Received frame #%ld, expctd #%d, phys sender %ld\n",
 frame_no,rx_seq[phys_sender],phys_sender);
}
else {
 printg("rx_gamma(): Received bcast frame #%ld, expctd #%d, phys sender %ld\n",
 frame_no,rx_seq_bcast[phys_sender],phys_sender);
}
);

/*
 *  The interval between consecutive checks for correct pkt numbering
 *  is restored to the base value (it was growing as long as no pkt losses were
 *  observed).
 */
max_polls = base_max_polls;

/*  A missing pkt could indicate a congestion.  Ask everybody to slow down.  */
gamma_kernel_send(BROADCAST,DONT_MIND,CONGESTION,NULL,0,1);
/*
 *  Send a NACK to the origin of this pkt, to enforce retransmission of pkts
 *  from "frame_no" to the actual count (go-back-N).
 */
if (!rec_is_bcast(h)) {
 err_msg[0] = rx_seq[phys_sender];
 err_msg[1] = par_pid;
 err_msg[2] = port;
 err_msg[3] = procnum;
 gamma_kernel_send(
  phys_sender,DONT_MIND,NACK,err_msg,4*sizeof(unsigned long),1
 );
}
else {
 err_msg[0] = rx_seq_bcast[phys_sender];
 err_msg[1] = par_pid;
 err_msg[2] = port;
 err_msg[3] = procnum;
 gamma_kernel_send(
  phys_sender,DONT_MIND,NACKB,err_msg,4*sizeof(unsigned long),1
 );
}
/*  We refuse other pkts from phys_sender until we get the missing ones.  */
rx_discard[phys_sender] = 1;

goto end_rx;
#endif

not_assigned:

/*
 *  Virtual GAMMA not activated or application not yet registered or port not in use.
 *  Before discarding the pkt, check for a request for a flow ctl
 *  credit renewal...
 */
if (rec_header_creq(h)) {

DEBUG_F(printg("got CREQ from node %ld, frame_no %ld\n",phys_sender,frame_no))

 gamma_kernel_send(phys_sender,DONT_MIND,FILL_CREDIT,NULL,0,1);
 rec_header_clean_creq(h);
}

DEBUG(
printg("rx_gamma(): h0 %8.8lx, h1 %8.8lx, h2 %8.8lx:  ",
 h[0].ll, h[1].ll, h[2].ll
);
printg("Par_pid or port not assigned.  Discarding frame (rx_status %8.8x)...\n",
rx_status())
);

goto end_rx;

pkt_is_not_gamma:

if (rec_header_frmtype(h) == ETH_GAMMA_KERNEL) {
/*
 *  GAMMA kernel msg.
 *  In this case ``procnum'' is the kind of kernel msg.
 */
#ifdef SEQCHECK
/****************************************************************************/
 if ((procnum == FILL_CREDIT) || (procnum == CRED_REQ)) {
/*  Are we refusing regular/flowctl GAMMA pkts from phys_sender?  */
  if (rx_discard[phys_sender])
   goto end_rx;
/*  Check for correct frame sequence numbering.  */
  if (rec_is_bcast(h)) {
/*  Bcast pkts have distinct numbering.  */
   if (frame_no != rx_seq_bcast[phys_sender])
    goto err_frame_no;
   rx_seq_bcast[phys_sender]++;
  }
  else {
   if (frame_no != rx_seq[phys_sender])
    goto err_frame_no;
   rx_seq[phys_sender]++;
  }
 }
/****************************************************************************/
#endif

DEBUG(printg(
"rx_gamma(): kernel! complete %d, parpid %ld, type %ld, port %ld, cnk_len %ld, phys_sender %ld\n",
rec_header_msg_complete(h), par_pid, procnum, port, cnk_len, phys_sender
));

DEBUG_F(
if (procnum == FILL_CREDIT)
 printg("got FILL_CREDIT from node %ld, frame_no %ld\n",
  phys_sender,frame_no);
if (procnum == CRED_REQ)
 printg("got CRED_REQ from node %ld, frame_no %ld\n",
  phys_sender,frame_no);
)

 gamma_kernel_rec(
  (unsigned char *)h+HEADER_SIZE,par_pid,cnk_len,procnum,phys_sender
 );
 goto end_rx;
}
#ifdef SEQCHECK
if (rec_header_frmtype(h) == ETH_GAMMA_RETX) {
/*
 *  Regular GAMMA data pkt, simply retransmitted.
 *  Can only arrive after a ETH_GAMMA_KERNEL pkt marked as "START_RETX"
 *  (start of retransmission sequence) has arrived from node "phys_sender",
 *  so we should find rx_discard[phys_sender] == 2 (accept retrasmission
 *  sequence from node "phys_sender").
 */
/*  Are we refusing regular GAMMA pkts from phys_sender?  */
 if (rx_discard[phys_sender] == 1) {

DEBUG_F(printg(
"rx_gamma():  refusing ETH_GAMMA_RETX pkt, frame_no %ld, sender %ld\n",
frame_no,phys_sender
));

  goto end_rx;
 }
 if (rx_discard[phys_sender] == 2) {
/*  Treat this pkt as it were a regular one  */

DEBUG_F(printg(
"rx_gamma(), pid %u: RETX! complete %d, par_pid %ld, procnum %ld, port %ld, cnk_len %ld, frame_no %ld, phys_sender %ld, creq %d\n",
current->pid,
rec_header_msg_complete(h), par_pid, procnum, port, cnk_len, frame_no,
phys_sender, rec_header_creq(h)
));

  goto accept_pkt;
 }
/*
 *  This retransmitted pkt is unexpected.  Maybe because
 *  it is a retransmitted bcast (in this case we must discard it);
 *  or, a RETX_START pkt has been lost
 *  (and we should rather ask the sender to restart its retransmission in
 *  this case; however, the protocol will eventually recover even if we
 *  discard this pkt).
 */

DEBUG(printg(
"rx_gamma(): unexpected ETH_GAMMA_RETX pkt, frame_no %ld, phys_sender %ld\n",
frame_no,phys_sender
));

  goto end_rx;
}
if (rec_header_frmtype(h) == ETH_GAMMA_KERNEL_RETX) {
/*
 *  GAMMA kernel msg, retransmitted.
 *  In this case ``procnum'' is the kind of kernel msg.
 *  Possible pkt kinds are:  FILL_CREDIT,CRED_REQ.
 */
/*  Are we refusing regular GAMMA pkts from phys_sender?  */
 if (rx_discard[phys_sender] == 1) {

DEBUG_F(printg(
"rx_gamma():  refusing ETH_GAMMA_KERNEL_RETX pkt, frame_no %ld, sender %ld\n",
frame_no,phys_sender
));

  goto end_rx;
 }
 if (rx_discard[phys_sender] == 2) {
/*  Treat this pkt as it were a regular one  */
/*  Check for correct frame sequence numbering.  */
  if (rec_is_bcast(h)) {
/*  Bcast pkts have distinct numbering.  */
   if (frame_no != rx_seq_bcast[phys_sender])
    goto err_frame_no;
   rx_seq_bcast[phys_sender]++;
  }
  else {
   if (frame_no != rx_seq[phys_sender])
    goto err_frame_no;
   rx_seq[phys_sender]++;
  }

DEBUG(printg(
"rx_gamma(): retx kernel! complete %d, parpid %ld, type %ld, port %ld, cnk_len %ld, phys_sender %ld\n",
rec_header_msg_complete(h), par_pid, procnum, port, cnk_len, phys_sender
));

DEBUG_F(
if (procnum == FILL_CREDIT)
 printg("got FILL_CREDIT (retx) from node %ld, frame_no %ld\n",
  phys_sender,frame_no);
if (procnum == CRED_REQ)
 printg("got CRED_REQ (retx) from node %ld, frame_no %ld\n",
  phys_sender,frame_no);
)

  gamma_kernel_rec(
   (unsigned char *)h+HEADER_SIZE,par_pid,cnk_len,procnum,phys_sender
  );
  goto end_rx;
 }
/*
 *  This retransmitted pkt is unexpected.  Maybe because
 *  it is a retransmitted bcast (in this case we must discard it);
 *  or, a RETX_START pkt has been lost
 *  (and we should rather ask the sender to restart its retransmission in
 *  this case; however, the protocol will eventually recover even if we
 *  discard this pkt).
 */

DEBUG(printg(
"rx_gamma(): unexpected ETH_GAMMA_KERNEL_RETX pkt, frame_no %ld, phys_sender %ld\n",
frame_no,phys_sender
));

 goto end_rx;
}
#endif
/*  Non-GAMMA frame.  "h" contains its first HEADER_SIZE bytes.  */

DEBUG_E(printg("rx_gamma(): non-GAMMA frame.\n"));

/*  Route the frame through the standard path.  */
invoke_standard_rx(dev);
goto end_rx_with_updated_ptrs;

#ifdef MORE_INSTANCES_ON_SAME_NODE
found_bcast_pkt:

/*
 *  Found a bcast pkt.  We suppose it is the first of a sequence.
 *  Such pkts will not be removed from the Rx ring until all receivers
 *  have copied them into user space.  Instead, remove all the other
 *  pkts received so far (by forwarding the Rx ring head to the current pos).
 *  Also, pre-compute the next local receiver and propagate SIGURG to it
 *  if required.
 */
/*  Is there any receiver next?  */
if (
 (procnum < MAX_NUM_LOCAL_INST-1) &&
 (par_proc_table[par_pid]->inst_table[procnum+1].whoami >= 0)
) {
/*  Yes, there is!  */

DEBUG(printg("rx_gamma():  start of bcast seq\n"));

 bcast_sequence = 1;
/*
 *  Take note of the current Rx ring head and frame numbering:
 *  next local receiver will have to re-read pkts from there on.
 */
 read_rx_ring_head(head);
 bcast_frame_no = frame_no;

#if 0
 if (!atomic_read(&gamma_susp_irq_count)) {
/*
 *  IRQ-driven receive is enabled.
 *  Issue a SIGURG to the next receiver process, to force it to eventually
 *  poll the shared buffer (the signal simulates an IRQ).
 */

DEBUG(printg(
"rx_gamma(): Propagating SIGURG to procnum %ld (bcast)\n", procnum+1
));

  force_sig(SIGURG,par_proc_table[par_pid]->inst_table[procnum+1].task);
 }
#endif

/*  Prepare to yield CPU to next receiver in the bcast group.  */
 gamma_must_yield = par_proc_table[par_pid]->inst_table[procnum+1].task;
}
/*  If no next local receiver, treat bcast as it were unicast.  */
goto found_bcast_pkt_back;
#endif

cannot_deliver_yet:

{
 task_t *p;
/*  Is the receiver process alive?  */
 p = find_task_by_pid(receiver->pid);
 if (p == NULL)
  goto dead_process;

 if (receiver->state & (TASK_ZOMBIE | TASK_DEAD))
  goto dead_process;

DEBUG(printg(
"rx_gamma(), pid %u (%08lx):  delaying recv of pkt #%lu, receiver is %u (%08lx), state is %ld\n",
current->pid,(unsigned long)current->mm,frame_no,
receiver->pid,(unsigned long)receiver->mm,receiver->state
));

 if (receiver->state != TASK_RUNNING)
/*
 *  This pkt and all subsequent ones for the same receiver must be
 *  copied to the Rx tmp ring.  The receiver will consume them from the
 *  Rx tmp ring when it becomes running.
 */
  goto diverge_to_rx_tmp_ring_2;

/*  Receiver is runnable.  No need for moving pkts to the Rx tmp ring  */
#ifdef SEQCHECK
/****************************************************************************/
/*  Rollback the frame sequence numbering.  */
 if (rec_is_bcast(h))
/*  Bcast pkts have distinct numbering.  */
  rx_seq_bcast[phys_sender]--;
 else
  rx_seq[phys_sender]--;
/****************************************************************************/
#endif
#if 0
/*
 * Issue a SIGURG to the receiver process, to force it to poll the device
 * so as to free the Rx ring ASAP (the signal simulates an IRQ).
 */
 if (!atomic_read(&gamma_susp_irq_count))
  force_sig(SIGURG,receiver);
#endif
/*  Prepare to yield CPU to receiver.  */
 gamma_must_yield = receiver;
 res = -GAMMA_ERR_MUSTYIELD;
/*  Exit the IRQ service routine, leaving the pkt on the NIC Rx queue.  */
 goto exit;

diverge_to_rx_tmp_ring_2:

/*  Receiver is not runnable.  Must diverge this pkt to the Rx tmp ring.  */

DEBUG(printg("rx_gamma(): moving pkt #%lu to Rx tmp ring\n",frame_no));

 rx_tmp_ring.buf[rx_tmp_ring.tail] = (void *)h;
 incr_modulo(rx_tmp_ring.tail,RX_TMP_RING_SIZE);
 if (!par_proc_table[par_pid]->inst_table[procnum].has_pkts_in_rx_tmp_ring) {
/*
 *  Signal the receiver process that it has pkts in the Rx tmp buffers
 *  (one time only; no need to repeat this for each subsequent pkt).
 */
  par_proc_table[par_pid]->inst_table[procnum].has_pkts_in_rx_tmp_ring = 1;
  force_sig(SIGPWR,receiver);
 }
/*  Remove pkt from rx queue.  */
 goto end_rx;
}

diverge_to_rx_tmp_ring:

rx_tmp_ring.buf[rx_tmp_ring.tail] = (void *)h;
incr_modulo(rx_tmp_ring.tail,RX_TMP_RING_SIZE);
goto end_rx;

dead_process:

/*  Receiver is no longer existing.  Discard this packet.  */

DEBUG(printg("rx_gamma(): recv proc does not exist.  Discarding packet...\n"));

goto end_rx;

credit_request:

DEBUG_F(printg(
"got CREQ from node %ld, frame_no %ld (is_bcast %d)\n",
phys_sender,frame_no,rec_is_bcast(h)
))

/*  This packet carries a credit request. Refill credit to sender.  */
#ifdef MORE_INSTANCES_ON_SAME_NODE
/*
 *  Exception: if this is a bcast, refill credit
 *  only if this receiver is the last one of the bcast.
 */
if (!(rec_is_bcast(h))) {
#endif
/*  Refill credit and clean the credit request.  */
 gamma_kernel_send(phys_sender,DONT_MIND,FILL_CREDIT,NULL,0,1);
 rec_header_clean_creq(h);
 goto credit_request_back;
#ifdef MORE_INSTANCES_ON_SAME_NODE
}
/*
 *  Bcast packet.
 *  Refill (and clean the request) iff no other receivers (excluding this one).
 */
if (
 (procnum >= MAX_NUM_LOCAL_INST-1) ||
 (par_proc_table[par_pid]->inst_table[procnum+1].whoami < 0)
) {
 gamma_kernel_send(phys_sender,DONT_MIND,FILL_CREDIT,NULL,0,1);
 rec_header_clean_creq(h);
}
goto credit_request_back;
#endif

check_yield:

if (likely(gamma_must_yield == current)) {
 gamma_must_yield = NULL;
 goto ok_yield;
}
spin_unlock(&gamma_rx_lock);
return -GAMMA_ERR_BUSY;

bogus_inv:

DEBUG_E(printg(
"rx_gamma(): bogus invocation (rx_status %8.8x), pid %u\n",
rx_status(),current->pid
));

spin_unlock(&gamma_rx_lock);
return -GAMMA_NO_RX_YET;

rec_reentering:

if (gamma_run_handler != current) {
/*
 *  A receiver has temporarily quit to run the rec handler, but will
 *  reenter soon.
 *  Prevent other processes to enter now.
 */
 spin_unlock(&gamma_rx_lock);
 return -GAMMA_ERR_BUSY;
}
/*  The receiver is reentering after running a rec handler (1sh piece).  */

DEBUG_E(printg("rx_gamma(), pid %u: reentering...\n",current->pid));

gamma_run_handler = NULL;
rx_read_pkt_info();
res = -GAMMA_RX_INCOMPLETE;

DEBUG_E(printg(
"rx_gamma(): rx_ring_head %d, rx_ring_tail %d, rx_status %4.4x\n",
rx_ring_head,rx_ring_tail,rx_status()
));

rx_header(h);

cnk_len = rec_header_cnk_len(h);
frame_no = rec_header_frame_no(h);
phys_sender = rec_header_phys_sender(h);
rec_header_parpid_port(h,par_pid,port,procnum);

DEBUG(printg(
"rx_gamma(), pid %u:  reentering: complete %d, par_pid %ld, procnum %ld, port %ld, cnk_len %ld, frame_no %ld, phys_sender %ld, creq %d\n",
current->pid,
rec_header_msg_complete(h), par_pid, procnum, port, cnk_len, frame_no,
phys_sender, rec_header_creq(h)
));

receiver =
 par_proc_table[par_pid]->inst_table[procnum].port_table[port].task;

#ifdef MORE_INSTANCES_ON_SAME_NODE
bcast_sequence = 0;
if (rec_is_bcast(h)) {
 if (
  (procnum < MAX_NUM_LOCAL_INST-1) &&
  (par_proc_table[par_pid]->inst_table[procnum+1].whoami >= 0)
 ) {
  bcast_sequence = 1;
  read_rx_ring_head(head);
  bcast_frame_no = frame_no;
  gamma_must_yield = par_proc_table[par_pid]->inst_table[procnum+1].task;
 }
}
#endif

first_piece_size =
 par_proc_table[par_pid]->inst_table[procnum].port_table[port].boundary_1p;

cnk_len -= first_piece_size;

goto get_remaining_payload;

}


/*
 *  This routine is similar to rx_gamma() but works in the Rx ring of
 *  temporary buffers, where rx_gamma() stores all packets addressed to
 *  non-runnable processes.  This routine is then invoked as soon as
 *  the receiver process becomes running.
 *  It will consume all pending pkts belonging to the invoking process.
 */
/*
 *  Returns:
 *
 *  P | IS_SAME_NODE (0 <= P <= MAX_PORT_NUMBER) if msg complete for port P
 *  (user-level caller will perform notification on sem #P and run rec handler
 *  if present; msg lenght is in user-level variable "gamma_msglen")
 *
 *  P | IS_FIRST_CHUNK | IS_SAME_NODE if msg not complete for port P but
 *  rec handler must run to manage the arrival of 1st piece of msg
 *  (user-level caller will run the handler, 1st piece lenght is in user-level
 *  variable "gamma_msglen")
 *
 *  -GAMMA_RX_INCOMPLETE if no complete msg and no handler to run
 *
 *  -GAMMA_ERR_BUSY in some other cases.
 */
int sys_gamma_read_rx_tmpbuf
(unsigned char my_par_pid, unsigned char my_procnum)
{

l2s *h;
unsigned long tail;
int res;
unsigned long cnk_len;
unsigned long frame_no;
unsigned long phys_sender=0;
unsigned long par_pid;
unsigned long port;
unsigned long procnum;
task_t *receiver;
#ifdef MORE_INSTANCES_ON_SAME_NODE
unsigned long bcast_sequence;
#endif
unsigned long rx_buffer = 0;
unsigned long first_piece_size;
#ifdef MORE_INSTANCES_ON_SAME_NODE
unsigned long head = 0;
#endif


if (!spin_trylock(&gamma_rx_lock))
 return -GAMMA_ERR_BUSY;

if (gamma_run_handler != NULL)
 goto rec_reentering;

DEBUG_E(printg("sys_gamma_read_rx_tmpbuf(): entering...\n"));

if (rx_tmp_ring.head == rx_tmp_ring.tail)
 goto bogus_inv;

tail = rx_tmp_ring.tail;
res = -GAMMA_RX_INCOMPLETE;
#ifdef MORE_INSTANCES_ON_SAME_NODE
bcast_sequence = 0;
#endif

for (;;) {

DEBUG_E(printg(
"sys_gamma_read_rx_tmpbuf(): loop: rx_ring_head %ld, rx_ring_tail %ld\n",
rx_tmp_ring.head,rx_tmp_ring.tail
));

 h = (l2s *)(rx_tmp_ring.buf[rx_tmp_ring.head]);

DEBUG_E(printg(
"sys_gamma_read_rx_tmpbuf(): h0 %8.8lx, h1 %8.8lx, h2 %8.8lx, h3 %8.8lx, h4 %8.8lx\n",
h[0].ll, h[1].ll, h[2].ll, h[3].ll, h[4].ll
));

/*  Interpret the header  */

 cnk_len = rec_header_cnk_len(h);
 frame_no = rec_header_frame_no(h);
 phys_sender = rec_header_phys_sender(h);
 rec_header_parpid_port(h,par_pid,port,procnum);

DEBUG(printg(
"sys_gamma_read_rx_tmpbuf(), pid %u: complete %d, par_pid %ld, procnum %ld, port %ld, cnk_len %ld, frame_no %ld, phys_sender %ld, creq %d\n",
current->pid,
rec_header_msg_complete(h), par_pid, procnum, port, cnk_len, frame_no,
phys_sender, rec_header_creq(h)
));

/*  Flow control:  manage the request for credit renewal, if present.  */
 if (rec_header_creq(h))
  goto credit_request;

credit_request_back:

/*  Check whether virtual GAMMA #par_pid exists and is assigned.  */
 if (
  (par_proc_table[par_pid] == (par_entry *)NULL)
  ||
  (par_proc_table[par_pid]->inst_table[procnum].pid == -1)
 )
  goto not_assigned;

/*  Check whether port is in use as an input one.  */
 receiver =
  par_proc_table[par_pid]->inst_table[procnum].port_table[port].task;
 if (receiver == NULL)
  goto not_assigned;

/*  Check whether receiver is the caller.  */
 if (receiver->mm != current->mm)
  goto cannot_deliver_yet;

#ifdef MORE_INSTANCES_ON_SAME_NODE
/* Check whether a seq of bcast pkts is beginning or has ended */
 if (rec_is_bcast(h) && (!bcast_sequence))
  goto found_bcast_pkt;
 else
  if (!(rec_is_bcast(h)) && bcast_sequence) {
/*
 *  A seq. of bcast pkts has ended.
 *  Quit the routine, to allow other receivers to get the bcast data
 *  before considering the current pkt.
 */

DEBUG(printg("sys_gamma_read_rx_tmpbuf(): end of bcast pkt seq\n"));

   break;
  }

found_bcast_pkt_back:
#endif

/*  Frame is OK!  */

/*  Empty messages:  nothing to download.  */
 if (!cnk_len)
  goto notif;

/*  Download the payload.  */

 par_proc_table[par_pid]->inst_table[procnum]
 .port_table[port].current_rx_msglen
  += cnk_len;

/*
 *  Is this packet to be scattered into two pieces?
 *  If present, 1st piece is carried by 1st packet of the msg.
 */

 first_piece_size = 0;

 if (rec_header_1st_frame(h)) {
  first_piece_size =
   par_proc_table[par_pid]->inst_table[procnum].port_table[port].boundary_1p;
  if ((first_piece_size > 0) && (first_piece_size < cnk_len)) {
/*
 *  Payload is to be scattered into two pieces.
 *  Download 1st piece.
 */

DEBUG(printg(
"sys_gamma_read_rx_tmpbuf():  Msg to be scattered, size of first piece is %ld\n", first_piece_size
));

   get_ptr_to_user_recv_buffer(par_pid,port,procnum,rx_buffer);
   recv_payload_1st_chunk(rx_buffer,HEADER_SIZE,first_piece_size);
   cnk_len -= first_piece_size;

/*
 *  Prepare to quit, to run the receiver handler of the port.
 *  Since this message is going to be scattered, we can assume
 *  that there is a handler to run.
 */

/*
 *  Tell the caller which handler to run, and len of 1st chunk of data:
 *  "inter-node (delayed) msg not complete but must run handler"
 */
   res = port | IS_FIRST_CHUNK;

   set_msglen(par_pid,procnum,first_piece_size);

/*
 *  Current process only is allowed to reenter later,
 *  to download remaining part of payload.
 *  Note that the user-level destination of 2nd piece may have been
 *  set/reset/changed by the receiver handler invoked above.
 *  If not, 2nd piece will overwrite 1st one at the same user-level
 *  destination address, since the macro
 *  update_ptr_to_user_recv_buffer() (see below)
 *  has not been invoked.
 */
   gamma_run_handler = current;

DEBUG(printg("Quit to run rec handler for port %ld (1st piece)\n",port));

   spin_unlock(&gamma_rx_lock);
   return res;

  }
  else first_piece_size = 0;
 }

get_remaining_payload:

/*
 *  Shift user-space rec buffer's current ptr forward.
 *  Check for room availability.
 *  If not enough room in rec buffer, truncate the msg chunk
 *  by setting cnk_len equal to the available room.
 *  Decrement the available room by cnk_len.
 *  The current msg chunk will be stored starting
 *  from address "rx_buffer".
 */
 update_ptr_to_user_recv_buffer(par_pid,port,procnum,cnk_len,rx_buffer);

/*  Copy the payload (2nd piece) into user space  */
#ifdef USE_DAMNICK

DEBUG_E(printg(
"sys_gamma_read_rx_tmpbuf():  Start copying payload to addr %8.8lx through DAMNICK\n",
(unsigned long)rx_buffer)
);

 if (!rec_header_msg_complete(h))
  damnick_icopy_k2u(
   (void *)rx_buffer,
   (void *)((u32)h+HEADER_SIZE+first_piece_size),
   cnk_len
  );
 else {
  unsigned long pre_notif_val;
  pre_notify_arrival(par_pid,port,procnum,pre_notif_val);
  damnick_icopy_k2u_nu(
   (void *)rx_buffer,
   (void *)((u32)h+HEADER_SIZE+first_piece_size),
   cnk_len,
   (unsigned long *)(par_proc_table[par_pid]->inst_table[procnum].sem1_table)
   + par_proc_table[par_pid]->inst_table[procnum].port_table[port].semaphore,
   pre_notif_val
  );
 }
#else

DEBUG_E(printg(
"sys_gamma_read_rx_tmpbuf():  Copying payload to addr %8.8lx\n",
(unsigned long)rx_buffer
));

 recv_payload_2nd_chunk(rx_buffer,HEADER_SIZE,first_piece_size,cnk_len);
#endif

 if (!rec_header_msg_complete(h))
  goto end_rx;

DEBUG(printg("Message complete\n"));

/*
 *  Do one of the following, depending on the kind of buffer attached
 *  to the recv port:
 *  case GO_BACK:
 *  reset the current position ptr of the recv buffer to the begin of
 *  the recv buffer itself.  Next msg will be stored at the begin
 *  of the buffer.
 *  case GO_AHEAD:
 *  If there is still room available in the attached buffer,
 *  leave the current position ptr untouched.
 *  Otherwise, reset the current position ptr of the recv buffer to
 *  the begin of the recv buffer itself; in this case, next msg will
 *  be stored at the begin of the buffer.
 */
 switch (
  par_proc_table[par_pid]->inst_table[procnum].port_table[port].buffer_kind
 ) {
  case GO_BACK:
   reset_current_ptr(par_pid,port,procnum);
   break;
  case GO_AHEAD:
   check_available_room_and_possibly_reset_current_ptr(par_pid,port,procnum);
   break;
 }

#ifdef USE_DAMNICK
/*
 *  Skip CPU-operated notification of arrival: it will be carried out by the
 *  DAMNICK device, because of the invocation of damnick_icopy_k2u_nu() .
 */
 goto set_msglen;
#endif

notif:

#ifdef USE_DAMNICK

DEBUG(printg("Notifying arrival at port %ld\n",port));

 {
 unsigned long pre_notif_val;
 pre_notify_arrival(par_pid,port,procnum,pre_notif_val);
 damnick_writel_u(
  (unsigned long *)
  (par_proc_table[par_pid]->inst_table[procnum].sem1_table)+
  par_proc_table[par_pid]->inst_table[procnum].port_table[port].semaphore,
  pre_notif_val
 );
 }

set_msglen:

#endif

/*
 *  Prepare to quit, to notify msg arrival and run the receiver handler
 *  (if a handler is present and it did not run already)
 */
 res = port;

DEBUG_E(printg(
"Setting \"gamma_msglen\" to %ld\n",
par_proc_table[par_pid]->inst_table[procnum].port_table[port].current_rx_msglen
));

 set_msglen(
  par_pid,procnum,
  par_proc_table[par_pid]->inst_table[procnum].port_table[port]
   .current_rx_msglen
 );

/*  Next msg starts from len = 0  */
 par_proc_table[par_pid]->inst_table[procnum]
 .port_table[port].current_rx_msglen = 0;

end_rx:

#ifdef MORE_INSTANCES_ON_SAME_NODE
/*
 *  Go to next packet in Rx tmp ring; bcast packets might however have
 *  to be read by other processes in this node.
 */
 if (bcast_sequence)
  rec_header_set_procnum(h,procnum+1);
#endif

end_rx_pktuntouched:

 incr_modulo(rx_tmp_ring.head,RX_TMP_RING_SIZE);

 if (res >= 0)
/*  New msg complete, so quit.  */
  break;

 if (rx_tmp_ring.head == rx_tmp_ring.tail) {
/*
 *  No further pending frames; quit the loop.
 *  The usual pkt path from NIC to user buffers can be restored.
 */
  break;
 }

/*  There is at least another pending frame.  */

DEBUG(printg("sys_gamma_read_rx_tmpbuf(): continuing loop...\n"));

continue;

};

#ifdef MORE_INSTANCES_ON_SAME_NODE
DEBUG_E(printg(
"sys_gamma_read_rx_tmpbuf(), pid %u: exiting: rx_ring_head %ld (%ld), rx_ring_tail %ld\n",
current->pid,rx_tmp_ring.head,head,rx_tmp_ring.tail
));
#else
DEBUG_E(printg(
"sys_gamma_read_rx_tmpbuf(), pid %u: exiting: rx_ring_head %ld, rx_ring_tail %ld\n",
current->pid,rx_tmp_ring.head,rx_tmp_ring.tail
));
#endif

if (rx_tmp_ring.head == rx_tmp_ring.tail) {

 if (my_procnum == PROCNUM_NONE) {
/*  "procnum" of invoker process is unknown.  Find it  */
  for (my_procnum=0; my_procnum<MAX_NUM_LOCAL_INST; my_procnum++)
   if (par_proc_table[my_par_pid]->inst_table[my_procnum].task == current)
    break;
  if (my_procnum >= MAX_NUM_LOCAL_INST)
   my_procnum = PROCNUM_NONE;	/*  not found  */
 }

DEBUG(printg(
"sys_gamma_read_rx_tmpbuf(): invoker was par_pid %d, procnum %d (pid %u mm %08lx)\n",
my_par_pid,my_procnum,current->pid,(unsigned long)current->mm
));

/*  Restore normal Rx path for incoming pkts.  */
 if (my_procnum != PROCNUM_NONE)
  par_proc_table[my_par_pid]->inst_table[my_procnum].has_pkts_in_rx_tmp_ring
   = 0;
}

#ifdef MORE_INSTANCES_ON_SAME_NODE
/*
 *  Bcast pkts:
 *  Next local receiver must re-read them; set Rx tmp ring head back!
 */
 if (bcast_sequence)
  rx_tmp_ring.head = head;
#endif

rx_tmp_ring.tail = tail;

spin_unlock(&gamma_rx_lock);
return res;


/*  Here all the unusual branches.  */

#ifdef MORE_INSTANCES_ON_SAME_NODE
found_bcast_pkt:

/*
 *  Found a bcast pkt.  We suppose it is the first of a sequence.
 *  Such pkts will not be removed from the Rx ring until all receivers
 *  have copied them into user space.  Instead, remove all the other
 *  pkts received so far (by forwarding the Rx ring head to the current pos).
 *  Also, pre-compute the next local receiver and propagate SIGPWR to it
 *  if required.
 */
/*  Is there any receiver next?  */
if (
 (procnum < MAX_NUM_LOCAL_INST-1) &&
 (par_proc_table[par_pid]->inst_table[procnum+1].whoami >= 0)
) {
/*  Yes, there is!  */

DEBUG(printg("sys_gamma_read_rx_tmpbuf():  start of bcast seq\n"));

 bcast_sequence = 1;
/*
 *  Take note of the current Rx ring head: next local receiver will have
 *  to re-read pkts from there on.
 */
 head = rx_tmp_ring.head;
/*  Signal the receiver process that it has pkts in the Rx tmp ring.  */

DEBUG(printg(
"sys_gamma_read_rx_tmpbuf(): Propagating SIGPWR to procnum %ld (bcast)\n", procnum+1
));

 force_sig(SIGPWR,par_proc_table[par_pid]->inst_table[procnum+1].task);
}
/*  If no next local receiver, treat bcast as it were unicast.  */
goto found_bcast_pkt_back;
#endif

cannot_deliver_yet:

{
 task_t *p;
/*  Is the receiver process alive?  */
 p = find_task_by_pid(receiver->pid);
 if (p == NULL)
  goto dead_process;

 if (receiver->state & (TASK_ZOMBIE | TASK_DEAD))
  goto dead_process;

DEBUG(printg(
"sys_gamma_read_rx_tmpbuf(), pid %u: recv %u not running.  Forward pkt in tmp ring...\n",
current->pid,receiver->pid
));

/*
 *  Receiver is not running.  Must move pkts after the tail of Rx tmp ring,
 *  so as to avoid stale in the ring.
 */
 rx_tmp_ring.buf[tail] = (void *)h;
/*  Forward the pseudo-tail not the real tail, to avoid endless looping.  */
 incr_modulo(tail,RX_TMP_RING_SIZE);
/*  Will remove pkt from Rx tmp ring head, but a copy will remain at tail.  */
 goto end_rx_pktuntouched;

dead_process:

/*  Receiver is no longer existing.  Discard this packet.  */

DEBUG(printg(
"sys_gamma_read_rx_tmpbuf(): recv proc does not exist.  Discarding packet...\n"
));

 goto end_rx_pktuntouched;
}

credit_request:

/*  This packet carries a credit request.  Refill credit to sender.  */
#ifdef MORE_INSTANCES_ON_SAME_NODE
/*
 *  Exception: if this is a bcast, refill credit
 *  only if this receiver is the last one of the bcast.
 */
if (!(rec_is_bcast(h))) {
#endif
/*  Refill credit and clean the credit request.  */
 gamma_kernel_send(phys_sender,DONT_MIND,FILL_CREDIT,NULL,0,1);
 rec_header_clean_creq(h);
 goto credit_request_back;
#ifdef MORE_INSTANCES_ON_SAME_NODE
}
/*
 *  Bcast packet.
 *  Refill (and clean the request) iff no other receivers (excluding this one).
 */
if (
 (procnum >= MAX_NUM_LOCAL_INST-1) ||
 (par_proc_table[par_pid]->inst_table[procnum+1].whoami < 0)
) {
 gamma_kernel_send(phys_sender,DONT_MIND,FILL_CREDIT,NULL,0,1);
 rec_header_clean_creq(h);
}
goto credit_request_back;
#endif

not_assigned:

DEBUG(
printg("sys_gamma_read_rx_tmpbuf(): h0 %8.8lx, h1 %8.8lx, h2 %8.8lx:  ",
 h[0].ll, h[1].ll, h[2].ll
);
printg("Par_pid not assigned.  Discarding frame...\n")
);

goto end_rx_pktuntouched;

bogus_inv:

DEBUG_E(printg("sys_gamma_read_rx_tmpbuf(): bogus invocation\n"));

spin_unlock(&gamma_rx_lock);
return -GAMMA_NO_RX_YET;

rec_reentering:

if (gamma_run_handler != current) {
/*
 *  A receiver has temporarily quit to run the rec handler, but will
 *  reenter soon.
 *  Prevent other processes to enter now.
 */
 spin_unlock(&gamma_rx_lock);
 return -GAMMA_ERR_BUSY;
}
/*  The receiver is reentering after running a rec handler (1sh piece).  */

DEBUG_E(printg("sys_gamma_read_rx_tmpbuf(): reentering...\n"));

gamma_run_handler = NULL;
tail = rx_tmp_ring.tail;
res = -GAMMA_RX_INCOMPLETE;

DEBUG_E(printg(
"sys_gamma_read_rx_tmpbuf(): loop: rx_ring_head %ld, rx_ring_tail %ld\n",
rx_tmp_ring.head,rx_tmp_ring.tail
));

h = (l2s *)(rx_tmp_ring.buf[rx_tmp_ring.head]);

cnk_len = rec_header_cnk_len(h);
frame_no = rec_header_frame_no(h);
phys_sender = rec_header_phys_sender(h);
rec_header_parpid_port(h,par_pid,port,procnum);

receiver =
 par_proc_table[par_pid]->inst_table[procnum].port_table[port].task;

#ifdef MORE_INSTANCES_ON_SAME_NODE
/* Check whether a seq of bcast pkts is beginning or has ended */
bcast_sequence = 0;
if (rec_is_bcast(h)) {
 if (
  (procnum < MAX_NUM_LOCAL_INST-1) &&
  (par_proc_table[par_pid]->inst_table[procnum+1].whoami >= 0)
 ) {
  bcast_sequence = 1;
  head = rx_tmp_ring.head;
 }
}
#endif

first_piece_size =
 par_proc_table[par_pid]->inst_table[procnum].port_table[port].boundary_1p;

cnk_len -= first_piece_size;

goto get_remaining_payload;

}


/*
 *  Allowed vals for "flag":
 *  0:  invoked by a recv(), check for inter-node msgs first
 *  1:  invoked by a recv(), check for intra-node msgs first
 *  2:  invoked by a send() while waiting for flow ctl credit renewal
 *      (see wait_credit_refill() in gamma_macro.h)
 */
int sys_gamma_poll
(unsigned long flag)
{

unsigned long i;
int res;
poll_common_decl_and_init;
time_586 a,b;


gamma_time(a);

#ifdef MORE_INSTANCES_ON_SAME_NODE
if (flag == 1)
 goto check_intra_node;
#endif

for (;;) {
/*  Check out the NIC rx ring (inter-node msgs)  */
 rx_read_pkt_info();
 if (!(rx_err_pktnotavail())) {
  res = rx_gamma(gamma_device,flag);
  clear_rx_irqs();
#ifdef SEQCHECK
  how_many_polls = 0;
#endif
  if (likely(res >= 0))
/*  New msg complete, or first chunk arrived and handler must run.  */
   return res;
/*
 *  res can be -GAMMA_RX_INCOMPLETE, -GAMMA_ERR_MUSTYIELD, -GAMMA_ERR_BUSY,
 *  -GAMMA_RX_SUSP
 */
  if ((res == -GAMMA_ERR_MUSTYIELD) || (res == -GAMMA_RX_SUSP))
   return res;
 }

#ifdef MORE_INSTANCES_ON_SAME_NODE

check_intra_node:

 if (shared_ring.head != shared_ring.tail) {
  res = rx_gamma_samenode(flag);
#ifdef SEQCHECK
  how_many_polls = 0;
#endif
  if (likely(res >= 0))
/*  New msg complete, or first chunk arrived and handler must run.  */
   return res;
  if ((res == -GAMMA_ERR_MUSTYIELD) || (res == -GAMMA_RX_SUSP))
   return res;
 }

#endif

 gamma_time(b);
 if (gamma_time_diff_ticks(b,a) >= REPEAT_POLL*CLOCK)
   break;
}

#ifdef SEQCHECK
if (++how_many_polls >= max_polls) {
 how_many_polls = 0;

DEBUG_L(printg(
"sys_gamma_poll():  no pkts for a long time.  Testing frame seq #s...\n"
));

DEBUG({
 unsigned long i;
 for (i=0;i<total_num_nodes;i++)
  printg("%ld ",local_credit[i]);
 printg("\n");
 for (i=0;i<total_num_nodes;i++)
  printg("%d ",rx_seq[i]);
 printg("\n");
})

/*  Ask everybody to spread their frame numbering tables.  */
 gamma_kernel_send(BROADCAST,DONT_MIND,TEST,NULL,0,1);
/*  Transmit the local frame numbering table (both unicast and bcast).  */
 for (
  i = (phys_whoami + 1) % total_num_nodes;
  i != phys_whoami;
  i = (i + 1) % total_num_nodes
 ) {
  unsigned long frame_no;

  frame_no = (unsigned long)(tx_tmp_ring[i].tx_seq);
  gamma_kernel_send(i,DONT_MIND,ACK,&frame_no,sizeof(unsigned long),1);
  frame_no = (unsigned long)(tx_tmp_ring[phys_whoami].tx_seq);
  gamma_kernel_send(i,DONT_MIND,ACKB,&frame_no,sizeof(unsigned long),1);
 }
}
#endif

return -GAMMA_NO_RX_YET;

}


/*
 *  Allowed vals for "flag":
 *  0:  invoked by a recv(), check for inter-node msgs first
 *  1:  invoked by a recv(), check for intra-node msgs first
 */
int sys_gamma_poll_nowait
(unsigned long flag)
{

unsigned long i;
int res;
#ifdef MORE_INSTANCES_ON_SAME_NODE
int res2;
#endif
poll_common_decl_and_init;


res = -GAMMA_NO_RX_YET;
#ifdef MORE_INSTANCES_ON_SAME_NODE
res2 = -GAMMA_NO_RX_YET;
#endif

#ifdef MORE_INSTANCES_ON_SAME_NODE
if (flag == 1)
 goto check_intra_node;
#endif

/*  Check out the NIC rx ring (inter-node msgs)  */
rx_read_pkt_info();
if (!rx_err_pktnotavail()) {
 res = rx_gamma(gamma_device,flag);
 clear_rx_irqs();
#ifdef SEQCHECK
 how_many_polls = 0;
#endif
 if (res >= 0)
/*  New msg complete, or first chunk arrived and handler must run.  */
  return res;
}

#ifdef MORE_INSTANCES_ON_SAME_NODE

check_intra_node:

if (shared_ring.head != shared_ring.tail) {
 res2 = rx_gamma_samenode(flag);
#ifdef SEQCHECK
 how_many_polls = 0;
#endif
 if (res2 >= 0)
/*  New msg complete, or first chunk arrived and handler must run.  */
  return res2;
}

#endif

#ifdef SEQCHECK
if (++how_many_polls >= (max_polls*10000)) {
			/*
			 *  "10000":  Adjust waiting time to take into account
			 *  that this routine does not spin over Rx (it is a
			 *  "one time only" check) and so returns quickly.
			 */
 how_many_polls = 0;

DEBUG_L(printg(
"sys_gamma_poll_nowait():  no pkts for a long time.  Testing frame seq #s...\n"
));

/*  Ask everybody to spread their frame numbering tables.  */
 gamma_kernel_send(BROADCAST,DONT_MIND,TEST,NULL,0,1);
/*  Transmit the local frame numbering table (both unicast and bcast).  */
 for (
  i = (phys_whoami + 1) % total_num_nodes;
  i != phys_whoami;
  i = (i + 1) % total_num_nodes
 ) {
  unsigned long frame_no;

  frame_no = (unsigned long)(tx_tmp_ring[i].tx_seq);
  gamma_kernel_send(i,DONT_MIND,ACK,&frame_no,sizeof(unsigned long),1);
  frame_no = (unsigned long)(tx_tmp_ring[phys_whoami].tx_seq);
  gamma_kernel_send(i,DONT_MIND,ACKB,&frame_no,sizeof(unsigned long),1);
 }
}
#endif

if (res == -GAMMA_ERR_MUSTYIELD)
 return res;
#ifdef MORE_INSTANCES_ON_SAME_NODE
if (res2 == -GAMMA_ERR_MUSTYIELD)
 return res2;
#endif
return -GAMMA_NO_RX_YET;

}


/************************************************************************
 *	Miscellaneous routines						*
 ************************************************************************/


/*
 *  Disable IRQ-driven message delivery, and block IP Tx path
 */
int sys_gamma_suspend_irq
(void)
{
	if (!atomic_read(&gamma_susp_irq_count))
		rx_suspend_nic();
	atomic_inc(&gamma_susp_irq_count);
	atomic_inc(&gamma_IP_tx_block);
	return 0;
}


/*
 *  Enable IRQ-driven message delivery, unblock IP Tx path
 */
int sys_gamma_resume_irq
(void)
{
	if (!atomic_read(&gamma_susp_irq_count))
		return 0;
	if (atomic_dec_and_test(&gamma_susp_irq_count)) {
		atomic_dec(&gamma_IP_tx_block);
		rx_resume_nic();
		return 0;
	}
	atomic_inc(&gamma_susp_irq_count);
	return -GAMMA_ERR_TRYAGAIN;
}


/*
 *  Print an integer value to "dmesg".
 */
int sys_gamma_printk
(int val)
{
	printg("sys_gamma_printk(): process %s (pid %d): %d (0x%8.8x)\n",
		current->comm,current->pid,val,val
	);
	return 0;
}


/*
 *  Set, reset, get the initial credit size.
 */
int sys_gamma_setcredit
(unsigned long size)
{
	unsigned long i;

	credit_size = size;
	for (i=0;i<total_num_nodes;i++) {
		initial_credit[i] = size;
		credit_ceil[i] = size;
		credit_growth_cnt[i] = 0;
	}
	return 0;
}

int sys_gamma_resetcredit
(void)
{
	unsigned long i;

	credit_size = MAX_CREDIT_SIZE;
	for (i=0;i<total_num_nodes;i++) {
		initial_credit[i] = MAX_CREDIT_SIZE;
		credit_ceil[i] = MAX_CREDIT_SIZE;
		credit_growth_cnt[i] = 0;
	}
	return 0;
}

int sys_gamma_getcredit
(void)
{
	unsigned long i;

	printg("sys_gamma_getcredit(): ");
	for (i=0;i<total_num_nodes;i++) {
		printg("-->%ld: %ld",i,initial_credit[i]);
		if (i<total_num_nodes-1)
			printg(", ");
	}
	printg("\n");
/*
 *  We only return the credit size towards node 0.  The complete picture has
 *  been printed by the statements above.
 */
	return initial_credit[0];
}


/*
 *  Set and reset the "max_polls" variables (threshold which triggers the
 *  collective check of frame numbering and credits.
 */
int sys_gamma_setmaxpolls
(unsigned long n)
{
	max_polls = n;
	base_max_polls = n;
	return 0;
}

int sys_gamma_resetmaxpolls
(void)
{
	max_polls = MAX_POLLS;
	base_max_polls = MAX_POLLS;
	return 0;
}

int sys_gamma_getmaxpolls
(void)
{
	return base_max_polls;
}


/************************************************************************
 *	Local routines							*
 ************************************************************************/


#define low_watermark_pt2pt(N) \
(LOW_WATERMARK < initial_credit[N] ? LOW_WATERMARK : initial_credit[N])

static inline int is_low_watermark
(unsigned long node, unsigned long credit, unsigned long curr_credit)
{
	unsigned long i;

	if (node != BROADCAST)
		return (curr_credit == low_watermark_pt2pt(node));
	for (i=0; i<total_num_nodes; i++)
		if (i != phys_whoami)
			if (
				(local_credit[i] - credit + curr_credit)
				== low_watermark_pt2pt(i)
			)
				return 1;
	return 0;
}

static inline int min_low_watermark
(unsigned long node)
{
	unsigned long min,i;

	if (node != BROADCAST)
		return low_watermark_pt2pt(node);

	min = 99999;
	for (i=0; i<total_num_nodes; i++)
		if (i != phys_whoami)
			if (low_watermark_pt2pt(i) < min)
				min = low_watermark_pt2pt(i);
	return min;
}

static inline int max_low_watermark
(unsigned long node)
{
	unsigned long max,i;

	if (node != BROADCAST)
		return low_watermark_pt2pt(node);

	max = 0;
	for (i=0; i<total_num_nodes; i++)
		if (i != phys_whoami)
			if (low_watermark_pt2pt(i) >= max)
				max = low_watermark_pt2pt(i);
	return max;
}


/*
 *  Other support routines.
 */
#include "sys_gamma_connect_port.c"
#include "sys_gamma_vm.c"
#include "sys_gamma_mem.c"
#include "sys_gamma_timers.c"
